/*
** Firmware for Marimbatron main module
** Version 4
** 
** this firmware will:
** - blink the red and green leds
** - use ISR timers to get micros
** - request sensor data using I2C (Wire1) from a sensorboard
** - convert sensorvalue to MIDI via USB
** - implement I2C communication protocol to communicate from main board to sensor board (see below for protocol)
** - implement arduinomenu library on buttons
**
** TODO
** - implement interrupt save I2C library so circular buffer for midi notes can be used properly
**   maybe this one: https://github.com/peterpanstechland/I2C_DMAC
**   or I should use https://www.arduino.cc/reference/en/language/functions/communication/serial/availableforwrite/
**   so that the USBMIDI doesn't block the main loop?
** - prevent retrigger when pad is pressed and hold
**
** - UI (when LCD display works...):
** - - Track last 4 played notes (South - West - North - East layout)
** - - Indicate velocity
** - - settings menu
** - - - show current sensor of last hit sensor
** - - - read/set thresholds
** - - - auto-set thresholds
** - - - set auto-threshold value
** - - - read/set max hit level per sensor (hitting with this value should send midi velocity 127)
** - - - read/set scantime
** - - - read/set masktime
**
**
**
** Required libraries:
** wire.h, MIDIUSB.h
**
** I2C Marimbatron protocol:
** C = type: Command to execute - followed by character that defines what command
** S = type: Set - followed by character that defines what to set (sometimes followed by uint_8 to indicate sensor number) 
** G = type: Get - followed by character that defines what to get (sometimes followed by uint_8 to indicate sensor number)

** main sends chars XR -> peripheral will reset (NVIC_SystemReset())
** main sends chars XA -> peripheral will do auto-threshold procedure
** main sends chars XS -> peripheral will stop sensor measurements. All detected hits that we're not sent yet will be flushed.

** main sends chars GH -> peripheral is set into Get sensor Hit mode -> 
                                                      on datarequest (3 bytes) it sends sensorvalue of sensor with hit = true (or 255,0,0 if no hit)
                                                      All detected hits that we're not sent yet will be flushed.
** main sends chars GC followed by byte sensornumber -> peripheral is set into Get Continuous actual value -> on datarequest (3 bytes) it sends sensorvalue of [sensornumber]
                                                      and on next datarequest it sends sensorvalue of sensor with hit = true (or 255,0,0 if no hit)
                                                      on next datarequest back to sensorvalue of [sensornumber], etc.
                                                      -> most probably not needed as GA already gets actual value of a single senso
** main sends chars GA followed by byte sensornumber -> peripheral set into Get single Actual values mode -> 
                                                        on datarequest (2 bytes) it sends actual sensorvalue of [sensornumber].
** main sends chars GT followed by byte sensornumber -> peripheral set to Get Threshold mode -> 
                                                        on datarequest (2 bytes) it sends current threshold value of [sensornumber].
** main sends chars GX followed by byte sensornumber -> peripheral set to Maximum mode -> 
                                                        on datarequest (2 bytes) it sends current maximum value of [sensornumber].
** main sends chars GM -> peripheral set into Masktime mode -> 
                                                        on datarequest it sends the current masktime.
** main sends chars GN -> peripheral set into Scantime mode -> 
                                                        on datarequest it sends the current scantime.

** main sends chars ST followed by byte sensornumber followed by uint_16 value -> peripheral sets threshold of sensor [sensornumber] to [value].
** main sends chars SX followed by byte sensornumber followed by uint_16 value -> peripheral sets maximum of sensor [sensornumber] to [value].
** main sends chars SM followed by byte sensornumber followed by uint_32 value -> peripheral sets Masktime of sensor [sensornumber] to [value].
** main sends chars SN followed by byte sensornumber followed by uint_32 value -> peripheral sets Scantime of sensor [sensornumber] to [value].
** main sends chars SL followed by byte layoutnumber -> peripheral sets layout to [value].
**                                                    layout 0 = wicki-hayden
**                                                    layout 1 = piano
**                                                    layout 2 = all off
** main sends chars SC followed by byte sensornumber followed by 3 bytes RGB value -> peripheral sets color of sensor [sensornumber] to RGB [value].
**
** This work may be reproduced, modified, distributed, performed, and 
** displayed for any non-commercial, non-profit purpose, but must acknowledge 
** the developer and the Marimbatron project. 
** Copyright is retained and must be preserved. The work is 
** provided as is; no warranty is provided, and users accept all liability.
** 
** Copyright 2024 Leo Kuipers.
*/


/*
pin assignments based on variant.h
using one uart, one wire, one spi, USB Config: WITH_CDC:

UART
PIN_USB_DM (24ul)-> PA24
PIN_USB_DP (25ul)-> PA25

I2C (sercom 1)
PIN_WIRE_SDA (16u)-> PA16
PIN_WIRE_SCL (17u)-> PA17

SPI (sercom3)
PIN_SPI_MISO (22u)-> PA22
PIN_SPI_MOSI (18u)-> PA18
PIN_SPI_SCK (19u)-> PA19
PIN_SPI_SS (23u)-> PA23
(HW SS isn't used)
*/
#define FIRMWAREVERSION 3
#define DEBUG
bool showHitInformation = false;

#define REDLEDPIN PORT_PA00 // PA00
#define GREENLEDPIN PORT_PA01 // PA01

#include <Wire.h>
#include "MIDIUSB.h"

// timers
// CPU clock is 48 MHz (48.000.000 Hz)
// so 1 CPU clock cycle is about 20,83ns. 
// timer_hander is called every time the timer overflows.
// which is here programmed as COUNT_PER_TICK * 1_CPU_CLOCKCYCLE.
// for a 100 us timer, COUNT_PER_TICK = 10.000 ticks because
// 48.000.000 / 10.000 = 4.800 and 4.800 * 20,83ns = 99,984us which is about 100us.

#define CPU_HZ 48000000 // The cpu clock is running at 48 MHz
#define TICK_HZ_100US 10000 // 100 us tick
#define COUNTS_PER_TICK_100US (CPU_HZ / TICK_HZ_100US) // number of clock cycles per tick (4800)
#define COUNTS_PER_TICK_TC5 60009 // This is the number of clock cycles per tick. TC5 is set up as a 16 bit timer with prescaler of 4, so this results in a 5ms timer.
#define COUNTS_PER_TICK_TC5 12000 // This is the number of clock cycles per tick. TC5 is set up as a 16 bit timer with prescaler of 4, so this results in a 1ms timer.

volatile uint32_t micros_counter = 0; // holds the count of microseconds since the MCU started (don't want to use micros())
const unsigned long intervalEvent1 = 5000; // in microseconds
const unsigned long intervalEvent2 = 30000; // in microseconds
// const unsigned long intervalEvent3 = 100000; // in microseconds
const unsigned long intervalEvent3 = 200000; // in microseconds
const unsigned long intervalEvent4 = 500000; // in microseconds
unsigned long previousTimeEvent1 = 0;
unsigned long previousTimeEvent2 = 0;
unsigned long previousTimeEvent3 = 0;
unsigned long previousTimeEvent4 = 0;
const uint8_t sensorArrayLength = 19;
bool requestSingleSensorActualValue = false;

/* ArduinoMenu */
#include <menu.h>
#include <menuIO/serialOut.h>
#include <menuIO/serialIn.h>
using namespace Menu;

// function prototypes so they can be used in the menu but defined later in the code
result menuGetSensorsActualValue();
result menuGetSingleSensorActualValue();
result menuGetSensorsThreshold();
result menuGetSensorsMaximum();
result menuExecuteAutoThreshold();
result menuExecuteReset();
result enableGetSingleSensorActualValue(bool state);

result zZz() {Serial.println("zZz");return proceed;}

TOGGLE(showHitInformation,setHitInfo,"Show hit info: ",doNothing,noEvent,noStyle//,doExit,enterEvent,noStyle
  ,VALUE("On",HIGH,doNothing,noEvent)
  ,VALUE("Off",LOW,doNothing,noEvent)
);

MENU(mainMenu,"*M* Main menu",zZz,noEvent,wrapStyle
  ,OP("Read all actual values",menuGetSensorsActualValue,enterEvent)
  ,OP("Read thresholds",menuGetSensorsThreshold,enterEvent)
  ,OP("Read maximums",menuGetSensorsMaximum,enterEvent)
  ,OP("Execute Auto threshold",menuExecuteAutoThreshold,enterEvent)
  ,OP("Reboot system",menuExecuteReset,enterEvent)
  ,SUBMENU(setHitInfo)
  ,EXIT("Exit menu")
);

#define MAX_DEPTH 1
serialIn serial(Serial); // menu input
MENU_OUTPUTS(out,MAX_DEPTH
  ,SERIAL_OUT(Serial)
  ,NONE//must have 2 items at least
);
NAVROOT(nav,mainMenu,MAX_DEPTH,serial,out);

result idle(menuOut &o, idleEvent e) { // when idle the menu+ and menu- buttons are used to change the octave
  switch(e) {
    // case idleStart:o.println("suspending menu!");break;
    case idling:o.println("*S* menu suspended... Hit enter to wake up.");break;
    // case idleEnd:o.println("resuming menu.");
      nav.reset();
      break;
  }
  return proceed;
}

uint32_t ledIndicatorTime = 0; // timer that holds info if hit-indicator led should be turned off

const uint32_t midiNoteOffTimeout = 100000; // 100ms = 100000us
const uint8_t midiChannel = 0;

const uint8_t numberOfSensorBoards = 4;
uint8_t I2CSensorboardAddresses[numberOfSensorBoards]; // array to hold the addresses of the sensorboards
bool requestHitData = false;
bool enableRequestSensorData = false;
uint8_t lastSelectedSensorboard = 0;
uint8_t lastSelectedSensor = 0;
bool getSingleSensorActualValue = false;
#define WICKIHAYDEN 0
#define PIANO 1
#define ALL_OFF 2

uint8_t currentLayout = ALL_OFF; // assign ALL OFF note layout = default
uint8_t currentOctave = 4; // Current octave refers to the octave of the first sensorboard. octave 4 is the standard center octave. C4 is MIDI note 60 (middle C).

// midi notes mapped to wicki hayden layout:
const int8_t noteArrayWickiHayden[sensorArrayLength] =  // int8_t is signed integer, values from -128 to 127. Midi defines 128 notes, from 0 to 127.
{
  -1, //s0 -> Cb-3 = B-3(not existing in midi)
  1, //s1 -> Dd-2 (C#-2)
  3, //s2 -> Eb-2 (D#-2)
  5, //s3 -> F-2
  7, //s4 -> G-2
  9, //s5 -> A-2
  11, //5s6 -> B-2
  13, //s7 -> C#-1 (Db-1)
  15, //s8 -> D#-1 (Eb-1)
  17, //s9 -> E#-1 = F

  6, //s10 -> Gb-2 (F#-2)
  8, //s11 -> Ab-2 (G#-2)
  10, //s12 -> Bb-2 (A#-2)
  12, //s13 -> C-1
  14, //s14 -> D-1
  16, //s15 -> E-1
  18, //s16 -> F#-1 (Gb)
  20, //s17 -> G#-1 (Ab)
  22, // s18 -> A#-1 (Bb)
};
// midi notes mapped to piano layout. -128 means note is disabled:
const int8_t noteArrayPiano[sensorArrayLength] = 
{
  0, //53, // s0 -> C-1
  2, //55, // s1 -> D-1
  4, //57, // s2 -> E-1
  5, //59, // s3 -> F-1
  7, //60, // s4 -> G-1
  9, //62, // s5 -> A-1
  11, //64, // s6 -> B-1
  12, //65, // s7 -> C0
  14, //67, // s8 -> G0
  16, //69, // s9 -> A0

  1, //54, // s10 -> C#-1
  3, //56, // s11 -> D#-1
  -128, //58, // s12 -> none
  6, // s13 -> F#-1
  8, //73, // s14 -> G#-1
  10, //75, // s15 -> A#-1
  -128, // s16 -> none
  13, //66 s17 -> C#0
  15 //68 s18 -> D#0
};

const uint16_t defaultMaximum[numberOfSensorBoards][sensorArrayLength] = 
{
  { // sensorboard 0
    390, // sensor 0
    250, // sensor 1
    690, // sensor 2
    350, // sensor 3
    450, // sensor 4
    575, // sensor 5
    1015, // sensor 6
    670, // sensor 7
    825, // sensor 8
    650, // sensor 9

    550, // sensor 10
    550, // sensor 11
    560, // sensor 12
    440, // sensor 13
    450, // sensor 14
    1200, // sensor 15
    660, // sensor 16
    670, // sensor 17
    670  // sensor 18 
  },
  { // sensorboard 1
    600, // sensor 0
    410, // sensor 1
    600, // sensor 2
    650, // sensor 3
    500, // sensor 4
    750, // sensor 5
    625, // sensor 6
    390, // sensor 7
    585, // sensor 8
    475, // sensor 9

    500, // sensor 10
    400, // sensor 11
    775, // sensor 12
    540, // sensor 13
    420, // sensor 14
    615, // sensor 15
    490, // sensor 16
    700, // sensor 17
    385  // sensor 18 
  },
  { // sensorboard 2
    325, // sensor 0
    225, // sensor 1
    475, // sensor 2
    600, // sensor 3
    390, // sensor 4
    600, // sensor 5
    750, // sensor 6
    745, // sensor 7
    275, // sensor 8
    625, // sensor 9

    380, // sensor 10
    525, // sensor 11
    340, // sensor 12
    325, // sensor 13
    350, // sensor 14
    825, // sensor 15
    560, // sensor 16
    1015, // sensor 17
    925  // sensor 18 
  },
  { // sensorboard 3
    490, // sensor 0
    550, // sensor 1
    350, // sensor 2
    325, // sensor 3
    310, // sensor 4
    325, // sensor 5
    340, // sensor 6
    360, // sensor 7
    475, // sensor 8
    400, // sensor 9

    250, // sensor 10
    605, // sensor 11
    390, // sensor 12
    480, // sensor 13
    1015, // sensor 14
    350, // sensor 15
    690, // sensor 16
    440, // sensor 17
    575  // sensor 18 
  }
};

uint8_t noteArray[numberOfSensorBoards][sensorArrayLength];  // array to hold currently assigned notes
uint16_t sensorThreshold[numberOfSensorBoards][sensorArrayLength];
uint16_t sensorMaximum[numberOfSensorBoards][sensorArrayLength];
uint16_t sensorLastActualValue[numberOfSensorBoards][sensorArrayLength];

// arrays to keep track of the sensor note status (will be set to true when hit and set to false after midi note-off timeout)
bool sensorHit[numberOfSensorBoards][sensorArrayLength];

// arrays to keep track of the last time (micros_counter) that a hit was detected on a sensor
uint32_t sensorHitTime[numberOfSensorBoards][sensorArrayLength];

// circular buffer for midi notes
const int MIDIBufferSize = 64;
struct buffer {
  uint8_t sensorboard;
  uint8_t sensorNumber;
  uint16_t sensorVelocity;
};

buffer MIDIBuffer[MIDIBufferSize];

volatile uint8_t MIDIBufferHead = 0;
volatile uint8_t MIDIBufferTail = 0;

void TC4_Handler() { // 100us timer, priority 0
  micros_counter = micros_counter + 100; // add 100 us to the micros_counter
  // Serial.print("micros_counter: ");Serial.println(micros_counter);
  TC4->COUNT16.INTFLAG.bit.MC0 = 1;  // Clear the interrupt
}

void TC5_Handler() { // request sensor data, priority 1
  if (enableRequestSensorData) {
    requestHitData = true;
  }
  TC5->COUNT16.INTFLAG.bit.MC0 = 1;  // Clear the interrupt
}

void setup() {
  uint8_t returnValue = 0;
  uint8_t numberOfI2CDevices = 0;
  // set SAMD to run at 48Mhz
  setup48MhzUSBClockRecoveryMode();
  setupHardwareTimers();  

  PORT->Group[0].DIRSET.reg = REDLEDPIN; // set led pin as output
  PORT->Group[0].DIRSET.reg = GREENLEDPIN; // set led pin as output
  PORT->Group[0].OUTSET.reg = REDLEDPIN; // turn RED on
  PORT->Group[0].OUTSET.reg = GREENLEDPIN; // turn GREEN on
  delay(200);
  PORT->Group[0].OUTCLR.reg = GREENLEDPIN; // turn GREEN off


  Serial.begin(115200);
  while (!Serial && micros_counter < 5000000); // wait for serial bus to activate (in order to communicate immediately after reset) or timeout
  delay(400);
  Serial.print("*R* Marimbatron Main board firmware v");
  Serial.println(FIRMWAREVERSION);Serial.flush();

  // set PA10 to input, enabling pull resistor and setting it to pull-up.
  PORT->Group[0].DIRCLR.reg = PORT_PA10; // change layout button
  PORT->Group[0].PINCFG[10].reg = PORT_PINCFG_PULLEN | PORT_PINCFG_INEN;
  PORT->Group[0].OUTSET.reg = PORT_PA10;
  // set PA11 to input, enabling pull resistor and setting it to pull-up.
  PORT->Group[0].DIRCLR.reg = PORT_PA11; // menu- button
  PORT->Group[0].PINCFG[11].reg = PORT_PINCFG_PULLEN | PORT_PINCFG_INEN;
  PORT->Group[0].OUTSET.reg = PORT_PA11;
  // set PA14 to input, enabling pull resistor and setting it to pull-up.
  PORT->Group[0].DIRCLR.reg = PORT_PA14; // menu+ button
  PORT->Group[0].PINCFG[14].reg = PORT_PINCFG_PULLEN | PORT_PINCFG_INEN;
  PORT->Group[0].OUTSET.reg = PORT_PA14;
  // set PA15 to input, enabling pull resistor and setting it to pull-up.
  PORT->Group[0].DIRCLR.reg = PORT_PA15; // menu enter button
  PORT->Group[0].PINCFG[15].reg = PORT_PINCFG_PULLEN | PORT_PINCFG_INEN;
  PORT->Group[0].OUTSET.reg = PORT_PA15;

  nav.timeOut=5;
  nav.idleTask=idle; //point a function to be used when menu is suspended
  nav.idleOn();// menu will start on idle state, press select to enter menu

  // Serial.println("init sensorHit and SensorHittime");
  // init arrays sensorHit[x][y] and sensorHitTime[x][y]
  for (uint8_t sensorboard = 0; sensorboard < numberOfSensorBoards; sensorboard++) { 
    for (uint8_t sensor = 0; sensor < sensorArrayLength; sensor++) {
      sensorHit[sensorboard][sensor] = false; // set all sensor hits to false
      sensorHitTime[sensorboard][sensor] = 0; // set all sensor hit times to 0
    }
  }
  Serial.println("checkAndSolveI2CDeadlock()");
  checkAndSolveI2CDeadlock();
  Wire.begin(); // join i2c bus as master (so no address is set)
  // Wire.setClock(10000); // 10kBaud -> low speed mode
  // Wire.setClock(100000); // 100kBaud -> standard speed mode
  Wire.setClock(400000); // 400kBaud -> high speed mode
  delay(1000);  
  numberOfI2CDevices = scanI2Cbus();
  delay(1000);  
  if (numberOfSensorBoards == numberOfI2CDevices) {
    Serial.println("found following sensorboards:");
    for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) {
      Serial.print(sensorboard);Serial.print(" @ ");Serial.println(I2CSensorboardAddresses[sensorboard]);
    }
    PORT->Group[0].OUTSET.reg = GREENLEDPIN; // turn GREEN on

    enableSensorMeasurements(false); // disable sensor measurements for all sensorboards
    delay(100);

    // set sensorthresholds and maximums
    // for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) { // iterate over the 4 sensor boards
    //   for (uint8_t sensor = 0; sensor < sensorArrayLength; sensor++) { // iterate over all 19 sensors
    //     setSensorThreshold(sensorboard, sensor, 1000);
    //     // delay(10);
    //   }
    // }
    // delay(100);
    
    // get sensorthresholds from sensorboards
    getSensorsThreshold();

    for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) { // iterate over the 4 sensor boards
      for (uint8_t sensor = 0; sensor < sensorArrayLength; sensor++) { // iterate over all 19 sensors
        setSensorMaximum(sensorboard, sensor, defaultMaximum[sensorboard][sensor]);
      }
    }

    setLayout(WICKIHAYDEN, true);
    PORT->Group[0].OUTCLR.reg = REDLEDPIN; // turn RED off
    PORT->Group[0].OUTCLR.reg = GREENLEDPIN; // turn GREEN off
    Serial.println("Setup done.");
    Serial.print("Initialization took ");Serial.print(micros_counter/1000);Serial.println(" ms.");
    enableSensorMeasurements(true);
  }
  else {
    Serial.print("Error connecting to sensorboards. Found "); Serial.print(numberOfI2CDevices);Serial.print(" instead of ");Serial.println(numberOfSensorBoards);
    while(1) { // error reading number I2C devices. Do not proceed.
      PORT->Group[0].OUTTGL.reg = REDLEDPIN; // toggle RED
      delay(100);
    }
  }
}

void loop() {
  unsigned long currentTime = micros_counter;
  static bool btn0State = 0;
  static bool btn1State = 0;
  static bool btn2State = 0;
  static bool btn3State = 0;
  static bool btn0StateOld = 0;
  static bool btn1StateOld = 0;
  static bool btn2StateOld = 0;
  static bool btn3StateOld = 0;

  if (requestHitData) {
    requestHitData = false;
    I2CRequestSensorData();
  }
  processMIDIBuffer(); // process incoming midi messages as fast as possible

  /* This is event 1 -> request sensor data */
  if ((unsigned long)(currentTime - previousTimeEvent1) >= intervalEvent1) {
    // if (enableRequestSensorData) {
    //   I2CRequestSensorData();
    // }
    // if (getSingleSensorActualValue) {
    //   uint16_t sensorVelocity;
    //   Wire.requestFrom(I2CSensorboardAddresses[lastSelectedSensorboard], 2); // requesting 3 bytes from sensorboard. Max number of bytes in a single transmission cannot be more than 32.
    //   Serial.print("Sensorboard ");Serial.print(lastSelectedSensorboard);Serial.print(" Sensor ");
    //   Serial.print(lastSelectedSensor);Serial.print(" -> ");Serial.print(Wire.read() | (Wire.read() << 8));
    // }

   /* Update the timing for the next time around */
    previousTimeEvent1 = currentTime;
  }

  /* This is event 2 -> request button states  */
  if ((unsigned long)(currentTime - previousTimeEvent2) >= intervalEvent2) {
    btn0State = ~PORT->Group[0].IN.reg & PORT_PA10 ? 1 : 0; // read value of input pin. Input has pull-up so the actual value is inverted. this converts back.
    btn1State = ~PORT->Group[0].IN.reg & PORT_PA14 ? 1 : 0; // read value of input pin. Input has pull-up so the actual value is inverted. this converts back.
    btn2State = ~PORT->Group[0].IN.reg & PORT_PA11 ? 1 : 0; // read value of input pin. Input has pull-up so the actual value is inverted. this converts back.
    btn3State = ~PORT->Group[0].IN.reg & PORT_PA15 ? 1 : 0; // read value of input pin. Input has pull-up so the actual value is inverted. this converts back.

    if (!nav.sleepTask) {
      nav.poll(); // deals with all I/O, driving the menu on a full automated mode.
    }
    // handle menu buttons
    if (btn0State != btn0StateOld) {
      // Serial.print("Button 0: ");Serial.println(btn0State);
      if (btn0State) {
        if (currentLayout == WICKIHAYDEN) {
          Serial.println("*L* Piano");
          setLayout(PIANO, true);
        }
        else if (currentLayout == PIANO) {
          Serial.println("*L* Isomorphic");
          setLayout(WICKIHAYDEN, true);
        }
        else {
          Serial.println("*L* Isomorphic");
          setLayout(WICKIHAYDEN, true);
        }
      }
      btn0StateOld = btn0State;
    }
    if (btn1State != btn1StateOld) {
      // Serial.print("Button 1: ");Serial.println(btn1State);
      if (btn1State) {
        if (nav.sleepTask) {
          octaveUp();
        }
        else {
          nav.doNav(downCmd);
        }
      }
      btn1StateOld = btn1State;
    }
    if (btn2State != btn2StateOld) {
      // Serial.print("Button 2: ");Serial.println(btn2State);
      if (btn2State) {
        if (nav.sleepTask) {
          octaveDown();
        }
        else {
          nav.doNav(upCmd);
        }
      }
      btn2StateOld = btn2State;
    }
    if (btn3State != btn3StateOld) {
      // Serial.print("Button 3: ");Serial.println(btn3State);
      if (btn3State) {
        nav.doNav(enterCmd);
      }
      btn3StateOld = btn3State;
    }
   /* Update the timing for the next time around */
    previousTimeEvent2 = currentTime;
  }


  /* This is event 3 -> check for and send noteoff */
  if ((unsigned long)(currentTime - previousTimeEvent3) >= intervalEvent3) {
    if ((unsigned long)(currentTime - ledIndicatorTime) >= midiNoteOffTimeout) {
      PORT->Group[0].OUTCLR.reg = REDLEDPIN; // turn led pin off
    }
    for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) { // iterate over the 4 sensor boards
      for (uint8_t sensor = 0; sensor < sensorArrayLength; sensor++) { // iterate over all 19 sensors
        if (sensorHit[sensorboard][sensor]) {
          if ((unsigned long)(currentTime - sensorHitTime[sensorboard][sensor]) >= midiNoteOffTimeout) {
            noteOff(midiChannel, noteArray[sensorboard][sensor], 0); // send note off
            sensorHit[sensorboard][sensor] = false;
          };
        }
      };
    };

   /* Update the timing for the next time around */
    previousTimeEvent3 = currentTime;
  }  


  /* This is event 4 -> blink Green LED */
  if ((unsigned long)(currentTime - previousTimeEvent4) >= intervalEvent4) {
    PORT->Group[0].OUTTGL.reg = GREENLEDPIN; // toggle green led pin as indicator that the main controller is still operating properly

   /* Update the timing for the next time around */
    previousTimeEvent4 = currentTime;
  }
}

void processMIDIBuffer() {  
  // read midi buffer:
  // send midi note on, making sure to get to the correct octave. 
  // only register the hit when there's a midi note number defined on this sensor
  // (midi note number -128 is used to "disable" the sensor)
  uint8_t sensorboard;
  uint8_t sensorNumber;
  uint16_t sensorVelocity;
  while (MIDIBufferTail != MIDIBufferHead) {
    // Serial.print("process midi buffer head "); Serial.print(MIDIBufferHead);Serial.print(" tail ");Serial.println(MIDIBufferTail);
    sensorboard = MIDIBuffer[MIDIBufferTail].sensorboard;
    sensorNumber = MIDIBuffer[MIDIBufferTail].sensorNumber;
    sensorVelocity = MIDIBuffer[MIDIBufferTail].sensorVelocity;

    if (noteArray[sensorboard][sensorNumber] != -128) { // only send midi note on if note is not disabled (-128)
      noteOn(
        midiChannel, 
        noteArray[sensorboard][sensorNumber],
        mapToMidiVelocity(
          sensorVelocity,
          sensorThreshold[sensorboard][sensorNumber], 
          sensorMaximum[sensorboard][sensorNumber]
        )
      ); 
      sensorHit[sensorboard][sensorNumber] = true; // store that sensor hit was detected
      sensorHitTime[sensorboard][sensorNumber] = micros_counter; // store what time the hit was detected
      if (showHitInformation) {
        Serial.print("*H* sending midi: ch"); Serial.print(midiChannel); Serial.print(" note "); 
        Serial.print(noteArray[sensorboard][sensorNumber]); 
        Serial.print(" velocity "); 
        Serial.println(
          mapToMidiVelocity(
            sensorVelocity,
            sensorThreshold[sensorboard][sensorNumber], 
            sensorMaximum[sensorboard][sensorNumber]
          )
        );
      }      
    }
    MIDIBufferTail = (MIDIBufferTail + 1) % MIDIBufferSize;
  }
}

void addToMIDIBuffer(uint8_t board, uint8_t sensor, uint16_t velocity) {
  uint8_t nextHead = (MIDIBufferHead + 1) % MIDIBufferSize;
  // Serial.print("Adding to midi buffer head "); Serial.print(MIDIBufferHead);Serial.print(" tail ");Serial.print(MIDIBufferTail);
  // Serial.print(" board ");Serial.print(board);Serial.print(" sensor ");Serial.print(sensor);Serial.print(" velocity ");Serial.println(velocity);  
  if (nextHead != MIDIBufferTail) { // Check if buffer is full
    MIDIBuffer[MIDIBufferHead].sensorboard = board;
    MIDIBuffer[MIDIBufferHead].sensorNumber = sensor;
    MIDIBuffer[MIDIBufferHead].sensorVelocity = velocity;
    MIDIBufferHead = nextHead;
  } else {
    // Buffer is full, give a warning
    Serial.println("WARNING: MIDI buffer full");
  }
}

void I2CRequestSensorData() {
  uint8_t sensorNumber = 255;
  uint16_t sensorVelocity=0;
  // Serial.println("requesting I2C sensordata");
  for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) { // iterate over the sensor boards
    Wire.requestFrom(I2CSensorboardAddresses[sensorboard], 3); // requesting 3 bytes from sensorboard. Max number of bytes in a single transmission cannot be more than 32.
    // Serial.print("requesting I2C from address ");
    // Serial.println(I2CSensorboardAddresses[sensorboard]);
    // Serial.println(Wire.available());
    sensorNumber = Wire.read(); // Wire.read reads 1 byte at a time. uint8_t is 1 byte.
    sensorVelocity = Wire.read() | (Wire.read() << 8); // sensorVelocity is uint16_t, which is 2 bytes. So need to combine 2 reads into one uint16_t.
    // Serial.print(sensorNumber);         // Print the value
    // Serial.print(" ");
    // Serial.println(sensorVelocity);

    if (sensorNumber < 255) { // if returned sensorNumber = 255 there is no hit
      if (showHitInformation) {
        Serial.print("*H* Hit on sensorboard "); Serial.print(sensorboard); Serial.print(" at address "); 
        Serial.print(I2CSensorboardAddresses[sensorboard]); Serial.print(" sensor "); Serial.print(sensorNumber); 
        Serial.print(" -> "); Serial.println(sensorVelocity);
      }
      // register the hit in the midibuffer
      ledIndicatorTime = micros_counter;
      PORT->Group[0].OUTSET.reg = REDLEDPIN; // turn RED on

      lastSelectedSensorboard = sensorboard;
      lastSelectedSensor = sensorNumber;

      // register the hit in the MIDIBuffer
      addToMIDIBuffer(sensorboard, sensorNumber, sensorVelocity);  
    }
  }
}

void octaveUp() {
  uint8_t layout = currentLayout;
  if (currentOctave <= 5) {
    currentOctave = currentOctave + 1;
    assignNotes();
  }
  setSensorColor(0, currentOctave-1, 128, 128, 255); // indicator on sensorboard 0
  delay(100);
  setLayout(currentLayout, true);
}

void octaveDown() {
  uint8_t layout = currentLayout;
  if (currentOctave >= 2) {
    currentOctave = currentOctave - 1;
    assignNotes();
  }
  setSensorColor(0, currentOctave-1, 128, 128, 255); // indicator on sensorboard 0
  delay(100);
  setLayout(layout, true);
  // enableSensorMeasurements(true); 
}

void assignNotes() {
  for (uint8_t sensorboard = 0; sensorboard < numberOfSensorBoards; sensorboard++) { 
    for (uint8_t sensor = 0; sensor < sensorArrayLength; sensor++) {
      if (currentLayout == 0) {
        noteArray[sensorboard][sensor] = noteArrayWickiHayden[sensor] + (currentOctave * 12) + (sensorboard * 12); // add 12 per higher octave
      }
      else if (currentLayout == 1) {
        if (noteArrayPiano[sensor] != -128) {
          noteArray[sensorboard][sensor] = noteArrayPiano[sensor] + (currentOctave * 12) + (sensorboard * 12); // add 12 per higher octave
        }
        else {
          noteArray[sensorboard][sensor] = -128; // disable note
        }
      }
      else if (currentLayout == 2) {
        noteArray[sensorboard][sensor] = -128; // disable note
      }
    }
  }
}

void enableSensorMeasurements(bool state) {
  for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) {
    Wire.beginTransmission(I2CSensorboardAddresses[sensorboard]); // transmit to device with address i
    if (state) {
      Wire.write('G'); // GH = Get Sensor Hit (start measurements)
      Wire.write('H'); // 
      enableRequestSensorData = true;
    }
    else {
      Wire.write('X'); // XS = execute stop measurements
      Wire.write('S'); // 
      enableRequestSensorData = false;
    }
    uint8_t returnValue = Wire.endTransmission(); // send I2C buffer 
    delay(1);
    if (returnValue > 0) { // someting went wrong
      Serial.print("return value of I2C transmission to "); Serial.print(sensorboard);Serial.print(" GS enableSensorMeasurements "); Serial.print(state);Serial.print(" = ");Serial.println(returnValue);
    }
  }
}

void executeReset() {
  for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) {
    Wire.beginTransmission(I2CSensorboardAddresses[sensorboard]); // transmit to device with address i
    Wire.write('X'); // XR = execute reset
    Wire.write('R'); // 
    enableRequestSensorData = false;
    uint8_t returnValue = Wire.endTransmission(); // send I2C buffer 
    if (returnValue > 0) { // someting went wrong
      Serial.print("return value of I2C transmission to "); Serial.print(sensorboard);Serial.print(" XR execute reset ");Serial.print(" = ");Serial.println(returnValue);
    }
  }
}

void executeAutoThreshold() {
  for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) {
    Wire.beginTransmission(I2CSensorboardAddresses[sensorboard]); // transmit to device with address i
    Wire.write('X'); // XR = execute auto thresold
    Wire.write('A'); // 
    enableRequestSensorData = false;
    uint8_t returnValue = Wire.endTransmission(); // send I2C buffer 
    if (returnValue > 0) { // someting went wrong
      Serial.print("return value of I2C transmission to "); Serial.print(sensorboard);Serial.print(" XA execute auto threshold ");Serial.print(" = ");Serial.println(returnValue);
    }
  }
}

void getSensorsActualValue() {
  uint16_t currentValue = 0;
  for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) {    
    for (uint8_t sensor = 0; sensor < sensorArrayLength; sensor++) {    
      Wire.beginTransmission(I2CSensorboardAddresses[sensorboard]); // transmit to device with address i
      Wire.write('G'); // GT = Get Sensor actual value
      Wire.write('A'); // 
      Wire.write(sensor); // get info of this sensor number
      uint8_t returnValue = Wire.endTransmission(); // send I2C buffer 
      if (returnValue > 0) { // someting went wrong
        Serial.print("return value of I2C transmission to "); Serial.print(sensorboard);Serial.print(" GA getSensorsActualValue "); Serial.print(sensor);Serial.print(" = ");Serial.println(returnValue);
      }
      delay(10);
      Wire.requestFrom(I2CSensorboardAddresses[sensorboard], 2); // requesting 2 bytes from sensorboard
      currentValue = Wire.read() | (Wire.read() << 8); // currentValue is uint16_t, which is 2 bytes. So need to combine 2 reads into one uint16_t.
      // Serial.print("Actual value of sensorboard ");Serial.print(sensorboard);Serial.print(" sensor ");Serial.print(sensor);Serial.print(" = ");Serial.println(currentValue);
      sensorLastActualValue[sensorboard][sensor] = currentValue; // set sensor actual value array item to received value
      delay(10);
    }
  }
}

result enableGetSingleSensorActualValue(bool state) {
  Serial.print("enableGetSingleSensorActualValue");
  // if (state) {
  //   enableSensorMeasurements(false);
  //   Wire.beginTransmission(I2CSensorboardAddresses[lastSelectedSensorboard]); // transmit to device with address i
  //   Wire.write('G'); // GT = Get Sensor actual value
  //   Wire.write('A'); // 
  //   Wire.write(lastSelectedSensor); // get info of this sensor number
  //   uint8_t returnValue = Wire.endTransmission(); // send I2C buffer 
  //   if (returnValue > 0) { // someting went wrong
  //     Serial.print("return value of I2C transmission to "); Serial.print(lastSelectedSensorboard);
  //     Serial.print(" GA getSensorsActualValue "); Serial.print(lastSelectedSensor);Serial.print(" = ");Serial.println(returnValue);
  //   }
  //   delay(10);
  //   getSingleSensorActualValue = true;
  // }
  // else {
  //   getSingleSensorActualValue = false;
  //   enableSensorMeasurements(true);
  // }
  return proceed;
}

void getSensorsThreshold() {
  uint16_t currentThreshold = 0;
  for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) {    
    for (uint8_t sensor = 0; sensor < sensorArrayLength; sensor++) {    
      Wire.beginTransmission(I2CSensorboardAddresses[sensorboard]); // transmit to device with address i
      Wire.write('G'); // GT = Get Sensor Threshold
      Wire.write('T'); // 
      Wire.write(sensor); // get info of this sensor number
      uint8_t returnValue = Wire.endTransmission(); // send I2C buffer 
      if (returnValue > 0) { // someting went wrong
        Serial.print("return value of I2C transmission to "); Serial.print(sensorboard);Serial.print(" GT getSensorsThreshold "); Serial.print(sensor);Serial.print(" = ");Serial.println(returnValue);
      }
      delay(10);
      Wire.requestFrom(I2CSensorboardAddresses[sensorboard], 2); // requesting 2 bytes from sensorboard
      currentThreshold = Wire.read() | (Wire.read() << 8); // currentThreshold is uint16_t, which is 2 bytes. So need to combine 2 reads into one uint16_t.
      // Serial.print("current threshold of sensorboard ");Serial.print(sensorboard);Serial.print(" sensor ");Serial.print(sensor);Serial.print(" = ");Serial.println(currentThreshold);
      sensorThreshold[sensorboard][sensor] = currentThreshold; // set sensor threshold array item to received value
      delay(10);
    }
  }
}

void getSensorsMaximum() {
  uint16_t currentMaximum = 0;
  for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) {
    for (uint8_t sensor = 0; sensor < sensorArrayLength; sensor++) {    
      Wire.beginTransmission(I2CSensorboardAddresses[sensorboard]); // transmit to device with address i
      Wire.write('G'); // GT = Get Sensor maximum
      Wire.write('X'); // 
      Wire.write(sensor); // get info of this sensor number
      uint8_t returnValue = Wire.endTransmission(); // send I2C buffer 
      if (returnValue > 0) { // someting went wrong
        Serial.print("return value of I2C transmission to "); Serial.print(sensorboard);Serial.print(" GX getSensorsMaximum "); Serial.print(sensor);Serial.print(" = ");Serial.println(returnValue);
      }
      delay(10);
      Wire.requestFrom(I2CSensorboardAddresses[sensorboard], 2); // requesting 2 bytes from sensorboard
      currentMaximum = Wire.read() | (Wire.read() << 8); // currentMaximum is uint16_t, which is 2 bytes. So need to combine 2 reads into one uint16_t.
      // Serial.print("current maximum of sensorboard ");Serial.print(sensorboard);Serial.print(" sensor ");Serial.print(sensor);Serial.print(" = ");Serial.println(currentMaximum);
      sensorMaximum[sensorboard][sensor] = currentMaximum; // set sensor Maximum array item to received value
      delay(10);
    }
  }
}

void setSensorThreshold(uint8_t sensorboard, uint8_t sensor, uint16_t setValue) {
  Wire.beginTransmission(I2CSensorboardAddresses[sensorboard]); // transmit to device with address i
  Wire.write('S'); // ST = Set Sensor threshold
  Wire.write('T'); // 
  Wire.write(sensor); // get info of this sensor number
  Wire.write(setValue & 0xFF); // first byte of setValue
  Wire.write((setValue >> 8) & 0xFF); // second byte of setValue
  uint8_t returnValue = Wire.endTransmission(); // send I2C buffer 
  if (returnValue > 0) { // someting went wrong
    Serial.print("return value of I2C transmission to "); Serial.print(sensorboard);Serial.print(" sensor ");Serial.print(sensor);Serial.print(" ST setSensorThreshold "); Serial.print(sensor);Serial.print(" = ");Serial.println(returnValue);
  }
  else {
    Serial.print("Succesfully set threshold ST of sensorboard ");Serial.print(sensorboard);Serial.print(" sensor ");Serial.print(sensor);Serial.print(" to ");Serial.println(setValue);
  }
}

void setSensorMaximum(uint8_t sensorboard, uint8_t sensor, uint16_t setValue) {
  Wire.beginTransmission(I2CSensorboardAddresses[sensorboard]); // transmit to device with address i
  Wire.write('S'); // SX = Set Sensor maximum
  Wire.write('X'); // 
  Wire.write(sensor); // get info of this sensor number
  Wire.write(setValue & 0xFF); // first byte of setValue
  Wire.write((setValue >> 8) & 0xFF); // second byte of setValue
  uint8_t returnValue = Wire.endTransmission(); // send I2C buffer 
  if (returnValue > 0) { // someting went wrong
    Serial.print("return value of I2C transmission to "); Serial.print(sensorboard);Serial.print(" sensor ");Serial.print(sensor);Serial.print(" SX setSensorMaximum "); Serial.print(sensor);Serial.print(" = ");Serial.println(returnValue);
  }
  else {
    Serial.print("Succesfully set maximum SX of sensorboard ");Serial.print(sensorboard);Serial.print(" sensor ");Serial.print(sensor);Serial.print(" to ");Serial.println(setValue);
  }
}

void setLayout(uint8_t layout, bool disableSensors) {
  if (disableSensors) {
    enableSensorMeasurements(false);
  }
  for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) { // iterate over the 4 sensor boards
    Wire.beginTransmission(I2CSensorboardAddresses[sensorboard]); // transmit to device with address i
    Wire.write('S'); // SL = Set layout
    Wire.write('L'); // 
    Wire.write(layout); // set to this value
    uint8_t returnValue = Wire.endTransmission(); // send I2C buffer 
    if (returnValue > 0) { // someting went wrong
      Serial.print("return value of I2C transmission to "); Serial.print(sensorboard);Serial.print(" SL setLAyout "); Serial.print(" = ");Serial.println(returnValue);
    }
    // else {
    //   Serial.print("Succesfully set layout SL of sensorboard ");Serial.print(sensorboard);Serial.print(" to ");Serial.println(layout);
    // }
  }
  currentLayout = layout;
  assignNotes();
  if (disableSensors) {
    enableSensorMeasurements(true);
  } 
}

void setSensorColor(uint8_t sensorboard, uint8_t sensor, uint8_t valueR, uint8_t valueG, uint8_t valueB) {
  Wire.beginTransmission(I2CSensorboardAddresses[sensorboard]); // transmit to device with address i
  Wire.write('S'); // SX = Set Sensor maximum
  Wire.write('C'); // 
  Wire.write(sensor); // get info of this sensor number
  Wire.write(valueR); // get info of this sensor number
  Wire.write(valueG); // get info of this sensor number
  Wire.write(valueB); // get info of this sensor number
  uint8_t returnValue = Wire.endTransmission(); // send I2C buffer 
  if (returnValue > 0) { // someting went wrong
    Serial.print("return value of I2C transmission to "); Serial.print(sensorboard);Serial.print(" sensor ");
    Serial.print(sensor);Serial.print(" SC setcolor "); Serial.print(" = ");Serial.println(returnValue);
  }
  // else {
  //   Serial.print("Succesfully set sensorColor of sensorboard ");Serial.print(sensorboard);Serial.print(" sensor ");
  //   Serial.print(sensor);Serial.print(" to ");Serial.print(valueR);Serial.print(" ");Serial.print(valueG);Serial.print(" ");
  //   Serial.println(valueB);
  // }
}

result menuGetSensorsActualValue() {
  enableSensorMeasurements(false);
  getSensorsActualValue();
  Serial.println();
  Serial.println("*V* Actual sensor values:");
  for (int8_t sensorboard=numberOfSensorBoards-1; sensorboard >= 0; sensorboard--) {
    Serial.print(sensorboard);Serial.print("| ");    
    for (uint8_t sensor = 10; sensor < sensorArrayLength; sensor++) {
      Serial.print(sensor);Serial.print(": ");Serial.print(sensorLastActualValue[sensorboard][sensor]);Serial.print(" ");
    }
    Serial.println();
    Serial.print(sensorboard);Serial.print("| ");    
    for (uint8_t sensor = 0; sensor < sensorArrayLength - 10; sensor++) {
      Serial.print("  ");Serial.print(sensor);Serial.print(": ");Serial.print(sensorLastActualValue[sensorboard][sensor]);
    }
    Serial.println();
  }

  enableSensorMeasurements(true);
  return proceed;
}

result menuGetSensorsThreshold() {
 enableSensorMeasurements(false);
  getSensorsThreshold();
  Serial.println();
  Serial.println("*T* Sensor threshold values:");
  for (int8_t sensorboard=numberOfSensorBoards-1; sensorboard >= 0; sensorboard--) {
    Serial.print(sensorboard);Serial.print("| ");
    for (uint8_t sensor = 10; sensor < sensorArrayLength; sensor++) {
      Serial.print(sensor);Serial.print(": ");Serial.print(sensorThreshold[sensorboard][sensor]);Serial.print(" ");
    }
    Serial.println();
    Serial.print(sensorboard);Serial.print("| ");
    for (uint8_t sensor = 0; sensor < sensorArrayLength - 10; sensor++) {
      Serial.print("  ");Serial.print(sensor);Serial.print(": ");Serial.print(sensorThreshold[sensorboard][sensor]);
    }
    Serial.println();
  }
  enableSensorMeasurements(true);
  return quit;
}

result menuGetSensorsMaximum() {
 enableSensorMeasurements(false);
  getSensorsThreshold();
  Serial.println();
  Serial.println("*X* Sensor maximum values:");
  for (int8_t sensorboard=numberOfSensorBoards-1; sensorboard >= 0; sensorboard--) {
    Serial.print(sensorboard);Serial.print("| ");
    for (uint8_t sensor = 10; sensor < sensorArrayLength; sensor++) {
      Serial.print(sensor);Serial.print(": ");Serial.print(sensorMaximum[sensorboard][sensor]);Serial.print(" ");
    }
    Serial.println();
    Serial.print(sensorboard);Serial.print("| ");
    for (uint8_t sensor = 0; sensor < sensorArrayLength - 10; sensor++) {
      Serial.print("  ");Serial.print(sensor);Serial.print(": ");Serial.print(sensorMaximum[sensorboard][sensor]);
    }
    Serial.println();
  }
  enableSensorMeasurements(true);
  return quit;
}

result menuExecuteAutoThreshold() {
  enableSensorMeasurements(false);
  Serial.println("*A* Executing auto threshold on sensorboard ");
  executeAutoThreshold();
  delay(100);
  getSensorsThreshold();
  Serial.println();
  Serial.println("*T* Sensor threshold values:");
  for (int8_t sensorboard=numberOfSensorBoards-1; sensorboard >= 0; sensorboard--) {
    Serial.print(sensorboard);Serial.print("| ");
    for (uint8_t sensor = 10; sensor < sensorArrayLength; sensor++) {
      Serial.print(sensor);Serial.print(": ");Serial.print(sensorThreshold[sensorboard][sensor]);Serial.print(" ");
    }
    Serial.println();
    Serial.print(sensorboard);Serial.print("| ");
    for (uint8_t sensor = 0; sensor < sensorArrayLength - 10; sensor++) {
      Serial.print("  ");Serial.print(sensor);Serial.print(": ");Serial.print(sensorThreshold[sensorboard][sensor]);
    }
    Serial.println();
  }
  enableSensorMeasurements(true);
  return quit;
}

result menuExecuteReset() {
  enableSensorMeasurements(false);
  executeReset(); // send execute reset to sensorboards
  delay(100);
  NVIC_SystemReset(); // reset this microcontroller
  return quit;
} 

int mapToMidiVelocity(uint16_t input, uint16_t threshold, uint16_t maximum) {
    if(input < maximum) input = maximum; // make sure that input value cannot be lower than maximum value
    if(input > threshold) input = threshold; // make sure that input value cannot be lower than threshold value
    return (uint8_t)(127 - ((input - maximum) * 127) / (threshold - maximum)); // linear map threshold ~ maximum to 0 ~ 127
}

// First parameter is the event type (0x09 = note on, 0x08 = note off).
// Second parameter is note-on/note-off, combined with the channel.
// Channel can be anything between 0-15. Typically reported to the user as 1-16.
// Third parameter is the note number (48 = middle C).
// Fourth parameter is the velocity (64 = normal, 127 = fastest).

void noteOn(byte channel, byte pitch, byte velocity) {
  midiEventPacket_t noteOn = {0x09, 0x90 | channel, pitch, velocity};
  MidiUSB.sendMIDI(noteOn);
  MidiUSB.flush();
  // serializeMidiToUART(0x09, 0x90 | channel, pitch, velocity);
}

void noteOff(byte channel, byte pitch, byte velocity) {
  midiEventPacket_t noteOff = {0x08, 0x80 | channel, pitch, velocity};
  MidiUSB.sendMIDI(noteOff);
  MidiUSB.flush();
  // serializeMidiToUART(0x08, 0x80 | channel, pitch, velocity);
}

void serializeMidiToLED(byte event, byte channel, byte pitch, byte velocity) {  
  byte midiData[] = {channel, pitch, velocity};
  Serial.write(midiData, 3);
}

uint8_t scanI2Cbus() { 
  byte error, address;
  int nDevices;
  Serial.println("Scanning I2C bus ...");
  nDevices = 0;
  for(address = 1; address < 127; address++ )
  {
    Wire.beginTransmission(address);
    error = Wire.endTransmission();
    if (error == 0)
    {
      Serial.print("I2C device found at address 0x");
      if (address<16)
      { Serial.print("0"); }
      Serial.println(address,HEX);
      I2CSensorboardAddresses[nDevices] = address; // store address in array
      nDevices++;
    }
    else if (error==4)
    {
      Serial.print("Unknown error at address 0x");
      if (address<16)
      { Serial.print("0"); }
      Serial.println(address,HEX);
    }    
  }
  if (nDevices == 0) { Serial.println("No I2C devices found\n"); }
  else { Serial.println("I2C bus scanning done\n"); }
  return nDevices;
}

void checkAndSolveI2CDeadlock(){
  // Pin configuration during I2C initialisation are set as inputs.
  // PIN_WIRE_SDA (16u)-> PA16
  // PIN_WIRE_SCL (17u)-> PA17
  Serial.println("Checking I2C deadlock");
  // set PA16 to input, enabling pull resistor and setting it to pull-up.
  PORT->Group[0].DIRCLR.reg = PORT_PA16;
  PORT->Group[0].PINCFG[PIN_PA16].reg = PORT_PINCFG_PULLEN | PORT_PINCFG_INEN;
  PORT->Group[0].OUTSET.reg = PORT_PA16;
  // set PA17 to input, enabling pull resistor and setting it to pull-up.
  PORT->Group[0].DIRCLR.reg = PORT_PA17;
  PORT->Group[0].PINCFG[PIN_PA17].reg = PORT_PINCFG_PULLEN | PORT_PINCFG_INEN;
  PORT->Group[0].OUTSET.reg = PORT_PA17;

  if (~PORT->Group[0].IN.reg & PORT_PA16 ? 1 : 0) {
    Serial.println("I2C deadlock detected!");
    PORT->Group[0].DIRSET.reg = PORT_PA17; // set SCL pin as output
    //following code checks the bus state and recovers if needed
    while(~PORT->Group[0].IN.reg & PORT_PA16 ? 1 : 0) { // if the SDA line is low we clock the SCL line to free it
      PORT->Group[0].OUTCLR.reg = PORT_PA17; // pull SCL low
      delay(10);
      PORT->Group[0].OUTSET.reg = PORT_PA17; // pull SCL high
      delay(10);
    }
  }
}

void setup48MhzUSBClockRecoveryMode(){
  // from https://blog.thea.codes/understanding-the-sam-d21-clocks/
  /* Set the correct number of wait states for 48 MHz @ 3.3v */
  NVMCTRL->CTRLB.bit.RWS = 1;

  /* This works around a quirk in the hardware (errata 1.2.1) -
    the DFLLCTRL register must be manually reset to this value before
    configuration. */
  while(!SYSCTRL->PCLKSR.bit.DFLLRDY);
  SYSCTRL->DFLLCTRL.reg = SYSCTRL_DFLLCTRL_ENABLE;
  while(!SYSCTRL->PCLKSR.bit.DFLLRDY);

  /* Write the coarse and fine calibration from NVM. */
  uint32_t coarse =
      ((*(uint32_t*)FUSES_DFLL48M_COARSE_CAL_ADDR) & FUSES_DFLL48M_COARSE_CAL_Msk) >> FUSES_DFLL48M_COARSE_CAL_Pos;
  uint32_t fine =
      ((*(uint32_t*)FUSES_DFLL48M_FINE_CAL_ADDR) & FUSES_DFLL48M_FINE_CAL_Msk) >> FUSES_DFLL48M_FINE_CAL_Pos;

  SYSCTRL->DFLLVAL.reg = SYSCTRL_DFLLVAL_COARSE(coarse) | SYSCTRL_DFLLVAL_FINE(fine);

  /* Wait for the write to finish. */
  while (!SYSCTRL->PCLKSR.bit.DFLLRDY) {};

  SYSCTRL->DFLLCTRL.reg |=
      /* Enable USB clock recovery mode */
      SYSCTRL_DFLLCTRL_USBCRM |
      /* Disable chill cycle as per datasheet to speed up locking.
        This is specified in section 17.6.7.2.2, and chill cycles
        are described in section 17.6.7.2.1. */
      SYSCTRL_DFLLCTRL_CCDIS;

  /* Configure the DFLL to multiply the 1 kHz clock to 48 MHz */
  SYSCTRL->DFLLMUL.reg =
      /* This value is output frequency / reference clock frequency,
        so 48 MHz / 1 kHz */
      SYSCTRL_DFLLMUL_MUL(48000) |
      /* The coarse and fine values can be set to their minimum
        since coarse is fixed in USB clock recovery mode and
        fine should lock on quickly. */
      SYSCTRL_DFLLMUL_FSTEP(1) |
      SYSCTRL_DFLLMUL_CSTEP(1);

  /* Closed loop mode */
  SYSCTRL->DFLLCTRL.bit.MODE = 1;

  /* Enable the DFLL */
  SYSCTRL->DFLLCTRL.bit.ENABLE = 1;

  /* Wait for the write to complete */
  while (!SYSCTRL->PCLKSR.bit.DFLLRDY) {};      

  /* Setup GCLK0 using the DFLL @ 48 MHz */
  GCLK->GENCTRL.reg =
      GCLK_GENCTRL_ID(0) |
      GCLK_GENCTRL_SRC_DFLL48M |
      /* Improve the duty cycle. */
      GCLK_GENCTRL_IDC |
      GCLK_GENCTRL_GENEN;

  /* Wait for the write to complete */
  while(GCLK->STATUS.bit.SYNCBUSY) {};

  // for debugging: temporarily set 48MHz clock to GCLK4 as well and output to a pin to check accuracy.
  // GCLK->GENCTRL.reg =
  //     GCLK_GENCTRL_ID(4) |
  //     GCLK_GENCTRL_SRC_DFLL48M |
  //     GCLK_GENCTRL_IDC |
  //     GCLK_GENCTRL_GENEN |
  //     /* Enable outputting the clock to an I/O pin. */
  //     GCLK_GENCTRL_OE;

  // /* Wait for the write to complete */
  // while(GCLK->STATUS.bit.SYNCBUSY);
  // PORT->Group[1].DIRSET.reg = (1 << 10);
  // PORT->Group[1].PINCFG[10].reg |= PORT_PINCFG_PMUXEN;
  // PORT->Group[1].PMUX[10 >> 1].bit.PMUXE |= PORT_PMUX_PMUXE_H;
}

void setupHardwareTimers() 
{
    // Set up the generic clock (GCLK4) used to clock timers
    GCLK->GENDIV.reg = GCLK_GENDIV_DIV(1) |          // Divide the 48MHz clock source by divisor 1: 48MHz/1=48MHz
                       GCLK_GENDIV_ID(4);            // Select Generic Clock (GCLK) 4
    while (GCLK->STATUS.bit.SYNCBUSY);               // Wait for synchronization

    GCLK->GENCTRL.reg = GCLK_GENCTRL_IDC |           // Set the duty cycle to 50/50 HIGH/LOW
                        GCLK_GENCTRL_GENEN |         // Enable GCLK4
                        GCLK_GENCTRL_SRC_DFLL48M |   // Set the 48MHz clock source
                        GCLK_GENCTRL_ID(4);          // Select GCLK4
    while (GCLK->STATUS.bit.SYNCBUSY);               // Wait for synchronization

    // // Feed GCLK4 to TC3
    // GCLK->CLKCTRL.reg = GCLK_CLKCTRL_CLKEN |         // Enable GCLK4
    //                     GCLK_CLKCTRL_GEN_GCLK4 |     // Select GCLK4
    //                     GCLK_CLKCTRL_ID_TCC2_TC3;     // Feed the GCLK4 to TC3 (and TCC2)
    // while (GCLK->STATUS.bit.SYNCBUSY);               // Wait for synchronization

    // Feed GCLK4 to TC4 and TC5
    GCLK->CLKCTRL.reg = GCLK_CLKCTRL_CLKEN |         // Enable GCLK4
                        GCLK_CLKCTRL_GEN_GCLK4 |     // Select GCLK4
                        GCLK_CLKCTRL_ID_TC4_TC5;     // Feed the GCLK4 to TC4 and TC5
    while (GCLK->STATUS.bit.SYNCBUSY);               // Wait for synchronization

    // // specific for TC3: 1ms
    // TC3->COUNT16.CTRLA.reg |= TC_CTRLA_MODE_COUNT16;  // Set Timer counter Mode to 16 bits
    // while (TC3->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // TC3->COUNT16.CTRLA.reg |= TC_CTRLA_WAVEGEN_MFRQ;  // Set TC as Match Frequency
    // while (TC3->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // TC3->COUNT16.CC[0].reg = COUNTS_PER_TICK_1MS;     // Set the TC3 CC0 register as the TOP value in match frequency mode
    // while (TC3->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // // Enable the TC3 interrupt request
    // TC3->COUNT16.INTENSET.bit.MC0 = 1;
    // while (TC3->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // // Enable TC3
    // TC3->COUNT16.CTRLA.reg |= TC_CTRLA_ENABLE;        // Enable TC3
    // while (TC3->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // // Enable the TC3 interrupt vector use priority 2
    // NVIC_SetPriority(TC3_IRQn, 2);                    // Set the NVIC priority for TC3 to 2
    // NVIC_EnableIRQ(TC3_IRQn);

    // specific for TC4: 100us
    TC4->COUNT16.CTRLA.reg |= TC_CTRLA_MODE_COUNT16;  // Set Timer counter Mode to 16 bits. Max value for COUNT16 is 65535, resulting in about 1,36ms.
    while (TC4->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    TC4->COUNT16.CTRLA.reg |= TC_CTRLA_WAVEGEN_MFRQ;  // Set TC as Match Frequency
    while (TC4->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    TC4->COUNT16.CC[0].reg = COUNTS_PER_TICK_100US;     // Set the TC4 CC0 register as the TOP value in match frequency mode
    while (TC4->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable the TC4 interrupt request
    TC4->COUNT16.INTENSET.bit.MC0 = 1;
    while (TC4->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable TC4
    TC4->COUNT16.CTRLA.reg |= TC_CTRLA_ENABLE;        // Enable TC4
    while (TC4->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable the TC4 interrupt vector use priority 0
    NVIC_SetPriority(TC4_IRQn, 0);                    // Set the NVIC priority for TC4 to 0
    NVIC_EnableIRQ(TC4_IRQn);

    // specific for TC5
    TC5->COUNT16.CTRLA.reg |= TC_CTRLA_MODE_COUNT16;  // Set Timer counter Mode to 16 bits
    while (TC5->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // TC5->COUNT16.CTRLA.reg |= TC_CTRLA_PRESCALER_DIV8;  // Set Timer prescaler to 4, so 48MHz/4 = 12MHz
    TC5->COUNT16.CTRLA.reg |= TC_CTRLA_PRESCALER_DIV4;  // Set Timer prescaler to 4, so 48MHz/4 = 12MHz
                                                      // with 16 bit timer the maximum period is 65535/12MHz = 5.46ms
    while (TC4->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    TC5->COUNT16.CTRLA.reg |= TC_CTRLA_WAVEGEN_MFRQ;  // Set TC as Match Frequency
    while (TC5->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    TC5->COUNT16.CC[0].reg = COUNTS_PER_TICK_TC5;     // Set the TC5 CC0 register as the TOP value in match frequency mode
    while (TC5->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable the TC5 interrupt request
    TC5->COUNT16.INTENSET.bit.MC0 = 1;
    while (TC5->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable TC5
    TC5->COUNT16.CTRLA.reg |= TC_CTRLA_ENABLE;        // Enable TC5
    while (TC5->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // use priority 1
    NVIC_SetPriority(TC5_IRQn, 1);                    // Set the NVIC priority for TC5 to 1
    NVIC_EnableIRQ(TC5_IRQn);    // Enable the TC5 timer (start checking sensorboards for hits)

}
