/*
** Firmware for Marimbatron Sensormodule
** Version 7
** 
** this firmware will:
** - read out 19 analog sensor inputs
** - blink the red and green leds
** - send data using I2C (Wire1) when requested -> 
**       using Wire1 because I2C pins are not the default ones (PA16 and PA17), 
**       but PA12 and PA13. Wire1 is using those.
** - use ISR timers to get micros
** - Use DMA to send data to 19 WS2812Bs using PB10. 
**       Credits to https://www.lucadavidian.com/wifi-controlled-neo-pixels-strips/
** - implement I2C communication protocol to communicate from main board to sensor board
** - store configurable values in non-volatile flash
**
** TODO - unroll ISR loops to speed things up (only if needed)
**
** Required libraries:
** wire.h
**
** for I2C Marimbatron protocol see comments in mainmodule code.
**
** This work may be reproduced, modified, distributed, performed, and 
** displayed for any non-commercial, non-profit purpose, but must acknowledge 
** the developer and the Marimbatron project. 
** Copyright is retained and must be preserved. The work is 
** provided as is; no warranty is provided, and users accept all liability.
** 
** Copyright 2024 Leo Kuipers.
*/

#define FIRMWAREVERSION 7
// #define DEBUG

// timers
#include "Clocks.h"
volatile uint32_t micros_counter = 0; // holds the count of microseconds since the MCU started (don't want to use micros())

const unsigned long intervalEvent1 = 50000; // in microseconds
const unsigned long intervalEvent2 = 40000; // 40ms = 25x per second (in microseconds)
unsigned long previousTimeEvent1 = 0;
unsigned long previousTimeEvent2 = 0;

const uint8_t sensorArrayLength = 19;

// non-volatile memory
#include <FlashStorage_SAMD.h>
typedef struct { // A struct that will be stored in flash
  uint16_t thresholdStored[sensorArrayLength]; // stored threshold values for all sensors
  uint16_t maximumStored[sensorArrayLength]; // stored threshold values for all sensors
  uint16_t scantimeStored[sensorArrayLength]; // stored threshold values for all sensors
  uint16_t masktimeStored[sensorArrayLength]; // stored threshold values for all sensors
  bool dataValid;       // A flag to check if data is valid (if the data is already stored or not)
} StorageData;

// Create an instance of FlashStorage
FlashStorage(storageData, StorageData);

// LEDs and WS2812B
#define GREENLEDPIN PORT_PB30 // PB30 = arduino 43
#define REDLEDPIN PORT_PB31 // PB31 = arduino 46
static bool flashLED;
const uint8_t ledOnRmax = 40; // was 127
const uint8_t ledOnGmax = 40; // was 127
const uint8_t ledOnBmax = 100; // was 255
bool sensorHit[sensorArrayLength];
#include "WS2812B.h" // 19 leds are defined in this file
typedef struct {
    uint8_t R;
    uint8_t G;
    uint8_t B;
} RGB;
const RGB colorLayoutWickiHayden[sensorArrayLength] = 
{
    {ledOnRmax, 0, 0}, // sensor 0
    {ledOnRmax, 0, 0}, // sensor 1
    {ledOnRmax, 0, 0}, // sensor 2
    {0, 0, ledOnBmax}, // sensor 3
    {0, 0, ledOnBmax}, // sensor 4
    {0, 0, ledOnBmax}, // sensor 5
    {0, 0, ledOnBmax}, // sensor 6
    {ledOnRmax, 0, 0}, // sensor 7
    {ledOnRmax, 0, 0}, // sensor 8
    {ledOnRmax, 0, 0}, // sensor 9

    {ledOnRmax, 0, 0}, // sensor 10
    {ledOnRmax, 0, 0}, // sensor 11
    {ledOnRmax, 0, 0}, // sensor 12
    {0, 0, ledOnBmax}, // sensor 13
    {0, 0, ledOnBmax}, // sensor 14
    {0, 0, ledOnBmax}, // sensor 15
    {ledOnRmax, 0, 0}, // sensor 16
    {ledOnRmax, 0, 0}, // sensor 17
    {ledOnRmax, 0, 0}  // sensor 18 
};
const RGB colorLayoutPiano[sensorArrayLength] = 
{
    {0, ledOnGmax, 0}, // sensor 0
    {0, ledOnGmax, 0}, // sensor 1
    {0, ledOnGmax, 0}, // sensor 2
    {0, ledOnGmax, 0}, // sensor 3
    {0, ledOnGmax, 0}, // sensor 4
    {0, ledOnGmax, 0}, // sensor 5
    {0, ledOnGmax, 0}, // sensor 6
    {0, ledOnGmax, 0}, // sensor 7
    {0, ledOnGmax, 0}, // sensor 8
    {0, ledOnGmax, 0}, // sensor 9

    {ledOnRmax, 0, 0}, // sensor 10
    {ledOnRmax, 0, 0}, // sensor 11
    {0, 0, 0}, // sensor 12
    {ledOnRmax, 0, 0}, // sensor 13
    {ledOnRmax, 0, 0}, // sensor 14
    {ledOnRmax, 0, 0}, // sensor 15
    {0, 0, 0}, // sensor 16
    {ledOnRmax, 0, 0}, // sensor 17
    {ledOnRmax, 0, 0} // sensor 18  
};
const RGB colorLayoutAllOff[sensorArrayLength] = 
{
    {0, 0, 0}, // sensor 0
    {0, 0, 0}, // sensor 1
    {0, 0, 0}, // sensor 2
    {0, 0, 0}, // sensor 3
    {0, 0, 0}, // sensor 4
    {0, 0, 0}, // sensor 5
    {0, 0, 0}, // sensor 6
    {0, 0, 0}, // sensor 7
    {0, 0, 0}, // sensor 8
    {0, 0, 0}, // sensor 9
    {0, 0, 0}, // sensor 10
    {0, 0, 0}, // sensor 11
    {0, 0, 0}, // sensor 12
    {0, 0, 0}, // sensor 13
    {0, 0, 0}, // sensor 14
    {0, 0, 0}, // sensor 15
    {0, 0, 0}, // sensor 16
    {0, 0, 0}, // sensor 17
    {0, 0, 0} // sensor 18  
};
RGB targetSensorColor[sensorArrayLength]; // array that holds the color that the sensors should be
RGB currentSensorColor[sensorArrayLength]; // array that holds the color that the sensors are currently
const float colorTransitionSpeed = 0.3;
const uint8_t WS2812BtoSensorMap[sensorArrayLength] = 
{
 0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 1, 3, 5, 7, 9, 11, 13, 15, 17
};
uint8_t currentLayout = 3; // 0 = wicki hayden, 1 = piano, 2 = all off
// ADC
#include "ADC.c"

// array of the sensors, to be used for iterating
const unsigned long sensorArray[sensorArrayLength] = 
{
  SENSOR0, SENSOR1, SENSOR2, SENSOR3, SENSOR4,
  SENSOR5, SENSOR6, SENSOR7, SENSOR8, SENSOR9,
  SENSOR10, SENSOR11, SENSOR12, SENSOR13, SENSOR14,
  SENSOR15, SENSOR16, SENSOR17, SENSOR18
};

#include "Sensor.h" // include sensor class
Sensor pads[sensorArrayLength]; // array that holds all sensor related data
const uint16_t DEFAULTTHRESHOLD = 2600; // 12-bit threshold, max value 4095.
const uint16_t DEFAULTMAXIMUM = 200; // 12-bit maximum, max value 4095.
const uint16_t TOOLOWSENSORVALUE = 800; // if sensor value is below this value at startup, it's considered not connected
const uint16_t TOOLOWTHRESHOLD = 0; // if sensor not connected at startup, set threshold to this value
const uint16_t DEFAULTSCANTIME = 2000; // scantime in us.
const uint16_t DEFAULTMASKTIME = 50000; // masktime in us.
uint8_t selectedSensor = sensorArrayLength; // variable to hold selected sensornumber when sending data to main controller. If set to sensorArrayLength: no sensor is selected.
bool enableMeasurement = false; // variable to enable or disable sensor measurements (in TC5_handler)

// I2C communication
#include <Wire.h> 
static char mode = 'H'; // default program mode: H = sensor Hit mode

// configurable variables, to be stored in flash
uint16_t autoThresholdValue = 200;


void TC4_Handler() 
{ // 100us timer, priority 0
  // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope
  micros_counter = micros_counter + 100;
  // Clear the interrupt
  TC4->COUNT16.INTFLAG.bit.MC0 = 1;
}

void TC5_Handler() { // 200 us timer, priority 1
  // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope
  if (enableMeasurement) {
    measureSensors();
  }
  // Clear the interrupt
  TC5->COUNT16.INTFLAG.bit.MC0 = 1;
}

void measureSensors(void) 
{
  uint16_t padThreshold;
  uint32_t lastADCTime;
  uint32_t maskTime;
  unsigned long startScan;
  uint16_t currentMeasurement;
  bool maskTimeCheck;
  // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope

  for (uint8_t i = 0; i < sensorArrayLength; i++) { // iterate through all sensors 
    // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope
    startScan = micros_counter; // save current time
    lastADCTime = pads[i].getLastADCTime(); // save last 
    maskTime = pads[i].getMaskTime();
    // Next line is shorthand IF statement: checks if sensor should be masked
    // true if the measurement should not be masked. i.e. the measurement should be done.
    // startscan - lastADCTime >= masktime is done in this order to prevent issues from micros_counter overflow.
    // note (unsigned long) is typecasting the calculation between () to make sure the overflow prevention works.
    // see https://www.norwegiancreations.com/2018/10/arduino-tutorial-avoiding-the-overflow-issue-when-using-millis-and-micros/
    ((unsigned long)(startScan - lastADCTime) >= maskTime) ? maskTimeCheck = true : maskTimeCheck = false;

    if (maskTimeCheck) {
      // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope
      padThreshold = pads[i].getThreshold();
      setADCToPin(sensorArray[i]);
      currentMeasurement = readADC(true); // true = first measurement should be disregarded due to channel change
      if (currentMeasurement < padThreshold) { // measurement below threshold = hit detected. Start peak detection for scantime us.
        uint32_t scanTime = pads[i].getScanTime();
        uint16_t peakMeasurement = padThreshold;
        while ((unsigned long)(micros_counter - startScan) <= scanTime) { // stay measuring for as long as scanTime is. Again using micros_counter overflow prevention
          // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope
          currentMeasurement = readADC(false); // disregarding first measurement is not neccesary because we didn't change the input channel
          if (currentMeasurement < peakMeasurement) { // peak detection for local minimum
            peakMeasurement = currentMeasurement;
          }
        }
        // when scantime is done, store the peak value as velocity and set hit to true to indicate a valid hit is detected.
        pads[i].setVelocity(peakMeasurement); // peak detection done. Store peak value.
        pads[i].setHit(true); // set hit to true to indicate the pad was hit.
        sensorHit[i] = true; // save sensor hit status
        pads[i].setLastADCTime(startScan + scanTime); // store lastADCTime, required for masktime calculation.
        #ifdef DEBUG
          Serial.print("hit on s"); Serial.print(i); Serial.print(" ");Serial.print(peakMeasurement);Serial.println();
        #endif // DEBUG  
      } 
      else {
        if (!pads[i].getHit()) { // do not do this if there's a hit detected. Wait for hit to be set to false.
          pads[i].setVelocity(currentMeasurement); // save measurement (which is > threshold) as velocity, as to be able to show the value
        }
      }
    }
  }
}

void I2CCommand(int length) {
  uint8_t sensorNumber;
  uint16_t value, valueR, valueG, valueB;
  uint32_t time;
  #ifdef DEBUG
  // Serial.print("Receiving command with length ");
  // Serial.println(length);
  #endif // DEBUG  

  if (length > 0) { // if there's no length, there are no command bytes. So it's a request for data which is handled by I2CEvent()
    enableMeasurement = false; // disable sensor measurements so they don't interfere
    #ifdef DEBUG
      Serial.print("I2CCommand() - Current mode is "); Serial.println(mode);
    #endif // DEBUG  

    char c1 = Wire1.read(); // read first byte
    char c2 = Wire1.read(); // read second byte
    if (c1 == -1 or c2 == -1) { // read error, no data available
      #ifdef DEBUG
        Serial.println("I2CCommand() - Error reading I2C data");
      #endif // DEBUG  
    }
    else {
      #ifdef DEBUG
        Serial.print("I2CCommand() - Received command ");Serial.print(c1);Serial.println(c2);
      #endif // DEBUG  
      if (c1 == 'X') { // execute command
        if (c2 == 'R') { // XR: reset
          #ifdef DEBUG
            Serial.println("I2CCommand() - Execute reset received.");
          #endif // DEBUG  
          NVIC_SystemReset(); // reset this microcontroller
        }
        else if (c2 == 'A') { // XA: auto threshold
          #ifdef DEBUG
            Serial.println("I2CCommand() - Execute autoThreshold recieved.");
          #endif // DEBUG  
          autoThreshold();
        }
        else if (c2 == 'S') { // XS: stop measurements
          #ifdef DEBUG
            Serial.println("I2CCommand() - Execute stop measurements recieved.");
          #endif // DEBUG  
          enableMeasurement = false; // stop measurements
          for (uint8_t i = 0; i < sensorArrayLength; i++) { // flush sensorHit array
            sensorHit[i] = false;
          }
        }
      }
      else if (c1 == 'G') { // get command
        if (c2 == 'H') { // GH: get sensor hit
          mode = 'H';
          for (uint8_t i = 0; i < sensorArrayLength; i++) { // flush sensorHit array
            sensorHit[i] = false;
          }
          enableMeasurement = true; // enable sensor measurements
        }
        else if (c2 == 'C') { // GC: get continous actual value 
          mode = 'C';
        }
        else if (c2 == 'A') { // GA: get single actual value
          mode = 'A';
          sensorNumber = Wire1.read(); // read third byte
          if (sensorNumber >= sensorArrayLength) {
            #ifdef DEBUG
              Serial.println("I2CCommand() - Error: received get single actual value with sensornumber out of bound");
            #endif // DEBUG  
            selectedSensor = sensorArrayLength; // means no sensor is selected
          }
          else {
            selectedSensor = sensorNumber;
          }
        }
        else if (c2 == 'T') { // GT: get threshold
          mode = 'T';
          sensorNumber = Wire1.read(); // read third byte
          if (sensorNumber >= sensorArrayLength) {
            #ifdef DEBUG
              Serial.println("I2CCommand() - Error: received get threshold with sensornumber out of bound");
            #endif // DEBUG  
            selectedSensor = sensorArrayLength; // means no sensor is selected
          }
          else {
            selectedSensor = sensorNumber;
          }
        }
        else if (c2 == 'X') { // GX: get maximum
          mode = 'X';
          sensorNumber = Wire1.read(); // read third byte
          if (sensorNumber >= sensorArrayLength) {
            #ifdef DEBUG
              Serial.println("I2CCommand() - Error: received get maximum with sensornumber out of bound");
            #endif // DEBUG  
            selectedSensor = sensorArrayLength; // means no sensor is selected
          }
          else {
            selectedSensor = sensorNumber;
          }
        }
        else if (c2 == 'M') { // GM: get masktime
          mode = 'M';
          sensorNumber = Wire1.read(); // read third byte
          if (sensorNumber >= sensorArrayLength) {
            #ifdef DEBUG
              Serial.println("I2CCommand() - Error: received get masktime with sensornumber out of bound");
            #endif // DEBUG  
            selectedSensor = sensorArrayLength; // means no sensor is selected
          }
          else {
            selectedSensor = sensorNumber;
          }
        }
        else if (c2 == 'N') { // GN: get scantime
          mode = 'N';
          sensorNumber = Wire1.read(); // read third byte
          if (sensorNumber >= sensorArrayLength) {
            #ifdef DEBUG
              Serial.println("I2CCommand() - Error: received get scantime with sensornumber out of bound");
            #endif // DEBUG  
            selectedSensor = sensorArrayLength; // means no sensor is selected
          }
          else {
            selectedSensor = sensorNumber;
          }
        }
        #ifdef DEBUG
          Serial.print("I2CCommand() - Mode is now ");Serial.println(mode);
        #endif // DEBUG  
      }
      else if (c1 == 'S') { // set
        if (c2 == 'T') { // ST: set threshold
          sensorNumber = Wire1.read(); // read third byte
          value = Wire1.read() | (Wire1.read() << 8); // combine 2 bytes into one uint16
          if (sensorNumber >= sensorArrayLength) {
          #ifdef DEBUG
            Serial.println("Error: received set threshold with sensornumber out of bound");
          #endif // DEBUG  
          }
          else {
            pads[sensorNumber].setThreshold(value);
            saveToFlash(); // save value to flash
          #ifdef DEBUG
            Serial.print("Set threshold of sensor ");Serial.print(sensorNumber);Serial.print(" to ");Serial.println(value);
          #endif // DEBUG  
          }
        }
        else if (c2 == 'X') { // SX: set maximum
          sensorNumber = Wire1.read(); // read third byte
          value = Wire1.read() | (Wire1.read() << 8); // combine 2 bytes into one uint16
          if (sensorNumber >= sensorArrayLength) {
          #ifdef DEBUG
            Serial.println("Error: received set threshold with sensornumber out of bound");
          #endif // DEBUG  
          }
          else {
            pads[sensorNumber].setMaximum(value);
            saveToFlash(); // save value to flash
            #ifdef DEBUG
              Serial.print("Set maximum of sensor ");Serial.print(sensorNumber);Serial.print(" to ");Serial.println(value);
            #endif // DEBUG  
          }
        }
        else if (c2 == 'M') { // SM: set masktime
          sensorNumber = Wire1.read(); // read third byte
          time = Wire1.read() | (Wire1.read() << 8) | (Wire1.read() << 16) | (Wire1.read() << 24); // combine 4 bytes into one uint32
          if (sensorNumber >= sensorArrayLength) {
            #ifdef DEBUG
              Serial.println("Error: received set masktime with sensornumber out of bound");
            #endif // DEBUG  
          }
          else {
            pads[sensorNumber].setMaskTime(time);
            saveToFlash(); // save value to flash
            #ifdef DEBUG
              Serial.print("Set masktime of sensor ");Serial.print(sensorNumber);Serial.print(" to ");Serial.println(time);
            #endif // DEBUG  
          }
        }
        else if (c2 == 'N') { // SN: set scantime
          sensorNumber = Wire1.read(); // read third byte
          time = Wire1.read() | (Wire1.read() << 8) | (Wire1.read() << 16) | (Wire1.read() << 24); // combine 4 bytes into one uint32
          if (sensorNumber >= sensorArrayLength) {
            #ifdef DEBUG
              Serial.println("Error: received set scantime with sensornumber out of bound");
            #endif // DEBUG  
          }
          else {
            pads[sensorNumber].setScanTime(time);
            saveToFlash(); // save value to flash
            #ifdef DEBUG
              Serial.print("Set scantime of sensor ");Serial.print(sensorNumber);Serial.print(" to ");Serial.println(time);
            #endif // DEBUG  
          }
        }
        else if (c2 == 'L') { // SL: set layout
          value = Wire1.read(); // read third byte
          if (value > 2) {
            #ifdef DEBUG
              Serial.println("Error: received set layout with value out of bound");
            #endif // DEBUG  
          }
          else {
            setLayoutColor(value);
            #ifdef DEBUG
              Serial.print("Set layout ");Serial.print(" to ");Serial.println(value);
            #endif // DEBUG  
          }
        }
        else if (c2 == 'C') { // SC: set color
          sensorNumber = Wire1.read(); // read third byte
          valueR = Wire1.read();
          valueG = Wire1.read();
          valueB = Wire1.read();
          if (sensorNumber >= sensorArrayLength) {
            #ifdef DEBUG
              Serial.println("Error: received set color with sensornumber out of bound");
            #endif // DEBUG  
          }
          else {
            setSensorColor(sensorNumber, valueR, valueG, valueB);
            #ifdef DEBUG
              Serial.print("Set color of sensor ");Serial.print(sensorNumber);Serial.print(" to ");
              Serial.print(valueR);Serial.print(" ");Serial.print(valueG);Serial.print(" ");Serial.println(valueB);
            #endif // DEBUG  
          }
        }
      }
    }
  }
}

void I2CEvent() { // modes: S C A T X M N
  uint16_t velocity;
  bool hitDetected = false;
  #ifdef DEBUG
    if (mode != 'H') {
      Serial.println("** Received I2C event **");
    }
  #endif // DEBUG  
  if (mode == 'H') { // GH Get sensor hit
    uint8_t data[] = {255,0,0};
    // send [255, 0, 0] when no hit is detected. Otherwise send sensor number and velocity of just 1 sensor and reset the hit state of that sensor
    for (uint8_t i = 0; i < sensorArrayLength; i++) {
      if (pads[i].getHit()){ // if true: hit is detected
        hitDetected = true;
        velocity = pads[i].getVelocity();
        #ifdef DEBUG
          Serial.print("send hit to I2C s"); Serial.print(i); Serial.print(" ");Serial.print(velocity);Serial.println();
        #endif // DEBUG  
        // set sensor number and velocity to data variable when hit is detected
        data[0] = i;
        data[1] = velocity & 0xFF;
        data[2] = (velocity >> 8) & 0xFF;
        pads[i].setHit(false); // reset hit
        break; // break the for loop
      }
    }
    #ifdef DEBUG
    // Serial.print("sending I2C data: ");
    // Serial.print(data[0]);Serial.print(" ");Serial.print(data[1]);Serial.print(" ");Serial.println(data[2]);
    #endif // DEBUG  
    Wire1.write(data, 3); // even when no hit is detected an answer will still be send using sensor number 255 (255, 0, 0)
  }
  else if (mode == 'C') { // continuous actual value
    #ifdef DEBUG
      Serial.println("I2CEvent() - mode C");
    #endif // DEBUG  
  }
  else if (mode == 'A') { // single actual value
    #ifdef DEBUG
      Serial.println("I2CEvent() - mode A");
    #endif // DEBUG  
    uint8_t data[] = {0,0};
    if (selectedSensor < sensorArrayLength) { // when set to sensorArrayLength no sensor is selected
      setADCToPin(sensorArray[selectedSensor]);
      uint16_t currentMeasurement = readADC(true); // true = first measurement should be disregarded due to channel change
      data[0] = currentMeasurement & 0xFF;
      data[1] = (currentMeasurement >> 8) & 0xFF;
      Wire1.write(data, 2); // send actual value of selected sensor
    }
  }
  else if (mode == 'T') { // GT threshold
    #ifdef DEBUG
      Serial.println("I2CEvent() - mode T");
    #endif // DEBUG  
    #ifdef DEBUG
      Serial.print("I2CEvent GT, sending threshold for sensor ");Serial.print(selectedSensor);Serial.print(" which is ");Serial.println(pads[selectedSensor].getThreshold());
    #endif // DEBUG  
    uint8_t data[] = {0,0};
    if (selectedSensor < sensorArrayLength) { // when set to sensorArrayLength no sensor is selected
      uint16_t currentValue = pads[selectedSensor].getThreshold();
      data[0] = currentValue & 0xFF;
      data[1] = (currentValue >> 8) & 0xFF;
      Wire1.write(data, 2); // send actual value of selected sensor
    }
  }
  else if (mode == 'X') { // GX maximum
    #ifdef DEBUG
      Serial.println("I2CEvent() - mode X");
    #endif // DEBUG  
    #ifdef DEBUG
      Serial.print("I2CEvent GX, sending maximum for sensor ");Serial.print(selectedSensor);Serial.print(" which is ");Serial.println(pads[selectedSensor].getMaximum());
    #endif // DEBUG  
    uint8_t data[] = {0,0};
    if (selectedSensor < sensorArrayLength) { // when set to sensorArrayLength no sensor is selected
      uint16_t currentValue = pads[selectedSensor].getMaximum();
      data[0] = currentValue & 0xFF;
      data[1] = (currentValue >> 8) & 0xFF;
      Wire1.write(data, 2); // send actual value of selected sensor
    }
  }
  else if (mode == 'M') { // GM masktime
    #ifdef DEBUG
      Serial.println("I2CEvent() - mode M");
    #endif // DEBUG  
    #ifdef DEBUG
      Serial.print("I2CEvent GM, sending masktime for sensor ");Serial.print(selectedSensor);Serial.print(" which is ");Serial.println(pads[selectedSensor].getMaskTime());
    #endif // DEBUG  
    uint8_t data[] = {0,0,0,0};
    if (selectedSensor < sensorArrayLength) { // when set to sensorArrayLength no sensor is selected
      uint32_t currentValue = pads[selectedSensor].getMaskTime();
      data[0] = currentValue & 0xFF;
      data[1] = (currentValue >> 8) & 0xFF;
      data[2] = (currentValue >> 16) & 0xFF;
      data[3] = (currentValue >> 24) & 0xFF;
      Wire1.write(data, 4); // send actual value of selected sensor
    }
  }
  else if (mode == 'N') { // GN scantime
    #ifdef DEBUG
      Serial.println("I2CEvent() - mode N");
    #endif // DEBUG  
    #ifdef DEBUG
      Serial.print("I2CEvent GN, sending scantime for sensor ");Serial.print(selectedSensor);Serial.print(" which is ");Serial.println(pads[selectedSensor].getScanTime());
    #endif // DEBUG  
    uint8_t data[] = {0,0,0,0};
    if (selectedSensor < sensorArrayLength) { // when set to sensorArrayLength no sensor is selected
      uint32_t currentValue = pads[selectedSensor].getScanTime();
      data[0] = currentValue & 0xFF;
      data[1] = (currentValue >> 8) & 0xFF;
      data[2] = (currentValue >> 16) & 0xFF;
      data[3] = (currentValue >> 24) & 0xFF;
      Wire1.write(data, 4); // send actual value of selected sensor
    }
  }
}

void setup() 
{
  uint8_t I2CAddress;
  bool I2CAddress0, I2CAddress1, I2CAddress2;

  // set SAMD21 to run at 48Mhz
  setup48MhzUSBClockRecoveryMode();
  // setup 3 hardware timers
  setupHardwareTimers();

  // configure serial port
  #ifdef DEBUG
    Serial.begin(115200); // enable serial communication  
    while (!Serial && micros_counter < 5000000); // wait for serial bus to activate (in order to communicate immediately after reset) or timeout
    delay(100);
    Serial.print("Marimbatron Sensor board firmware v");
    Serial.println(FIRMWAREVERSION);
    Serial.println(BOARD_NAME);
    Serial.println(FLASH_STORAGE_SAMD_VERSION);
  #endif // DEBUG  

  // configure LED pins
  PORT->Group[1].DIRSET.reg = GREENLEDPIN | REDLEDPIN; // set led pins as output
  PORT->Group[1].OUTCLR.reg = REDLEDPIN; // turn red led pin off
  PORT->Group[1].OUTSET.reg = GREENLEDPIN; // turn green led pin on

  // get stored data from flash
  StorageData settings;
  storageData.read(settings);
    // Check if the data is valid
  if (!settings.dataValid) {     // Data is not valid, initialize the array
    #ifdef DEBUG
      Serial.println("Data is not valid, initializing the array");
    #endif // DEBUG
    for (int i = 0; i < sensorArrayLength; i++) {
      settings.thresholdStored[i] = DEFAULTTHRESHOLD; // set default threshold
      settings.maximumStored[i] = DEFAULTMAXIMUM; // set default maximum
      settings.scantimeStored[i] = DEFAULTSCANTIME; // set default scantime
      settings.masktimeStored[i] = DEFAULTMASKTIME; // set default masktime      
    }
    settings.dataValid = true;
    storageData.write(settings);  // Write the initialized data to flash
  }

  #ifdef DEBUG
    Serial.println("Init data from flash");
  #endif // DEBUG
  for (int i = 0; i < sensorArrayLength; i++) {
    #ifdef DEBUG
      Serial.print("threshold ");Serial.print(i);Serial.print(" = ");Serial.println(settings.thresholdStored[i]);
    #endif // DEBUG
    pads[i].setThreshold(settings.thresholdStored[i]);
    pads[i].setMaximum(settings.maximumStored[i]);
    pads[i].setScanTime(settings.scantimeStored[i]);
    pads[i].setMaskTime(settings.masktimeStored[i]);
    pads[i].setLastADCTime(micros_counter);
  }

  // configure WS2812B
  WS2812B_init();
  setLayoutColor(currentLayout); // set layout colors to All Off

  // configure ADCs
  setupADC();

  // configure I2C
  // set PA14 to input, enabling pull resistor and setting it to pull-up.
  PORT->Group[0].DIRCLR.reg = PORT_PA14;
  PORT->Group[0].PINCFG[14].reg = PORT_PINCFG_PULLEN | PORT_PINCFG_INEN;
  PORT->Group[0].OUTSET.reg = PORT_PA14;
  // set PB15 to input, enabling pull resistor and setting it to pull-up.
  PORT->Group[1].DIRCLR.reg = PORT_PB15;
  PORT->Group[1].PINCFG[15].reg = PORT_PINCFG_PULLEN | PORT_PINCFG_INEN;
  PORT->Group[1].OUTSET.reg = PORT_PB15;
  // set PA15 to input, enabling pull resistor and setting it to pull-up.
  PORT->Group[0].DIRCLR.reg = PORT_PA15;
  PORT->Group[0].PINCFG[15].reg = PORT_PINCFG_PULLEN | PORT_PINCFG_INEN;
  PORT->Group[0].OUTSET.reg = PORT_PA15;

  I2CAddress0 = ~PORT->Group[0].IN.reg & PORT_PA14 ? 1 : 0; // read value of input pin. Input has pull-up so the actual value is inverted. this converts back.
  I2CAddress1 = ~PORT->Group[0].IN.reg & PORT_PA15 ? 1 : 0; // read value of input pin
  I2CAddress2 = ~PORT->Group[1].IN.reg & PORT_PB15 ? 1 : 0; // read value of input pin
  I2CAddress = ((I2CAddress2 << 2) | (I2CAddress1 << 1) | I2CAddress0) + 1; // combine values of input pins into one integer and add 1 (otherwise you can get the reserved address 0)
  
  I2CAddress = 5;

  Wire1.begin(I2CAddress); // start I2C with address based on jumpers
  #ifdef DEBUG
    Serial.print("I2C address = ");
    Serial.println(I2CAddress);
  #endif // DEBUG  

  for (uint8_t i = 0; i < sensorArrayLength; i++) { // sensorHit array is used to track which sensor was hit to be able to let that LED light up in different color. default state is false
    sensorHit[i] = false;
  }
  Wire1.onRequest(I2CEvent); // start listening for I2C requests from main where it is requested to send data back
  Wire1.onReceive(I2CCommand); // start listening for I2C commands send by main to this peripheral  
  delay(1000);
  PORT->Group[1].OUTCLR.reg = GREENLEDPIN; // turn green led pin off

}

void loop() 
{
  unsigned long currentTime = micros_counter;
  static uint32_t counter = 0;
  /* This is event 1 -> check for pad hit */
  if ((unsigned long)(currentTime - previousTimeEvent1) >= intervalEvent1) {
    if (enableMeasurement) {
      PORT->Group[1].OUTTGL.reg = REDLEDPIN; // toggle red led pin
    }
    else {
      counter++;
      if (counter % 10 == 0) {
        PORT->Group[1].OUTTGL.reg = REDLEDPIN; // toggle red led pin
      }
    }
    if (flashLED) {
      PORT->Group[1].OUTCLR.reg = GREENLEDPIN; // turn green led pin off
      flashLED = false;
    }

    for (uint8_t i = 0; i < sensorArrayLength; i++) {
      #ifdef DEBUG
        // Serial.print(" s"); Serial.print(i); Serial.print(" ");Serial.print(pads[i].getVelocity());
      #endif // DEBUG  
      if (sensorHit[i]){ // if true: hit is detected
        PORT->Group[1].OUTSET.reg = GREENLEDPIN; // turn green led pin on
        flashLED = true;
        currentSensorColor[i].R = ledOnRmax;
        currentSensorColor[i].G = ledOnGmax;
        currentSensorColor[i].B = ledOnBmax;
        #ifdef DEBUG
            // Serial.print("*"); // hit indicator
        #endif // DEBUG  
        sensorHit[i] = false; // reset hit
      }
      #ifdef DEBUG
        // else {
        //   Serial.print(" "); // add a white space for better alignment
        // }
      #endif // DEBUG  
    }
    #ifdef DEBUG
      // Serial.println();
    #endif // DEBUG  

   /* Update the timing for the next time around */
    previousTimeEvent1 = currentTime;
  }

  /* This is event 2 -> set LED and update colors */
  if ((unsigned long)(currentTime - previousTimeEvent2) >= intervalEvent2) {
    updateSensorColors();
    /* Update the timing for the next time around */
    previousTimeEvent2 = currentTime;
  }
}

void autoThreshold()
{
  // read sensor values when they're not pressed. subtract autoThresholdValue and save that as threshold for that sensor
  uint16_t averageSensorValue, measuredSensorValue;
  for (uint8_t i = 0; i < sensorArrayLength; i++) {
    setADCToPin(sensorArray[i]);
    measuredSensorValue = readADC(false); // false = first measurement is not disregarded
    measuredSensorValue = 0; // but we still disregard it.
    for (uint8_t j = 0; j < 15; j++) { // measuredSensorValue is uint16_t so max value is 65535. ADC is 12 bit, so max value is 4095. Max ADC value fits 16 times in a uint16_t. For safety taking 15 samples. 
        measuredSensorValue = measuredSensorValue + readADC(false); // false = first measurement is not disregarded
    }
    averageSensorValue = measuredSensorValue / 15;
    pads[i].setThreshold(averageSensorValue - autoThresholdValue);
    #ifdef DEBUG
      Serial.print("autoThreshold Pad ");
      Serial.print(i);
      Serial.print(" -> ");
      Serial.print(averageSensorValue);
      Serial.print(" - ");
      Serial.print(autoThresholdValue);
      Serial.print(" = ");
      Serial.println(averageSensorValue - autoThresholdValue);
    #endif // DEBUG  
    if (averageSensorValue <= TOOLOWSENSORVALUE) {
      #ifdef DEBUG
        Serial.println("ERROR: VALUE TOO LOW. No sensor connected? Setting threshold to minimum.");
      #endif // DEBUG  
      pads[i].setThreshold(TOOLOWTHRESHOLD);
    }
  }
  saveToFlash(); // save values to flash
}

void setLayoutColor(uint8_t layout) {
  if (layout == 0) {
    for (uint8_t i = 0; i < sensorArrayLength; i++){
      targetSensorColor[i].R = colorLayoutWickiHayden[i].R;
      targetSensorColor[i].G = colorLayoutWickiHayden[i].G;
      targetSensorColor[i].B = colorLayoutWickiHayden[i].B;
    }  
  }
  else if (layout == 1) {
    for (uint8_t i = 0; i < sensorArrayLength; i++){
      targetSensorColor[i].R = colorLayoutPiano[i].R;
      targetSensorColor[i].G = colorLayoutPiano[i].G;
      targetSensorColor[i].B = colorLayoutPiano[i].B;
    }  
  }
  else if (layout == 2) {
    for (uint8_t i = 0; i < sensorArrayLength; i++){
      targetSensorColor[i].R = colorLayoutAllOff[i].R;
      targetSensorColor[i].G = colorLayoutAllOff[i].G;
      targetSensorColor[i].B = colorLayoutAllOff[i].B;
    }  
  }
}

void setSensorColor(uint8_t sensorNumber,uint8_t valueR,uint8_t valueG,uint8_t valueB) {
  targetSensorColor[sensorNumber].R = valueR;
  targetSensorColor[sensorNumber].G = valueG;
  targetSensorColor[sensorNumber].B = valueB;
}

void transitionColor(RGB* current, RGB target, float speed) { // calculate interpolated color
    current->R += speed * (target.R - current->R);
    current->G += speed * (target.G - current->G);
    current->B += speed * (target.B - current->B);
}

void updateSensorColors()
{
  for (uint8_t i = 0; i < sensorArrayLength; i++){
    transitionColor(&currentSensorColor[i], targetSensorColor[i], colorTransitionSpeed);
    WS2812B_set_pixel(WS2812BtoSensorMap[i], currentSensorColor[i].R, currentSensorColor[i].G, currentSensorColor[i].B); // set light levels
  }
  WS2812B_update();  
}

void saveToFlash() {
  StorageData settingsToStore;
  for (int i = 0; i < sensorArrayLength; i++) {
    settingsToStore.thresholdStored[i] = pads[i].getThreshold();
    settingsToStore.maximumStored[i] = pads[i].getMaximum();
    settingsToStore.scantimeStored[i] = pads[i].getScanTime();
    settingsToStore.masktimeStored[i] = pads[i].getMaskTime();
  }
  settingsToStore.dataValid = true;
  storageData.write(settingsToStore);  // Write the data to flash
}