/*
** Firmware for Marimbatron main module
** Version 1
** 
** this firmware will:
** - blink the red and green leds
** - use ISR timers to get micros
** - request sensor data using I2C (Wire1) from a sensorboard
** - convert sensorvalue to MIDI via USB
**
** TODO
** - wait for sensormodules to start before sending I2C messages (prevent deadlock) -> maybe I2C address scan first?
** - implement wicki-hayden and piano keylayout
** - implement communication to the sensor modules
** - implement buttons and screen for UI
** - Buttons:
** - - switch wicki-hayden <-> piano keylayout
** - - menu up
** - - menu down
** - - menu enter
** - UI:
** - - Track last 4 played notes (South - West - North - East layout)
** - - Indicate velocity
** - - settings menu
** - - - show current sensor level(s)
** - - - read/set threshold per sensor
** - - - auto-set threshold
** - - - set auto-threshold value
** - - - read/set max hit level per sensor (hitting with this value should send midi velocity 127)
** - - - read/set scantime
** - - - read/set masktime
**
** - store configurable values in flash
**
**
** This work may be reproduced, modified, distributed, performed, and 
** displayed for any non-commercial, non-profit purpose, but must acknowledge 
** the developer and the Marimbatron project. 
** Copyright is retained and must be preserved. The work is 
** provided as is; no warranty is provided, and users accept all liability.
** 
** Copyright 2024 Leo Kuipers.
*/


/*
pin assignments based on variant.h
using one uart, one wire, one spi, USB Config: WITH_CDC:

UART
PIN_USB_DM (24ul)-> PA24
PIN_USB_DP (25ul)-> PA25

I2C (sercom 1)
PIN_WIRE_SDA (16u)-> PA16
PIN_WIRE_SCL (17u)-> PA17

SPI (sercom3)
PIN_SPI_MISO (22u)-> PA22
PIN_SPI_MOSI (18u)-> PA18
PIN_SPI_SCK (19u)-> PA19
PIN_SPI_SS (23u)-> PA23
(HW SS isn't used)
*/
#define FIRMWAREVERSION 1

#define REDLEDPIN PORT_PA00 // PA00
#define GREENLEDPIN PORT_PA01 // PA01

#include <Wire.h>
#include "MIDIUSB.h"

// timers
#define CPU_HZ 48000000 // The cpu clock is running at 48 MHz
#define TICK_HZ_100US 10000 // We want a 100 us tick.
#define COUNTS_PER_TICK_100US (CPU_HZ / TICK_HZ_100US) // This is the number of clock cycles per tick
#define TICK_HZ_1MS 100000 // We want a 1ms tick.
#define COUNTS_PER_TICK_1MS (CPU_HZ / TICK_HZ_1MS) // This is the number of clock cycles per tick
#define NUMBER_OF_OCTAVES 1
volatile uint32_t micros_counter = 0; // holds the count of microseconds since the MCU started (don't want to use micros())
const unsigned long intervalEvent1 = 5000; // in microseconds
const unsigned long intervalEvent2 = 100000; // in microseconds
unsigned long previousTimeEvent1 = 0;
unsigned long previousTimeEvent2 = 0;
const uint8_t sensorArrayLength = 19;
bool ledIndicator = false;
const uint32_t midiNoteOffTimeout = 100000; // 100ms = 100000us
const uint8_t midiChannel = 0;
uint8_t currentOctave = 4; // Current octave refers to the octave of the first sensorboard. octave 4 is the standard center octave. C4 is MIDI note 60 (middle C).

// midi notes mapped to wicki hayden layout:
const uint8_t noteArrayWickiHayden[sensorArrayLength] = 
{
  47, // s0 -> Cb3
  49, // s1 -> Dd3
  51, // s2 -> Eb3
  53, // s3 -> F3
  55, // s4 -> G3
  57, // s5 -> A3
  59, // s6 -> B3
  49, // s7 -> C#3
  51, // s8 -> D#3
  53, // s9 -> E#3
  54, // s10 -> Gb3
  56, // s11 -> Ab3
  58, // s12 -> Bb3
  60, // s13 -> C4
  62, // s14 -> D4
  64, // s15 -> E4
  66, // s16 -> F#4
  68, // s17 -> G#4
  70 // s18 -> A#4
};
// midi notes mapped to wicki hayden layout:
const uint8_t noteArrayPiano[sensorArrayLength] = 
{
  53, // s0 -> F3
  55, // s1 -> G3
  57, // s2 -> A3
  59, // s3 -> B3
  60, // s4 -> C4
  62, // s5 -> D4
  64, // s6 -> E4
  65, // s7 -> F4
  67, // s8 -> G4
  69, // s9 -> A4
  54, // s10 -> F#3
  56, // s11 -> G#3
  70, // s12 -> A#4
  0, // s13 -> none
  73, // s14 -> C#4
  75, // s15 -> D#4
  0, // s16 -> none
  78, // s17 -> F#4
  80 // s18 -> G#4
};

uint8_t noteArray[sensorArrayLength];  // array to hold currently assigned notes
const uint32_t sensorThreshold[sensorArrayLength] = 
{
  3300, // sensor 0
  2700, // sensor 1
  1700, // sensor 2
  2600, // sensor 3
  2600, // sensor 4
  2600,// sensor 5
  2600,// sensor 6
  2600,// sensor 7
  2600,// sensor 8
  2600,// sensor 9
  2600,// sensor 10
  2600,// sensor 11
  2600,// sensor 12
  2600,// sensor 13
  2600,// sensor 14
  2600,// sensor 15
  2600,// sensor 16
  2600,// sensor 17
  2600 // sensor 18
};
const uint32_t sensorMaximum[sensorArrayLength] = 
{
  1900, // sensor 0
  450, // sensor 1
  600, // sensor 2
  400, // sensor 3
  400, // sensor 4
  400,// sensor 5
  400,// sensor 6
  400,// sensor 7
  400,// sensor 8
  400,// sensor 9
  400,// sensor 10
  400,// sensor 11
  400,// sensor 12
  400,// sensor 13
  400,// sensor 14
  400,// sensor 15
  400,// sensor 16
  400,// sensor 17
  400 // sensor 18
};
// arrays to keep track of the sensor note status (will be set to true when hit and set to false after midi note-off timeout)
bool sensorHit1[sensorArrayLength];
bool sensorHit2[sensorArrayLength];
bool sensorHit3[sensorArrayLength];
bool sensorHit4[sensorArrayLength];
bool* sensorHit[NUMBER_OF_OCTAVES]{ // array of pointers to sensorHit arrays (1 per octave)
  sensorHit1
  // sensorHit1,
  // sensorHit2,
  // sensorHit3,
  // sensorHit4
};

// arrays to keep track of the last time (micros_counter) that a hit was detected on a sensor
uint32_t sensorHitTime1[sensorArrayLength];
uint32_t sensorHitTime2[sensorArrayLength];
uint32_t sensorHitTime3[sensorArrayLength];
uint32_t sensorHitTime4[sensorArrayLength];
uint32_t* sensorHitTime[NUMBER_OF_OCTAVES]{ // array of pointers to sensorHitTime arrays (1 per octave)
  sensorHitTime1
  // sensorHitTime1,
  // sensorHitTime2,
  // sensorHitTime3,
  // sensorHitTime4
};

void TC4_Handler() { // 100us timer, priority 0
  // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope
  micros_counter = micros_counter + 100;
  // Clear the interrupt
  TC4->COUNT16.INTFLAG.bit.MC0 = 1;
}

void TC5_Handler() { // 1 ms timer, priority 1
  // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope
  // I2CRequestData();
  // Clear the interrupt
  TC5->COUNT16.INTFLAG.bit.MC0 = 1;
}

void setup() {
  // set SAMD to run at 48Mhz
  setup48MhzUSBClockRecoveryMode();
  setupHardwareTimers();  
  Wire.begin(); // join i2c bus as master (so no address is set)
  Wire.setClock(100000);

  Serial.begin(115200);
  while (!Serial && micros_counter < 5000000); // wait for serial bus to activate (in order to communicate immediately after reset) or timeout
  delay(100);
  Serial.print("Marimbatron Main board firmware v");
  Serial.println(FIRMWAREVERSION);

  PORT->Group[0].DIRSET.reg = REDLEDPIN; // set led pin as output
  PORT->Group[0].DIRSET.reg = GREENLEDPIN; // set led pin as output
  PORT->Group[0].OUTSET.reg = REDLEDPIN; // turn led pin on
  PORT->Group[0].OUTCLR.reg = GREENLEDPIN; // turn led pin off

  for (uint8_t i = 0; i < sensorArrayLength; i++) { // assign wicki hayden note layout as default, init arrays sensorHitx and SensorHitTimex
    noteArray[i] = noteArrayWickiHayden[i];
    sensorHit1[i] = false;
    sensorHit2[i] = false;
    sensorHit3[i] = false;
    sensorHit4[i] = false;
    sensorHitTime1[i] = 0;
    sensorHitTime2[i] = 0;
    sensorHitTime3[i] = 0;
    sensorHitTime4[i] = 0;
  }

  delay(2000);
  PORT->Group[0].OUTCLR.reg = REDLEDPIN; // turn led pin off
  PORT->Group[0].OUTSET.reg = GREENLEDPIN; // turn led pin on
  Serial.println("Start I2C bus scan.");
  scanI2Cbus();
  delay(5000);
  PORT->Group[0].OUTSET.reg = REDLEDPIN; // turn led pin on
  PORT->Group[0].OUTCLR.reg = GREENLEDPIN; // turn led pin off
}

void loop() {
  unsigned long currentTime = micros_counter;
  /* This is event 1 -> check for pad hit */
  if ((unsigned long)(currentTime - previousTimeEvent1) >= intervalEvent1) {
    // Serial.println("I2Crequestdata()");
    I2CRequestData();
   /* Update the timing for the next time around */
    previousTimeEvent1 = currentTime;
  }

  /* This is event 2 -> set LED colors based on incoming I2C master commands */
  if ((unsigned long)(currentTime - previousTimeEvent2) >= intervalEvent2) {
    PORT->Group[0].OUTTGL.reg = REDLEDPIN; // toggle led pin
    if(ledIndicator)
    {
      PORT->Group[0].OUTCLR.reg = GREENLEDPIN; // pin off
      ledIndicator = false;
    }
    for (uint8_t address=1; address <= NUMBER_OF_OCTAVES; address++) { // iterate over the 4 sensor boards
      for (int i = 0; i <= sensorArrayLength; i++) 
      {
        // if (sensorHit[address][i]) {
        //   if ((unsigned long)(currentTime - sensorHitTime[address][i]) >= midiNoteOffTimeout) {
        //     noteOff(midiChannel, noteArray[i], 0); // send note off
        //     sensorHit[address][i] = false;
        //   };
        // }
        if (sensorHit1[i]) {
          if ((unsigned long)(currentTime - sensorHitTime1[i]) >= midiNoteOffTimeout) {
            noteOff(midiChannel, noteArray[i], 0); // send note off
            sensorHit1[i] = false;
          };
        }
      };
    };
   /* Update the timing for the next time around */
    previousTimeEvent2 = currentTime;
  }  
}

void scanI2Cbus() { 
  byte error, address;
  int nDevices;
  Serial.println("Scanning...");
  nDevices = 0;
  for(address = 1; address < 127; address++ )
  {
    Wire.beginTransmission(address);
    error = Wire.endTransmission();
    if (error == 0)
    {
      Serial.print("I2C device found at address 0x");
      if (address<16)
      { Serial.print("0"); }
      Serial.print(address,HEX);
      Serial.println("  !");
      nDevices++;
    }
    else if (error==4)
    {
      Serial.print("Unknown error at address 0x");
      if (address<16)
      { Serial.print("0"); }
      Serial.println(address,HEX);
    }    
  }
  if (nDevices == 0)
  { Serial.println("No I2C devices found\n"); }
  else
  { Serial.println("done\n"); }
}

void I2CRequestData() {
  uint8_t sensorNumber;
  uint16_t sensorVelocity;
  // Serial.println("requesting I2C");
  for (uint8_t address=1; address <= NUMBER_OF_OCTAVES; address++) { // iterate over the 4 sensor boards
    Wire.requestFrom(address, 3);    // requesting 5 bytes from address 0. Max number of bytes in a single transmission cannot be more than 32.
    // Serial.print("requesting I2C from address ");
    // Serial.println(address);
    // Serial.println(Wire.available());
    while(Wire.available()) {
        sensorNumber = Wire.read(); // Wire.read reads 1 byte at a time. uint8_t is 1 byte.
  //      sensorVelocity = Wire.read() | (Wire.read() << 8) | (Wire.read() << 16) | (Wire.read() << 24); // sensorVelocity is uint32_t, which is 4 bytes. So need to combine 4 reads into one uint32_t.
        sensorVelocity = Wire.read() | (Wire.read() << 8); // sensorVelocity is uint16_t, which is 3 bytes. So need to combine 2 reads into one uint16_t.
        // Serial.print(sensorNumber);         // Print the value
        // Serial.print(" ");
        // Serial.println(sensorVelocity);
        // if (sensorNumber < 255) {
          // Serial.print("Hit on sensor "); Serial.print(sensorNumber); Serial.print(" -> "); Serial.println(sensorVelocity);
          // Serial.println("Sending note on");

          if (sensorNumber < 255) {
            Serial.print("Hit on octave "); Serial.print(address); Serial.print(" sensor "); Serial.print(sensorNumber); Serial.print(" -> "); Serial.println(sensorVelocity);
            // send midi note on, making sure to get to the correct octave. 
            // there are 12 notes in an octave, so 
            // noteArray[sensorNumber] + (address - 1) * 12
            // gets the base octave (the one defined in noteArray), keeps those values for address 1 and for higher addresses adds multiples of 12, effectively increasing the octave.
            if (noteArray[sensorNumber] != 0) { // only register the hit when there's a midi note number defined on this sensor (midi note number 0 is used to "disable" the sensor)
            Serial.print("sending midi ch"); Serial.print(midiChannel); Serial.print(" note "); Serial.print(noteArray[sensorNumber] + (address - 1) * 12); Serial.print(" velocity "); Serial.println(mapToMidiVelocity(sensorVelocity, sensorThreshold[sensorNumber], sensorMaximum[sensorNumber]));
              noteOn(midiChannel, noteArray[sensorNumber] + (address - 1) * 12, mapToMidiVelocity(sensorVelocity, sensorThreshold[sensorNumber], sensorMaximum[sensorNumber])); 
              // sensorHit[address-1][sensorNumber] = true; // store that sensor hit was detected (array indices start at 0)
              // sensorHitTime[address-1][sensorNumber] = micros_counter; // store what time the hit was detected
              sensorHit1[sensorNumber] = true; // store that sensor hit was detected (array indices start at 0)
              sensorHitTime1[sensorNumber] = micros_counter; // store what time the hit was detected
            }
          }
          // switch(sensorNumber) {
          // case 0: 
          //   noteOn(0, 48, mapToMidiVelocity(sensorVelocity, 3300, 1900));   // Channel 0, middle C, normal velocity
          //   ledIndicator = true;
          //   PORT->Group[0].OUTSET.reg = GREENLEDPIN; // led pin on
          //   break;
          // case 1: 
          //   noteOn(0, 49, mapToMidiVelocity(sensorVelocity, 2700, 450));   // Channel 0, middle C, normal velocity
          //   ledIndicator = true;
          //   PORT->Group[0].OUTSET.reg = GREENLEDPIN; // led pin on
          //   break;
          // case 2: 
          //   noteOn(0, 50, mapToMidiVelocity(sensorVelocity, 1700, 600));   // Channel 0, middle C, normal velocity
          //   ledIndicator = true;
          //   PORT->Group[0].OUTSET.reg = GREENLEDPIN; // led pin on
          //   break;
          // }        
        // if (sensorNumber < 255) {
        //   delay(10);
        //   noteOff(0, 48, 0);  // Channel 0, middle C, normal velocity
        // }
          // noteOn(0, sensorNumber, mapToMidiVelocity(sensorVelocity, 2700, 800));   // Channel 0, middle C, normal velocity
          // noteOn(0, 48, mapToMidiVelocity(sensorVelocity, 2700, 800));   // Channel 0, middle C, normal velocity
          // MidiUSB.flush();
          // Serial.println("Sending note off");
          // MidiUSB.flush();
        // }
        // else{
        //   Serial.println("no hit");
        // }
    }
  }
}

int mapToMidiVelocity(uint16_t input, uint16_t threshold, uint16_t maximum) {
    if(input < maximum) input = maximum; // make sure that input value cannot be lower than maximum value
    if(input > threshold) input = threshold; // make sure that input value cannot be lower than threshold value
    return (uint8_t)(127 - ((input - maximum) * 127) / (threshold - maximum)); // linear map threshold ~ maximum to 0 ~ 127
}

// First parameter is the event type (0x09 = note on, 0x08 = note off).
// Second parameter is note-on/note-off, combined with the channel.
// Channel can be anything between 0-15. Typically reported to the user as 1-16.
// Third parameter is the note number (48 = middle C).
// Fourth parameter is the velocity (64 = normal, 127 = fastest).

void noteOn(byte channel, byte pitch, byte velocity) {
  midiEventPacket_t noteOn = {0x09, 0x90 | channel, pitch, velocity};
  MidiUSB.sendMIDI(noteOn);
  // serializeMidiToUART(0x09, 0x90 | channel, pitch, velocity);
}

void noteOff(byte channel, byte pitch, byte velocity) {
  midiEventPacket_t noteOff = {0x08, 0x80 | channel, pitch, velocity};
  MidiUSB.sendMIDI(noteOff);
  // serializeMidiToUART(0x08, 0x80 | channel, pitch, velocity);
}

void serializeMidiToLED(byte event, byte channel, byte pitch, byte velocity) {  
  byte midiData[] = {channel, pitch, velocity};
  Serial.write(midiData, 3);
}

void setup48MhzUSBClockRecoveryMode(){
  // from https://blog.thea.codes/understanding-the-sam-d21-clocks/
  /* Set the correct number of wait states for 48 MHz @ 3.3v */
  NVMCTRL->CTRLB.bit.RWS = 1;

  /* This works around a quirk in the hardware (errata 1.2.1) -
    the DFLLCTRL register must be manually reset to this value before
    configuration. */
  while(!SYSCTRL->PCLKSR.bit.DFLLRDY);
  SYSCTRL->DFLLCTRL.reg = SYSCTRL_DFLLCTRL_ENABLE;
  while(!SYSCTRL->PCLKSR.bit.DFLLRDY);

  /* Write the coarse and fine calibration from NVM. */
  uint32_t coarse =
      ((*(uint32_t*)FUSES_DFLL48M_COARSE_CAL_ADDR) & FUSES_DFLL48M_COARSE_CAL_Msk) >> FUSES_DFLL48M_COARSE_CAL_Pos;
  uint32_t fine =
      ((*(uint32_t*)FUSES_DFLL48M_FINE_CAL_ADDR) & FUSES_DFLL48M_FINE_CAL_Msk) >> FUSES_DFLL48M_FINE_CAL_Pos;

  SYSCTRL->DFLLVAL.reg = SYSCTRL_DFLLVAL_COARSE(coarse) | SYSCTRL_DFLLVAL_FINE(fine);

  /* Wait for the write to finish. */
  while (!SYSCTRL->PCLKSR.bit.DFLLRDY) {};

  SYSCTRL->DFLLCTRL.reg |=
      /* Enable USB clock recovery mode */
      SYSCTRL_DFLLCTRL_USBCRM |
      /* Disable chill cycle as per datasheet to speed up locking.
        This is specified in section 17.6.7.2.2, and chill cycles
        are described in section 17.6.7.2.1. */
      SYSCTRL_DFLLCTRL_CCDIS;

  /* Configure the DFLL to multiply the 1 kHz clock to 48 MHz */
  SYSCTRL->DFLLMUL.reg =
      /* This value is output frequency / reference clock frequency,
        so 48 MHz / 1 kHz */
      SYSCTRL_DFLLMUL_MUL(48000) |
      /* The coarse and fine values can be set to their minimum
        since coarse is fixed in USB clock recovery mode and
        fine should lock on quickly. */
      SYSCTRL_DFLLMUL_FSTEP(1) |
      SYSCTRL_DFLLMUL_CSTEP(1);

  /* Closed loop mode */
  SYSCTRL->DFLLCTRL.bit.MODE = 1;

  /* Enable the DFLL */
  SYSCTRL->DFLLCTRL.bit.ENABLE = 1;

  /* Wait for the write to complete */
  while (!SYSCTRL->PCLKSR.bit.DFLLRDY) {};      

  /* Setup GCLK0 using the DFLL @ 48 MHz */
  GCLK->GENCTRL.reg =
      GCLK_GENCTRL_ID(0) |
      GCLK_GENCTRL_SRC_DFLL48M |
      /* Improve the duty cycle. */
      GCLK_GENCTRL_IDC |
      GCLK_GENCTRL_GENEN;

  /* Wait for the write to complete */
  while(GCLK->STATUS.bit.SYNCBUSY) {};

  // for debugging: temporarily set 48MHz clock to GCLK4 as well and output to a pin to check accuracy.
  // GCLK->GENCTRL.reg =
  //     GCLK_GENCTRL_ID(4) |
  //     GCLK_GENCTRL_SRC_DFLL48M |
  //     GCLK_GENCTRL_IDC |
  //     GCLK_GENCTRL_GENEN |
  //     /* Enable outputting the clock to an I/O pin. */
  //     GCLK_GENCTRL_OE;

  // /* Wait for the write to complete */
  // while(GCLK->STATUS.bit.SYNCBUSY);
  // PORT->Group[1].DIRSET.reg = (1 << 10);
  // PORT->Group[1].PINCFG[10].reg |= PORT_PINCFG_PMUXEN;
  // PORT->Group[1].PMUX[10 >> 1].bit.PMUXE |= PORT_PMUX_PMUXE_H;
}

void setupHardwareTimers() 
{
    // Set up the generic clock (GCLK4) used to clock timers
    GCLK->GENDIV.reg = GCLK_GENDIV_DIV(1) |          // Divide the 48MHz clock source by divisor 1: 48MHz/1=48MHz
                       GCLK_GENDIV_ID(4);            // Select Generic Clock (GCLK) 4
    while (GCLK->STATUS.bit.SYNCBUSY);               // Wait for synchronization

    GCLK->GENCTRL.reg = GCLK_GENCTRL_IDC |           // Set the duty cycle to 50/50 HIGH/LOW
                        GCLK_GENCTRL_GENEN |         // Enable GCLK4
                        GCLK_GENCTRL_SRC_DFLL48M |   // Set the 48MHz clock source
                        GCLK_GENCTRL_ID(4);          // Select GCLK4
    while (GCLK->STATUS.bit.SYNCBUSY);               // Wait for synchronization

    // // Feed GCLK4 to TC3
    // GCLK->CLKCTRL.reg = GCLK_CLKCTRL_CLKEN |         // Enable GCLK4
    //                     GCLK_CLKCTRL_GEN_GCLK4 |     // Select GCLK4
    //                     GCLK_CLKCTRL_ID_TCC2_TC3;     // Feed the GCLK4 to TC3 (and TCC2)
    // while (GCLK->STATUS.bit.SYNCBUSY);               // Wait for synchronization

    // Feed GCLK4 to TC4 and TC5
    GCLK->CLKCTRL.reg = GCLK_CLKCTRL_CLKEN |         // Enable GCLK4
                        GCLK_CLKCTRL_GEN_GCLK4 |     // Select GCLK4
                        GCLK_CLKCTRL_ID_TC4_TC5;     // Feed the GCLK4 to TC4 and TC5
    while (GCLK->STATUS.bit.SYNCBUSY);               // Wait for synchronization

    // // specific for TC3: 1ms
    // TC3->COUNT16.CTRLA.reg |= TC_CTRLA_MODE_COUNT16;  // Set Timer counter Mode to 16 bits
    // while (TC3->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // TC3->COUNT16.CTRLA.reg |= TC_CTRLA_WAVEGEN_MFRQ;  // Set TC as Match Frequency
    // while (TC3->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // TC3->COUNT16.CC[0].reg = COUNTS_PER_TICK_1MS;     // Set the TC3 CC0 register as the TOP value in match frequency mode
    // while (TC3->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // // Enable the TC3 interrupt request
    // TC3->COUNT16.INTENSET.bit.MC0 = 1;
    // while (TC3->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // // Enable TC3
    // TC3->COUNT16.CTRLA.reg |= TC_CTRLA_ENABLE;        // Enable TC3
    // while (TC3->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // // Enable the TC3 interrupt vector use priority 2
    // NVIC_SetPriority(TC3_IRQn, 2);                    // Set the NVIC priority for TC3 to 2
    // NVIC_EnableIRQ(TC3_IRQn);

    // specific for TC4: 100us
    TC4->COUNT16.CTRLA.reg |= TC_CTRLA_MODE_COUNT16;  // Set Timer counter Mode to 16 bits
    while (TC4->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    TC4->COUNT16.CTRLA.reg |= TC_CTRLA_WAVEGEN_MFRQ;  // Set TC as Match Frequency
    while (TC4->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    TC4->COUNT16.CC[0].reg = COUNTS_PER_TICK_100US;     // Set the TC4 CC0 register as the TOP value in match frequency mode
    while (TC4->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable the TC4 interrupt request
    TC4->COUNT16.INTENSET.bit.MC0 = 1;
    while (TC4->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable TC4
    TC4->COUNT16.CTRLA.reg |= TC_CTRLA_ENABLE;        // Enable TC4
    while (TC4->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable the TC4 interrupt vector use priority 0
    NVIC_SetPriority(TC4_IRQn, 0);                    // Set the NVIC priority for TC4 to 0
    NVIC_EnableIRQ(TC4_IRQn);

    // specific for TC5: 1ms
    TC5->COUNT16.CTRLA.reg |= TC_CTRLA_MODE_COUNT16;  // Set Timer counter Mode to 16 bits
    while (TC5->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    TC5->COUNT16.CTRLA.reg |= TC_CTRLA_WAVEGEN_MFRQ;  // Set TC as Match Frequency
    while (TC5->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    TC5->COUNT16.CC[0].reg = COUNTS_PER_TICK_1MS;     // Set the TC5 CC0 register as the TOP value in match frequency mode
    while (TC5->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable the TC5 interrupt request
    TC5->COUNT16.INTENSET.bit.MC0 = 1;
    while (TC5->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable TC5
    TC5->COUNT16.CTRLA.reg |= TC_CTRLA_ENABLE;        // Enable TC5
    while (TC5->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable the TC5 interrupt vector use priority 1
    NVIC_SetPriority(TC5_IRQn, 1);                    // Set the NVIC priority for TC5 to 1
    NVIC_EnableIRQ(TC5_IRQn);
}
