/*
** Firmware for Marimbatron main module
** Version 2
** 
** this firmware will:
** - blink the red and green leds
** - use ISR timers to get micros
** - request sensor data using I2C (Wire1) from a sensorboard
** - convert sensorvalue to MIDI via USB
** - implement I2C communication protocol to communicate from main board to sensor board (see below for protocol)
**
** TODO
** - implement buttons and screen for UI
** - Buttons:
** - - switch wicki-hayden <-> piano keylayout (in menu mode: back button)
** - - menu up (octave up)
** - - menu down (octave down)
** - - menu enter (enter to enter menu and to confirm)
** - UI:
** - - Track last 4 played notes (South - West - North - East layout)
** - - Indicate velocity
** - - settings menu
** - - - show current sensor of last hit sensor
** - - - read/set thresholds
** - - - auto-set thresholds
** - - - set auto-threshold value
** - - - read/set max hit level per sensor (hitting with this value should send midi velocity 127)
** - - - read/set scantime
** - - - read/set masktime
**
** - store configurable values in non-volatile flash
**
**
** Required libraries:
** wire.h
**
** I2C Marimbatron protocol:
** C = type: Command to execute - followed by character that defines what command
** S = type: Set - followed by character that defines what to set (sometimes followed by uint_8 to indicate sensor number) 
** G = type: Get - followed by character that defines what to get (sometimes followed by uint_8 to indicate sensor number)

** main sends chars XR -> peripheral will reset (NVIC_SystemReset())
** main sends chars XA -> peripheral will do auto-threshold procedure
** main sends chars XS -> peripheral will stop sensor measurements. All detected hits that we're not sent yet will be flushed.

** main sends chars GH -> peripheral is set into Get sensor Hit mode -> 
                                                      on datarequest (3 bytes) it sends sensorvalue of sensor with hit = true (or 255,0,0 if no hit)
                                                      All detected hits that we're not sent yet will be flushed.
** main sends chars GC followed by byte sensornumber -> peripheral is set into Get Continuous actual value -> on datarequest (3 bytes) it sends sensorvalue of [sensornumber]
                                                      and on next datarequest it sends sensorvalue of sensor with hit = true (or 255,0,0 if no hit)
                                                      on next datarequest back to sensorvalue of [sensornumber], etc.
** main sends chars GA followed by byte sensornumber -> peripheral set into Get single Actual values mode -> 
                                                        on datarequest (2 bytes) it sends actual sensorvalue of [sensornumber].
** main sends chars GT followed by byte sensornumber -> peripheral set to Get Threshold mode -> 
                                                        on datarequest (2 bytes) it sends current threshold value of [sensornumber].
** main sends chars GX followed by byte sensornumber -> peripheral set to Maximum mode -> 
                                                        on datarequest (2 bytes) it sends current maximum value of [sensornumber].
** main sends chars GM -> peripheral set into Masktime mode -> 
                                                        on datarequest it sends the current masktime.
** main sends chars GN -> peripheral set into Scantime mode -> 
                                                        on datarequest it sends the current scantime.

** main sends chars ST followed by byte sensornumber followed by uint_16 value -> peripheral sets threshold of sensor [sensornumber] to [value].
** main sends chars SX followed by byte sensornumber followed by uint_16 value -> peripheral sets maximum of sensor [sensornumber] to [value].
** main sends chars SM followed by byte sensornumber followed by uint_32 value -> peripheral sets Masktime of sensor [sensornumber] to [value].
** main sends chars SN followed by byte sensornumber followed by uint_32 value -> peripheral sets Scantime of sensor [sensornumber] to [value].
**
** This work may be reproduced, modified, distributed, performed, and 
** displayed for any non-commercial, non-profit purpose, but must acknowledge 
** the developer and the Marimbatron project. 
** Copyright is retained and must be preserved. The work is 
** provided as is; no warranty is provided, and users accept all liability.
** 
** Copyright 2024 Leo Kuipers.
*/


/*
pin assignments based on variant.h
using one uart, one wire, one spi, USB Config: WITH_CDC:

UART
PIN_USB_DM (24ul)-> PA24
PIN_USB_DP (25ul)-> PA25

I2C (sercom 1)
PIN_WIRE_SDA (16u)-> PA16
PIN_WIRE_SCL (17u)-> PA17

SPI (sercom3)
PIN_SPI_MISO (22u)-> PA22
PIN_SPI_MOSI (18u)-> PA18
PIN_SPI_SCK (19u)-> PA19
PIN_SPI_SS (23u)-> PA23
(HW SS isn't used)
*/
#define FIRMWAREVERSION 2

#define REDLEDPIN PORT_PA00 // PA00
#define GREENLEDPIN PORT_PA01 // PA01

#include <Wire.h>
#include "MIDIUSB.h"

// timers
// CPU clock is 48 MHz (48.000.000 Hz)
// so 1 CPU clock cycle is about 20,83ns. 
// timer_hander is called every time the timer overflows.
// which is here programmed as COUNT_PER_TICK * 1_CPU_CLOCKCYCLE.
// for a 100 us timer, COUNT_PER_TICK = 10.000 ticks because
// 48.000.000 / 10.000 = 4.800 and 48.000 * 20,83ns = 99,984us which is about 100us.

#define CPU_HZ 48000000 // The cpu clock is running at 48 MHz
#define TICK_HZ_100US 10000 // 100 us tick
#define COUNTS_PER_TICK_100US (CPU_HZ / TICK_HZ_100US) // number of clock cycles per tick (4800)
#define TICK_HZ_TC5 1000000 // 10ms tick
#define COUNTS_PER_TICK_TC5 (CPU_HZ / TICK_HZ_TC5) // This is the number of clock cycles per tick (480000)

volatile uint32_t micros_counter = 0; // holds the count of microseconds since the MCU started (don't want to use micros())
const unsigned long intervalEvent1 = 10000; // in microseconds
const unsigned long intervalEvent2 = 50000; // in microseconds
const unsigned long intervalEvent3 = 500000; // in microseconds
unsigned long previousTimeEvent1 = 0;
unsigned long previousTimeEvent2 = 0;
unsigned long previousTimeEvent3 = 0;
const uint8_t sensorArrayLength = 19;
uint32_t ledIndicatorTime = 0; // timer that holds info if hit-indicator led should be turned off
const uint32_t midiNoteOffTimeout = 100000; // 100ms = 100000us
const uint8_t midiChannel = 0;
bool enableRequestSensorData = false;
const uint8_t numberOfSensorBoards = 4;
uint8_t currentLayout = 0; // assign wicki hayden note layout = default
uint8_t currentOctave = 4; // Current octave refers to the octave of the first sensorboard. octave 4 is the standard center octave. C4 is MIDI note 60 (middle C).
uint8_t I2CSensorboardAddresses[numberOfSensorBoards]; // array to hold the addresses of the sensorboards

// midi notes mapped to wicki hayden layout:
const int8_t noteArrayWickiHayden[sensorArrayLength] =  // int8_t is signed integer, values from -128 to 127. Midi defines 128 notes, from 0 to 127.
{
  -1, //47, // s0 -> Cb3
  1, //49, // s1 -> Dd3
  3, //51, // s2 -> Eb3
  5, //53, // s3 -> F3
  7, //55, // s4 -> G3
  9, //57, // s5 -> A3
  11, //59, // s6 -> B3
  1, //49, // s7 -> C#3
  3, //51, // s8 -> D#3
  5, //53, // s9 -> E#3
  6, //54, // s10 -> Gb3
  8, //56, // s11 -> Ab3
  10, //58, // s12 -> Bb3
  12, //60, // s13 -> C4
  14, //62, // s14 -> D4
  16, //64, // s15 -> E4
  18, //66, // s16 -> F#4
  20, //68, // s17 -> G#4
  22, //70 // s18 -> A#4
};
// midi notes mapped to piano layout. -128 means note is disabled:
const uint8_t noteArrayPiano[sensorArrayLength] = 
{
  5, //53, // s0 -> F3
  7, //55, // s1 -> G3
  9, //57, // s2 -> A3
  11, //59, // s3 -> B3
  12, //60, // s4 -> C4
  14, //62, // s5 -> D4
  16, //64, // s6 -> E4
  17, //65, // s7 -> F4
  19, //67, // s8 -> G4
  21, //69, // s9 -> A4
  6, //54, // s10 -> F#3
  8, //56, // s11 -> G#3
  10, //58, // s12 -> A#3
  -128, // s13 -> none
  25, //73, // s14 -> C#4
  27, //75, // s15 -> D#4
  -128, // s16 -> none
  18, //66 s17 -> F#4
  20 //68 s18 -> G#4
};

uint8_t noteArray[numberOfSensorBoards][sensorArrayLength];  // array to hold currently assigned notes
uint16_t sensorThreshold[numberOfSensorBoards][sensorArrayLength];
uint16_t sensorMaximum[numberOfSensorBoards][sensorArrayLength];

// arrays to keep track of the sensor note status (will be set to true when hit and set to false after midi note-off timeout)
bool sensorHit[numberOfSensorBoards][sensorArrayLength];

// arrays to keep track of the last time (micros_counter) that a hit was detected on a sensor
uint32_t sensorHitTime[numberOfSensorBoards][sensorArrayLength];

void TC4_Handler() { // 100us timer, priority 0
  micros_counter = micros_counter + 100; // add 100 us to the micros_counter
  TC4->COUNT16.INTFLAG.bit.MC0 = 1;  // Clear the interrupt
}

void TC5_Handler() { // sensor timer, priority 1
  TC5->COUNT16.INTFLAG.bit.MC0 = 1;  // Clear the interrupt
}

void setup() {
  uint8_t returnValue = 0;
  uint8_t numberOfI2CDevices = 0;
  // set SAMD to run at 48Mhz
  setup48MhzUSBClockRecoveryMode();
  setupHardwareTimers();  

  Serial.begin(115200);
  while (!Serial && micros_counter < 5000000); // wait for serial bus to activate (in order to communicate immediately after reset) or timeout
  delay(100);
  Serial.print("Marimbatron Main board firmware v");
  Serial.println(FIRMWAREVERSION);

  PORT->Group[0].DIRSET.reg = REDLEDPIN; // set led pin as output
  PORT->Group[0].DIRSET.reg = GREENLEDPIN; // set led pin as output
  PORT->Group[0].OUTSET.reg = REDLEDPIN; // turn RED on
  PORT->Group[0].OUTCLR.reg = GREENLEDPIN; // turn GREEN off

  Serial.println("init sensorHit and SensorHittime");
  // init arrays sensorHit[x][y] and sensorHitTime[x][y]
  for (uint8_t sensorboard = 0; sensorboard < numberOfSensorBoards; sensorboard++) { 
    for (uint8_t sensor = 0; sensor < sensorArrayLength; sensor++) {
      sensorHit[sensorboard][sensor] = false; // set all sensor hits to false
      sensorHitTime[sensorboard][sensor] = 0; // set all sensor hit times to 0
    }
  }
  Serial.println("checkAndSolveI2CDeadlock()");
  checkAndSolveI2CDeadlock();
  Wire.begin(); // join i2c bus as master (so no address is set)
  // Wire.setClock(10000); // 10kBaud -> low speed mode
  // Wire.setClock(100000); // 100kBaud -> standard speed mode
  Wire.setClock(400000); // 400kBaud -> high speed mode
  delay(1000);  
  numberOfI2CDevices = scanI2Cbus();
  delay(1000);  
  if (numberOfSensorBoards == numberOfI2CDevices) {
    Serial.println("found following sensorboards:");
    for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) {
      Serial.print(sensorboard);Serial.print(" @ ");Serial.println(I2CSensorboardAddresses[sensorboard]);
    }
    PORT->Group[0].OUTSET.reg = GREENLEDPIN; // turn GREEN on

    Serial.println("assignNotes()");
    assignNotes();

    enableSensorMeasurements(false); // disable sensor measurements for all sensorboards
    delay(100);

    // for testing without sensors attached only only
    delay(100);
    for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) { // iterate over the 4 sensor boards
      for (uint8_t sensor = 0; sensor < sensorArrayLength; sensor++) { // iterate over all 19 sensors
        setSensorThreshold(sensorboard, sensor, 1000);
        // delay(10);
      }
    }
    delay(100);
    for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) { // iterate over the 4 sensor boards
      for (uint8_t sensor = 0; sensor < sensorArrayLength; sensor++) { // iterate over all 19 sensors
        setSensorMaximum(sensorboard, sensor, 100);
        // delay(10);
      }
    }
    // end testing

    Serial.println("executeAutoThreshold()");
    //executeAutoThreshold();
    Serial.println("getSensorsThreshold()");
    getSensorsThreshold();
    Serial.println("getSensorsMaximum()");
    getSensorsMaximum();

    PORT->Group[0].OUTCLR.reg = REDLEDPIN; // turn RED off
    PORT->Group[0].OUTCLR.reg = GREENLEDPIN; // turn GREEN off
    Serial.println("Entering loop()");
    enableSensorMeasurements(true);
  }
  else {
    Serial.print("Error connecting to sensorboards. Found "); Serial.print(numberOfI2CDevices);Serial.print(" instead of ");Serial.println(numberOfSensorBoards);
    while(1) { // error reading number I2C devices. Do not proceed.
      PORT->Group[0].OUTTGL.reg = REDLEDPIN; // toggle RED
      delay(100);
    }
  }
}

void loop() {
  unsigned long currentTime = micros_counter;
  /* This is event 1 -> request sensor data */
  if ((unsigned long)(currentTime - previousTimeEvent1) >= intervalEvent1) {
    if (enableRequestSensorData) {
      I2CRequestSensorData();
    }

   /* Update the timing for the next time around */
    previousTimeEvent1 = currentTime;
  }

  /* This is event 2 -> check for and send noteoff */
  if ((unsigned long)(currentTime - previousTimeEvent2) >= intervalEvent2) {
    if ((unsigned long)(currentTime - ledIndicatorTime) >= midiNoteOffTimeout) {
      PORT->Group[0].OUTCLR.reg = REDLEDPIN; // turn led pin off
    }
    for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) { // iterate over the 4 sensor boards
      for (uint8_t sensor = 0; sensor < sensorArrayLength; sensor++) { // iterate over all 19 sensors
        if (sensorHit[sensorboard][sensor]) {
          if ((unsigned long)(currentTime - sensorHitTime[sensorboard][sensor]) >= midiNoteOffTimeout) {
            noteOff(midiChannel, noteArray[sensorboard][sensor], 0); // send note off
            sensorHit[sensorboard][sensor] = false;
          };
        }
      };
    };
   /* Update the timing for the next time around */
    previousTimeEvent2 = currentTime;
  }  
  /* This is event 3 -> blink Green LED */
  if ((unsigned long)(currentTime - previousTimeEvent3) >= intervalEvent3) {
    PORT->Group[0].OUTTGL.reg = GREENLEDPIN; // toggle green led pin as indicator that the main controller is still operating properly

   /* Update the timing for the next time around */
    previousTimeEvent3 = currentTime;
  }
}

void I2CRequestSensorData() {
  uint8_t sensorNumber;
  uint16_t sensorVelocity;
  // Serial.println("requesting I2C sensordata");
  for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) { // iterate over the sensor boards
    Wire.requestFrom(I2CSensorboardAddresses[sensorboard], 3); // requesting 3 bytes from sensorboard. Max number of bytes in a single transmission cannot be more than 32.
    // Serial.print("requesting I2C from address ");
    // Serial.println(I2CSensorboardAddresses[sensorboard]);
    // Serial.println(Wire.available());

    sensorNumber = Wire.read(); // Wire.read reads 1 byte at a time. uint8_t is 1 byte.
    sensorVelocity = Wire.read() | (Wire.read() << 8); // sensorVelocity is uint16_t, which is 2 bytes. So need to combine 2 reads into one uint16_t.
    // Serial.print(sensorNumber);         // Print the value
    // Serial.print(" ");
    // Serial.println(sensorVelocity);

    if (sensorNumber < 255) { // if returned sensorNumber = 255 there is no hit
      Serial.print("Hit on sensorboard "); Serial.print(sensorboard); Serial.print(" at address "); Serial.print(I2CSensorboardAddresses[sensorboard]); Serial.print(" sensor "); Serial.print(sensorNumber); Serial.print(" -> "); Serial.println(sensorVelocity);
      // send midi note on, making sure to get to the correct octave. 
      // only register the hit when there's a midi note number defined on this sensor
      // (midi note number -128 is used to "disable" the sensor)

      ledIndicatorTime = micros_counter;
      PORT->Group[0].OUTSET.reg = REDLEDPIN; // turn RED on
      if (noteArray[sensorboard][sensorNumber] != -128) { 
        Serial.print("sending midi: ch"); Serial.print(midiChannel); Serial.print(" note "); Serial.print(noteArray[sensorboard][sensorNumber]); 
        Serial.print(" velocity "); Serial.println(mapToMidiVelocity(sensorVelocity, sensorThreshold[sensorboard][sensorNumber], sensorMaximum[sensorboard][sensorNumber]));
        noteOn(midiChannel, noteArray[sensorboard][sensorNumber], mapToMidiVelocity(sensorVelocity, sensorThreshold[sensorboard][sensorNumber], sensorMaximum[sensorboard][sensorNumber])); 
        sensorHit[sensorboard][sensorNumber] = true; // store that sensor hit was detected
        sensorHitTime[sensorboard][sensorNumber] = micros_counter; // store what time the hit was detected
      }
    }
    // else{
    //   Serial.println("no hit");
    // }
  }
}

void octaveUp() {
  if (currentOctave <= 5) {
    currentOctave = currentOctave + 1;
    assignNotes();
  }
}

void octaveDown() {
  if (currentOctave >= 2) {
    currentOctave = currentOctave - 1;
    assignNotes();
  }
}

void assignNotes() {
  for (uint8_t sensorboard = 0; sensorboard < numberOfSensorBoards; sensorboard++) { 
    for (uint8_t sensor = 0; sensor < sensorArrayLength; sensor++) {
      if (currentLayout == 0) {
        noteArray[sensorboard][sensor] = noteArrayWickiHayden[sensor] + (currentOctave * 12) + (sensorboard * 12); // add 12 per higher octave
      }
      else if (currentLayout == 1) {
        noteArray[sensorboard][sensor] = noteArrayPiano[sensor] + (currentOctave * 12) + (sensorboard * 12); // add 12 per higher octave
      }
    }
  }
}

void enableSensorMeasurements(bool state) {
  for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) {
    Wire.beginTransmission(I2CSensorboardAddresses[sensorboard]); // transmit to device with address i
    if (state) {
      Wire.write('G'); // GH = Get Sensor Hit (start measurements)
      Wire.write('H'); // 
      enableRequestSensorData = true;
    }
    else {
      Wire.write('X'); // XS = execute stop measurements
      Wire.write('S'); // 
      enableRequestSensorData = false;
    }
    uint8_t returnValue = Wire.endTransmission(); // send I2C buffer 
    delay(1);
    if (returnValue > 0) { // someting went wrong
      Serial.print("return value of I2C transmission to "); Serial.print(sensorboard);Serial.print(" GS enableSensorMeasurements "); Serial.print(state);Serial.print(" = ");Serial.println(returnValue);
    }
  }
}

void executeReset() {
  for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) {
    Wire.beginTransmission(I2CSensorboardAddresses[sensorboard]); // transmit to device with address i
    Wire.write('X'); // XR = execute reset
    Wire.write('R'); // 
    enableRequestSensorData = false;
    uint8_t returnValue = Wire.endTransmission(); // send I2C buffer 
    if (returnValue > 0) { // someting went wrong
      Serial.print("return value of I2C transmission to "); Serial.print(sensorboard);Serial.print(" XR execute reset ");Serial.print(" = ");Serial.println(returnValue);
    }
  }
}

void executeAutoThreshold() {
  for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) {
    Wire.beginTransmission(I2CSensorboardAddresses[sensorboard]); // transmit to device with address i
    Wire.write('X'); // XR = execute auto thresold
    Wire.write('A'); // 
    enableRequestSensorData = false;
    uint8_t returnValue = Wire.endTransmission(); // send I2C buffer 
    if (returnValue > 0) { // someting went wrong
      Serial.print("return value of I2C transmission to "); Serial.print(sensorboard);Serial.print(" XA execute auto threshold ");Serial.print(" = ");Serial.println(returnValue);
    }
  }
}

void getSensorsThreshold() {
  uint16_t currentThreshold = 0;
  for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) {    
    for (uint8_t sensor = 0; sensor < sensorArrayLength; sensor++) {    
      Wire.beginTransmission(I2CSensorboardAddresses[sensorboard]); // transmit to device with address i
      Wire.write('G'); // GT = Get Sensor Threshold
      Wire.write('T'); // 
      Wire.write(sensor); // get info of this sensor number
      uint8_t returnValue = Wire.endTransmission(); // send I2C buffer 
      if (returnValue > 0) { // someting went wrong
        Serial.print("return value of I2C transmission to "); Serial.print(sensorboard);Serial.print(" GT getSensorsThreshold "); Serial.print(sensor);Serial.print(" = ");Serial.println(returnValue);
      }
      delay(10);
      Wire.requestFrom(I2CSensorboardAddresses[sensorboard], 2); // requesting 2 bytes from sensorboard
      currentThreshold = Wire.read() | (Wire.read() << 8); // currentThreshold is uint16_t, which is 2 bytes. So need to combine 2 reads into one uint16_t.
      Serial.print("current threshold of sensorboard ");Serial.print(sensorboard);Serial.print(" sensor ");Serial.print(sensor);Serial.print(" = ");Serial.println(currentThreshold);
      sensorThreshold[sensorboard][sensor] = currentThreshold; // set sensor threshold array item to received value
      delay(10);
    }
  }
}

void getSensorsMaximum() {
  uint16_t currentMaximum = 0;
  for (uint8_t sensorboard=0; sensorboard < numberOfSensorBoards; sensorboard++) {
    for (uint8_t sensor = 0; sensor < sensorArrayLength; sensor++) {    
      Wire.beginTransmission(I2CSensorboardAddresses[sensorboard]); // transmit to device with address i
      Wire.write('G'); // GT = Get Sensor maximum
      Wire.write('X'); // 
      Wire.write(sensor); // get info of this sensor number
      uint8_t returnValue = Wire.endTransmission(); // send I2C buffer 
      if (returnValue > 0) { // someting went wrong
        Serial.print("return value of I2C transmission to "); Serial.print(sensorboard);Serial.print(" GX getSensorsMaximum "); Serial.print(sensor);Serial.print(" = ");Serial.println(returnValue);
      }
      delay(10);
      Wire.requestFrom(I2CSensorboardAddresses[sensorboard], 2); // requesting 2 bytes from sensorboard
      currentMaximum = Wire.read() | (Wire.read() << 8); // currentMaximum is uint16_t, which is 2 bytes. So need to combine 2 reads into one uint16_t.
      Serial.print("current maximum of sensorboard ");Serial.print(sensorboard);Serial.print(" sensor ");Serial.print(sensor);Serial.print(" = ");Serial.println(currentMaximum);
      sensorMaximum[sensorboard][sensor] = currentMaximum; // set sensor Maximum array item to received value
      delay(10);
    }
  }
}

void setSensorThreshold(uint8_t sensorboard, uint8_t sensor, uint16_t setValue) {
  Wire.beginTransmission(I2CSensorboardAddresses[sensorboard]); // transmit to device with address i
  Wire.write('S'); // ST = Set Sensor maximum
  Wire.write('T'); // 
  Wire.write(sensor); // get info of this sensor number
  Wire.write(setValue & 0xFF); // first byte of setValue
  Wire.write((setValue >> 8) & 0xFF); // second byte of setValue
  uint8_t returnValue = Wire.endTransmission(); // send I2C buffer 
  if (returnValue > 0) { // someting went wrong
    Serial.print("return value of I2C transmission to "); Serial.print(sensorboard);Serial.print(" sensor ");Serial.print(sensor);Serial.print(" ST setSensorThreshold "); Serial.print(sensor);Serial.print(" = ");Serial.println(returnValue);
  }
  else {
    Serial.print("Succesfully set threshold ST of sensorboard ");Serial.print(sensorboard);Serial.print(" sensor ");Serial.print(sensor);Serial.print(" to ");Serial.println(setValue);
  }
}

void setSensorMaximum(uint8_t sensorboard, uint8_t sensor, uint16_t setValue) {
  Wire.beginTransmission(I2CSensorboardAddresses[sensorboard]); // transmit to device with address i
  Wire.write('S'); // SX = Set Sensor maximum
  Wire.write('X'); // 
  Wire.write(sensor); // get info of this sensor number
  Wire.write(setValue & 0xFF); // first byte of setValue
  Wire.write((setValue >> 8) & 0xFF); // second byte of setValue
  uint8_t returnValue = Wire.endTransmission(); // send I2C buffer 
  if (returnValue > 0) { // someting went wrong
    Serial.print("return value of I2C transmission to "); Serial.print(sensorboard);Serial.print(" sensor ");Serial.print(sensor);Serial.print(" SX setSensorMaximum "); Serial.print(sensor);Serial.print(" = ");Serial.println(returnValue);
  }
  else {
    Serial.print("Succesfully set maximum SX of sensorboard ");Serial.print(sensorboard);Serial.print(" sensor ");Serial.print(sensor);Serial.print(" to ");Serial.println(setValue);
  }
}

int mapToMidiVelocity(uint16_t input, uint16_t threshold, uint16_t maximum) {
    if(input < maximum) input = maximum; // make sure that input value cannot be lower than maximum value
    if(input > threshold) input = threshold; // make sure that input value cannot be lower than threshold value
    return (uint8_t)(127 - ((input - maximum) * 127) / (threshold - maximum)); // linear map threshold ~ maximum to 0 ~ 127
}

// First parameter is the event type (0x09 = note on, 0x08 = note off).
// Second parameter is note-on/note-off, combined with the channel.
// Channel can be anything between 0-15. Typically reported to the user as 1-16.
// Third parameter is the note number (48 = middle C).
// Fourth parameter is the velocity (64 = normal, 127 = fastest).

void noteOn(byte channel, byte pitch, byte velocity) {
  midiEventPacket_t noteOn = {0x09, 0x90 | channel, pitch, velocity};
  MidiUSB.sendMIDI(noteOn);
  // serializeMidiToUART(0x09, 0x90 | channel, pitch, velocity);
}

void noteOff(byte channel, byte pitch, byte velocity) {
  midiEventPacket_t noteOff = {0x08, 0x80 | channel, pitch, velocity};
  MidiUSB.sendMIDI(noteOff);
  // serializeMidiToUART(0x08, 0x80 | channel, pitch, velocity);
}

void serializeMidiToLED(byte event, byte channel, byte pitch, byte velocity) {  
  byte midiData[] = {channel, pitch, velocity};
  Serial.write(midiData, 3);
}

uint8_t scanI2Cbus() { 
  byte error, address;
  int nDevices;
  Serial.println("Scanning I2C bus ...");
  nDevices = 0;
  for(address = 1; address < 127; address++ )
  {
    Wire.beginTransmission(address);
    error = Wire.endTransmission();
    if (error == 0)
    {
      Serial.print("I2C device found at address 0x");
      if (address<16)
      { Serial.print("0"); }
      Serial.println(address,HEX);
      I2CSensorboardAddresses[nDevices] = address; // store address in array
      nDevices++;
    }
    else if (error==4)
    {
      Serial.print("Unknown error at address 0x");
      if (address<16)
      { Serial.print("0"); }
      Serial.println(address,HEX);
    }    
  }
  if (nDevices == 0) { Serial.println("No I2C devices found\n"); }
  else { Serial.println("I2C bus scanning done\n"); }
  return nDevices;
}

void checkAndSolveI2CDeadlock(){
  // Pin configuration during I2C initialisation are set as inputs.
  // PIN_WIRE_SDA (16u)-> PA16
  // PIN_WIRE_SCL (17u)-> PA17
  Serial.println("Checking I2C deadlock");
  // set PA16 to input, enabling pull resistor and setting it to pull-up.
  PORT->Group[0].DIRCLR.reg = PORT_PA16;
  PORT->Group[0].PINCFG[PIN_PA16].reg = PORT_PINCFG_PULLEN | PORT_PINCFG_INEN;
  PORT->Group[0].OUTSET.reg = PORT_PA16;
  // set PA17 to input, enabling pull resistor and setting it to pull-up.
  PORT->Group[0].DIRCLR.reg = PORT_PA17;
  PORT->Group[0].PINCFG[PIN_PA17].reg = PORT_PINCFG_PULLEN | PORT_PINCFG_INEN;
  PORT->Group[0].OUTSET.reg = PORT_PA17;

  if (~PORT->Group[0].IN.reg & PORT_PA16 ? 1 : 0) {
    Serial.println("I2C deadlock detected!");
    PORT->Group[0].DIRSET.reg = PORT_PA17; // set SCL pin as output
    //following code checks the bus state and recovers if needed
    while(~PORT->Group[0].IN.reg & PORT_PA16 ? 1 : 0) { // if the SDA line is low we clock the SCL line to free it
      PORT->Group[0].OUTCLR.reg = PORT_PA17; // pull SCL low
      delay(10);
      PORT->Group[0].OUTSET.reg = PORT_PA17; // pull SCL high
      delay(10);
    }
  }
}

void setup48MhzUSBClockRecoveryMode(){
  // from https://blog.thea.codes/understanding-the-sam-d21-clocks/
  /* Set the correct number of wait states for 48 MHz @ 3.3v */
  NVMCTRL->CTRLB.bit.RWS = 1;

  /* This works around a quirk in the hardware (errata 1.2.1) -
    the DFLLCTRL register must be manually reset to this value before
    configuration. */
  while(!SYSCTRL->PCLKSR.bit.DFLLRDY);
  SYSCTRL->DFLLCTRL.reg = SYSCTRL_DFLLCTRL_ENABLE;
  while(!SYSCTRL->PCLKSR.bit.DFLLRDY);

  /* Write the coarse and fine calibration from NVM. */
  uint32_t coarse =
      ((*(uint32_t*)FUSES_DFLL48M_COARSE_CAL_ADDR) & FUSES_DFLL48M_COARSE_CAL_Msk) >> FUSES_DFLL48M_COARSE_CAL_Pos;
  uint32_t fine =
      ((*(uint32_t*)FUSES_DFLL48M_FINE_CAL_ADDR) & FUSES_DFLL48M_FINE_CAL_Msk) >> FUSES_DFLL48M_FINE_CAL_Pos;

  SYSCTRL->DFLLVAL.reg = SYSCTRL_DFLLVAL_COARSE(coarse) | SYSCTRL_DFLLVAL_FINE(fine);

  /* Wait for the write to finish. */
  while (!SYSCTRL->PCLKSR.bit.DFLLRDY) {};

  SYSCTRL->DFLLCTRL.reg |=
      /* Enable USB clock recovery mode */
      SYSCTRL_DFLLCTRL_USBCRM |
      /* Disable chill cycle as per datasheet to speed up locking.
        This is specified in section 17.6.7.2.2, and chill cycles
        are described in section 17.6.7.2.1. */
      SYSCTRL_DFLLCTRL_CCDIS;

  /* Configure the DFLL to multiply the 1 kHz clock to 48 MHz */
  SYSCTRL->DFLLMUL.reg =
      /* This value is output frequency / reference clock frequency,
        so 48 MHz / 1 kHz */
      SYSCTRL_DFLLMUL_MUL(48000) |
      /* The coarse and fine values can be set to their minimum
        since coarse is fixed in USB clock recovery mode and
        fine should lock on quickly. */
      SYSCTRL_DFLLMUL_FSTEP(1) |
      SYSCTRL_DFLLMUL_CSTEP(1);

  /* Closed loop mode */
  SYSCTRL->DFLLCTRL.bit.MODE = 1;

  /* Enable the DFLL */
  SYSCTRL->DFLLCTRL.bit.ENABLE = 1;

  /* Wait for the write to complete */
  while (!SYSCTRL->PCLKSR.bit.DFLLRDY) {};      

  /* Setup GCLK0 using the DFLL @ 48 MHz */
  GCLK->GENCTRL.reg =
      GCLK_GENCTRL_ID(0) |
      GCLK_GENCTRL_SRC_DFLL48M |
      /* Improve the duty cycle. */
      GCLK_GENCTRL_IDC |
      GCLK_GENCTRL_GENEN;

  /* Wait for the write to complete */
  while(GCLK->STATUS.bit.SYNCBUSY) {};

  // for debugging: temporarily set 48MHz clock to GCLK4 as well and output to a pin to check accuracy.
  // GCLK->GENCTRL.reg =
  //     GCLK_GENCTRL_ID(4) |
  //     GCLK_GENCTRL_SRC_DFLL48M |
  //     GCLK_GENCTRL_IDC |
  //     GCLK_GENCTRL_GENEN |
  //     /* Enable outputting the clock to an I/O pin. */
  //     GCLK_GENCTRL_OE;

  // /* Wait for the write to complete */
  // while(GCLK->STATUS.bit.SYNCBUSY);
  // PORT->Group[1].DIRSET.reg = (1 << 10);
  // PORT->Group[1].PINCFG[10].reg |= PORT_PINCFG_PMUXEN;
  // PORT->Group[1].PMUX[10 >> 1].bit.PMUXE |= PORT_PMUX_PMUXE_H;
}

void setupHardwareTimers() 
{
    // Set up the generic clock (GCLK4) used to clock timers
    GCLK->GENDIV.reg = GCLK_GENDIV_DIV(1) |          // Divide the 48MHz clock source by divisor 1: 48MHz/1=48MHz
                       GCLK_GENDIV_ID(4);            // Select Generic Clock (GCLK) 4
    while (GCLK->STATUS.bit.SYNCBUSY);               // Wait for synchronization

    GCLK->GENCTRL.reg = GCLK_GENCTRL_IDC |           // Set the duty cycle to 50/50 HIGH/LOW
                        GCLK_GENCTRL_GENEN |         // Enable GCLK4
                        GCLK_GENCTRL_SRC_DFLL48M |   // Set the 48MHz clock source
                        GCLK_GENCTRL_ID(4);          // Select GCLK4
    while (GCLK->STATUS.bit.SYNCBUSY);               // Wait for synchronization

    // // Feed GCLK4 to TC3
    // GCLK->CLKCTRL.reg = GCLK_CLKCTRL_CLKEN |         // Enable GCLK4
    //                     GCLK_CLKCTRL_GEN_GCLK4 |     // Select GCLK4
    //                     GCLK_CLKCTRL_ID_TCC2_TC3;     // Feed the GCLK4 to TC3 (and TCC2)
    // while (GCLK->STATUS.bit.SYNCBUSY);               // Wait for synchronization

    // Feed GCLK4 to TC4 and TC5
    GCLK->CLKCTRL.reg = GCLK_CLKCTRL_CLKEN |         // Enable GCLK4
                        GCLK_CLKCTRL_GEN_GCLK4 |     // Select GCLK4
                        GCLK_CLKCTRL_ID_TC4_TC5;     // Feed the GCLK4 to TC4 and TC5
    while (GCLK->STATUS.bit.SYNCBUSY);               // Wait for synchronization

    // // specific for TC3: 1ms
    // TC3->COUNT16.CTRLA.reg |= TC_CTRLA_MODE_COUNT16;  // Set Timer counter Mode to 16 bits
    // while (TC3->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // TC3->COUNT16.CTRLA.reg |= TC_CTRLA_WAVEGEN_MFRQ;  // Set TC as Match Frequency
    // while (TC3->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // TC3->COUNT16.CC[0].reg = COUNTS_PER_TICK_1MS;     // Set the TC3 CC0 register as the TOP value in match frequency mode
    // while (TC3->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // // Enable the TC3 interrupt request
    // TC3->COUNT16.INTENSET.bit.MC0 = 1;
    // while (TC3->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // // Enable TC3
    // TC3->COUNT16.CTRLA.reg |= TC_CTRLA_ENABLE;        // Enable TC3
    // while (TC3->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // // Enable the TC3 interrupt vector use priority 2
    // NVIC_SetPriority(TC3_IRQn, 2);                    // Set the NVIC priority for TC3 to 2
    // NVIC_EnableIRQ(TC3_IRQn);

    // specific for TC4: 100us
    TC4->COUNT16.CTRLA.reg |= TC_CTRLA_MODE_COUNT16;  // Set Timer counter Mode to 16 bits
    while (TC4->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    TC4->COUNT16.CTRLA.reg |= TC_CTRLA_WAVEGEN_MFRQ;  // Set TC as Match Frequency
    while (TC4->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    TC4->COUNT16.CC[0].reg = COUNTS_PER_TICK_100US;     // Set the TC4 CC0 register as the TOP value in match frequency mode
    while (TC4->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable the TC4 interrupt request
    TC4->COUNT16.INTENSET.bit.MC0 = 1;
    while (TC4->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable TC4
    TC4->COUNT16.CTRLA.reg |= TC_CTRLA_ENABLE;        // Enable TC4
    while (TC4->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable the TC4 interrupt vector use priority 0
    NVIC_SetPriority(TC4_IRQn, 0);                    // Set the NVIC priority for TC4 to 0
    NVIC_EnableIRQ(TC4_IRQn);

    // specific for TC5
    TC5->COUNT16.CTRLA.reg |= TC_CTRLA_MODE_COUNT16;  // Set Timer counter Mode to 16 bits
    while (TC5->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    TC5->COUNT16.CTRLA.reg |= TC_CTRLA_WAVEGEN_MFRQ;  // Set TC as Match Frequency
    while (TC5->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    TC5->COUNT16.CC[0].reg = COUNTS_PER_TICK_TC5;     // Set the TC5 CC0 register as the TOP value in match frequency mode
    while (TC5->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable the TC5 interrupt request
    TC5->COUNT16.INTENSET.bit.MC0 = 1;
    while (TC5->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable TC5
    TC5->COUNT16.CTRLA.reg |= TC_CTRLA_ENABLE;        // Enable TC5
    while (TC5->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // use priority 1
    NVIC_SetPriority(TC5_IRQn, 1);                    // Set the NVIC priority for TC5 to 1
    NVIC_EnableIRQ(TC5_IRQn);    // Enable the TC5 timer (start checking sensorboards for hits)

}
