/*
 * I2C to USB Midi
 *
 * Wire.h on SAMD11C14 with ONE_UART_ONE_WIRE_NO_SPI uses
 * SAMD11CDEVBOARD pin 1 = 3.3V
 * SAMD11CDEVBOARD pin 2 = I2C SDA (PA14) SERCOM0
 * SAMD11CDEVBOARD pin 3 = I2C SCL (PA15) SERCOM0
 * SAMD11CDEVBOARD pin 4 = AIN2 (PA2)
 * SAMD11CDEVBOARD pin 5 = AIN0 (PA4)
 * SAMD11CDEVBOARD pin 6 = GND
 *
 * SERCOM0 is used for I2C
 * you cannot have TX on pad 1 or pad 3
 * see https://learn.adafruit.com/using-atsamd21-sercom-to-add-more-spi-i2c-serial-ports?view=all
 *
 * PA02 = pin AIN2 -> no sercom
 * PA04 = pin AIN0 -> SERCOM0 PAD[2] or alt SERCOM0 PAD[0]
 * PA05 = pin WS2812B-> SERCOM0 PAD[3] or alt SERCOM0 PAD[1]
 * PA08 = pin SWITCH -> SERCOM1 PAD[2] or alt SERCOM0 PAD[2]
 * PA09 = PIN LED -> SERCOM1 PAD[3] or alt SERCOM0 PAD[3]
 * PA14 = PIN I2C SDA -> SERCOM0 PAD[2] or alt SERCOM2 PAD[2]
 * PA15 = PIN I2C SCL -> SERCOM0 PAD[1] or alt SERCOM0 PAD[1]
 * PA24 = PIN USB D-
 * PA25 = PIN USB D+
 * PA28 = PIN RST
 * PA30 = PIN SWCLK -> SERCOM1 PAD[0] -> TX
 * PA31 = PIN SWDIO -> SERCOM1 PAD[1] -> RX
 *
** This work may be reproduced, modified, distributed, performed, and 
** displayed for any non-commercial, non-profit purpose, but must acknowledge 
** the developer and the Marimbatron project. 
** Copyright is retained and must be preserved. The work is 
** provided as is; no warranty is provided, and users accept all liability.
** 
** Copyright 2024 Leo Kuipers.
 */ 
#include "wiring_private.h"
#include <Wire.h>
#include "MIDIUSB.h"

// Uart(SERCOM *_s, uint8_t _pinRX, uint8_t _pinTX, SercomRXPad _padRX, SercomUartTXPad _padTX);
Uart pinSerial(&sercom1, PIN_PA31, PIN_PA30, SERCOM_RX_PAD_1, UART_TX_PAD_0); // Create the new UART instance

// timers
#define CPU_HZ 48000000 // The cpu clock is running at 48 MHz
#define TICK_HZ_100US 10000 // We want a 100 us tick.
#define COUNTS_PER_TICK_100US (CPU_HZ / TICK_HZ_100US) // This is the number of clock cycles per tick
#define TICK_HZ_ADC 100000 // We want a 1ms tick for the ADC.
#define COUNTS_PER_TICK_ADC (CPU_HZ / TICK_HZ_ADC) // This is the number of clock cycles per tick
volatile uint32_t micros_counter = 0; // holds the count of microseconds since the MCU started (don't want to use micros())
const unsigned long intervalEvent1 = 5000; // in microseconds
const unsigned long intervalEvent2 = 1000000; // in microseconds
unsigned long previousTimeEvent1 = 0;
unsigned long previousTimeEvent2 = 0;

#define LED PORT_PA09

void TC1_Handler() { // 100us timer, priority 0
  // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope
  micros_counter = micros_counter + 100;
  // Clear the interrupt
  TC1->COUNT16.INTFLAG.bit.MC0 = 1;
}

void TC2_Handler() { // 1 ms timer, priority 1
  // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope
  // I2CRequestData();
  // Clear the interrupt
  TC2->COUNT16.INTFLAG.bit.MC0 = 1;
}

void SERCOM1_Handler()
{
  pinSerial.IrqHandler();
}

void setup() {
  // set SAMD to run at 48Mhz
  setup48MhzUSBClockRecoveryMode();
  setupHardwareTimers();  
  Wire.begin(); // join i2c bus (address optional for master)
  Wire.setClock(100000);

  Serial.begin(9600);
  pinPeripheral(PIN_PA30, PIO_SERCOM); //Assign sercom function
  pinPeripheral(PIN_PA31, PIO_SERCOM); //Assign sercom function

  pinSerial.begin(31250);
  PORT->Group[0].DIRSET.reg = LED; // set led pin as output
  PORT->Group[0].OUTCLR.reg = LED; // turn led pin off
}

void loop() {
  unsigned long currentTime = micros_counter;
  /* This is event 1 -> check for pad hit */
  if ((unsigned long)(currentTime - previousTimeEvent1) >= intervalEvent1) {
    I2CRequestData();
   /* Update the timing for the next time around */
    previousTimeEvent1 = currentTime;
  }

  /* This is event 2 -> set LED colors based on incoming I2C master commands */
  if ((unsigned long)(currentTime - previousTimeEvent2) >= intervalEvent2) {
    PORT->Group[0].OUTTGL.reg = LED; // toggle led pin
   /* Update the timing for the next time around */
    previousTimeEvent2 = currentTime;
  }  
}

void I2CRequestData() {
  uint8_t sensorNumber;
  uint32_t sensorVelocity;

  Wire.requestFrom(1, 5);    // requesting 5 bytes from address 1. Max number of bytes in a single transmission cannot be more than 32.
  // Serial.println(Wire.available());
  while(Wire.available()) {
      sensorNumber = Wire.read(); // Wire.read reads 1 byte at a time. uint8_t is 1 byte.
      sensorVelocity = Wire.read() | (Wire.read() << 8) | (Wire.read() << 16) | (Wire.read() << 24); // sensorVelocity is uint32_t, which is 4 bytes. So need to combine 4 reads into one uint32_t.
      // Serial.print(sensorNumber);         // Print the value
      // Serial.print(" ");
      // Serial.println(sensorVelocity);
      // if (sensorNumber < 255) {
        // Serial.print("Hit on sensor "); Serial.print(sensorNumber); Serial.print(" -> "); Serial.println(sensorVelocity);
        // Serial.println("Sending note on");
        switch(sensorNumber) {
        case 0: noteOn(0, 48, mapToMidiVelocity(sensorVelocity, 3300, 2900));   // Channel 0, middle C, normal velocity
          break;
        case 1: noteOn(0, 49, mapToMidiVelocity(sensorVelocity, 2700, 1100));   // Channel 0, middle C, normal velocity
          break;
        case 2: noteOn(0, 50, mapToMidiVelocity(sensorVelocity, 2700, 800));   // Channel 0, middle C, normal velocity
          break;
        }        
      // if (sensorNumber < 255) {
      //   delay(10);
      //   noteOff(0, 48, 0);  // Channel 0, middle C, normal velocity
      // }
        // noteOn(0, sensorNumber, mapToMidiVelocity(sensorVelocity, 2700, 800));   // Channel 0, middle C, normal velocity
        // noteOn(0, 48, mapToMidiVelocity(sensorVelocity, 2700, 800));   // Channel 0, middle C, normal velocity
        // MidiUSB.flush();
        // Serial.println("Sending note off");
        // MidiUSB.flush();
      // }
      // else{
      //   Serial.println("no hit");
      // }
  }
}

int mapToMidiVelocity(int input, int threshold, int maximum) {
    if(input < maximum) input = maximum; // make sure that input value cannot be lower than maximum value
    if(input > threshold) input = threshold; // make sure that input value cannot be lower than threshold value
    return (int)(127 - ((input - maximum) * 127) / (threshold - maximum)); // linear map threshold ~ maximum to 0 ~ 127
}

// First parameter is the event type (0x09 = note on, 0x08 = note off).
// Second parameter is note-on/note-off, combined with the channel.
// Channel can be anything between 0-15. Typically reported to the user as 1-16.
// Third parameter is the note number (48 = middle C).
// Fourth parameter is the velocity (64 = normal, 127 = fastest).

void noteOn(byte channel, byte pitch, byte velocity) {
  midiEventPacket_t noteOn = {0x09, 0x90 | channel, pitch, velocity};
  MidiUSB.sendMIDI(noteOn);
  serializeMidiToLED(0x09, 0x90 | channel, pitch, velocity);
}

void noteOff(byte channel, byte pitch, byte velocity) {
  midiEventPacket_t noteOff = {0x08, 0x80 | channel, pitch, velocity};
  MidiUSB.sendMIDI(noteOff);
  serializeMidiToLED(0x08, 0x80 | channel, pitch, velocity);
}

void serializeMidiToLED(byte event, byte channel, byte pitch, byte velocity) {  
  byte midiData[] = {channel, pitch, velocity};
  pinSerial.write(midiData, 3);
  Serial.write(midiData, 3);
}

void setup48MhzUSBClockRecoveryMode(){
  // from https://blog.thea.codes/understanding-the-sam-d21-clocks/
  /* Set the correct number of wait states for 48 MHz @ 3.3v */
  NVMCTRL->CTRLB.bit.RWS = 1;

  /* This works around a quirk in the hardware (errata 1.2.1) -
    the DFLLCTRL register must be manually reset to this value before
    configuration. */
  while(!SYSCTRL->PCLKSR.bit.DFLLRDY);
  SYSCTRL->DFLLCTRL.reg = SYSCTRL_DFLLCTRL_ENABLE;
  while(!SYSCTRL->PCLKSR.bit.DFLLRDY);

  /* Write the coarse and fine calibration from NVM. */
  uint32_t coarse =
      ((*(uint32_t*)FUSES_DFLL48M_COARSE_CAL_ADDR) & FUSES_DFLL48M_COARSE_CAL_Msk) >> FUSES_DFLL48M_COARSE_CAL_Pos;
  uint32_t fine =
      ((*(uint32_t*)FUSES_DFLL48M_FINE_CAL_ADDR) & FUSES_DFLL48M_FINE_CAL_Msk) >> FUSES_DFLL48M_FINE_CAL_Pos;

  SYSCTRL->DFLLVAL.reg = SYSCTRL_DFLLVAL_COARSE(coarse) | SYSCTRL_DFLLVAL_FINE(fine);

  /* Wait for the write to finish. */
  while (!SYSCTRL->PCLKSR.bit.DFLLRDY) {};

  SYSCTRL->DFLLCTRL.reg |=
      /* Enable USB clock recovery mode */
      SYSCTRL_DFLLCTRL_USBCRM |
      /* Disable chill cycle as per datasheet to speed up locking.
        This is specified in section 17.6.7.2.2, and chill cycles
        are described in section 17.6.7.2.1. */
      SYSCTRL_DFLLCTRL_CCDIS;

  /* Configure the DFLL to multiply the 1 kHz clock to 48 MHz */
  SYSCTRL->DFLLMUL.reg =
      /* This value is output frequency / reference clock frequency,
        so 48 MHz / 1 kHz */
      SYSCTRL_DFLLMUL_MUL(48000) |
      /* The coarse and fine values can be set to their minimum
        since coarse is fixed in USB clock recovery mode and
        fine should lock on quickly. */
      SYSCTRL_DFLLMUL_FSTEP(1) |
      SYSCTRL_DFLLMUL_CSTEP(1);

  /* Closed loop mode */
  SYSCTRL->DFLLCTRL.bit.MODE = 1;

  /* Enable the DFLL */
  SYSCTRL->DFLLCTRL.bit.ENABLE = 1;

  /* Wait for the write to complete */
  while (!SYSCTRL->PCLKSR.bit.DFLLRDY) {};      

  /* Setup GCLK0 using the DFLL @ 48 MHz */
  GCLK->GENCTRL.reg =
      GCLK_GENCTRL_ID(0) |
      GCLK_GENCTRL_SRC_DFLL48M |
      /* Improve the duty cycle. */
      GCLK_GENCTRL_IDC |
      GCLK_GENCTRL_GENEN;

  /* Wait for the write to complete */
  while(GCLK->STATUS.bit.SYNCBUSY) {};

  // for debugging: temporarily set 48MHz clock to GCLK4 as well and output to a pin to check accuracy.
  // GCLK->GENCTRL.reg =
  //     GCLK_GENCTRL_ID(4) |
  //     GCLK_GENCTRL_SRC_DFLL48M |
  //     GCLK_GENCTRL_IDC |
  //     GCLK_GENCTRL_GENEN |
  //     /* Enable outputting the clock to an I/O pin. */
  //     GCLK_GENCTRL_OE;

  // /* Wait for the write to complete */
  // while(GCLK->STATUS.bit.SYNCBUSY);
  // PORT->Group[1].DIRSET.reg = (1 << 10);
  // PORT->Group[1].PINCFG[10].reg |= PORT_PINCFG_PMUXEN;
  // PORT->Group[1].PMUX[10 >> 1].bit.PMUXE |= PORT_PMUX_PMUXE_H;
}

void setupHardwareTimers() {
    // Set up the generic clock (GCLK4) used to clock timers
    GCLK->GENDIV.reg = GCLK_GENDIV_DIV(1) |          // Divide the 48MHz clock source by divisor 1: 48MHz/1=48MHz
                       GCLK_GENDIV_ID(4);            // Select Generic Clock (GCLK) 4
    while (GCLK->STATUS.bit.SYNCBUSY);               // Wait for synchronization

    GCLK->GENCTRL.reg = GCLK_GENCTRL_IDC |           // Set the duty cycle to 50/50 HIGH/LOW
                        GCLK_GENCTRL_GENEN |         // Enable GCLK4
                        GCLK_GENCTRL_SRC_DFLL48M |   // Set the 48MHz clock source
                        GCLK_GENCTRL_ID(4);          // Select GCLK4
    while (GCLK->STATUS.bit.SYNCBUSY);               // Wait for synchronization

    // // Feed GCLK4 to TC3
    // GCLK->CLKCTRL.reg = GCLK_CLKCTRL_CLKEN |         // Enable GCLK4
    //                     GCLK_CLKCTRL_GEN_GCLK4 |     // Select GCLK4
    //                     GCLK_CLKCTRL_ID_TCC2_TC3;     // Feed the GCLK4 to TC3 (and TCC2)
    // while (GCLK->STATUS.bit.SYNCBUSY);               // Wait for synchronization

    // Feed GCLK4 to TC1 and TC2
    GCLK->CLKCTRL.reg = GCLK_CLKCTRL_CLKEN |         // Enable GCLK4
                        GCLK_CLKCTRL_GEN_GCLK4 |     // Select GCLK4
                        GCLK_CLKCTRL_ID_TC1_TC2;     // Feed the GCLK4 to TC1 and TC2
    while (GCLK->STATUS.bit.SYNCBUSY);               // Wait for synchronization

    // specific for TC1: 100us
    TC1->COUNT16.CTRLA.reg |= TC_CTRLA_MODE_COUNT16;  // Set Timer counter Mode to 16 bits
    while (TC1->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    TC1->COUNT16.CTRLA.reg |= TC_CTRLA_WAVEGEN_MFRQ;  // Set TC as Match Frequency
    while (TC1->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    TC1->COUNT16.CC[0].reg = COUNTS_PER_TICK_100US;     // Set the TC1 CC0 register as the TOP value in match frequency mode
    while (TC1->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable the TC1 interrupt request
    TC1->COUNT16.INTENSET.bit.MC0 = 1;
    while (TC1->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable TC1
    TC1->COUNT16.CTRLA.reg |= TC_CTRLA_ENABLE;        // Enable TC1
    while (TC1->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable the TC1 interrupt vector use priority 0
    NVIC_SetPriority(TC1_IRQn, 0);                    // Set the NVIC priority for TC1 to 0
    NVIC_EnableIRQ(TC1_IRQn);

    // specific for TC2: 1ms
    TC2->COUNT16.CTRLA.reg |= TC_CTRLA_MODE_COUNT16;  // Set Timer counter Mode to 16 bits
    while (TC2->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    TC2->COUNT16.CTRLA.reg |= TC_CTRLA_WAVEGEN_MFRQ;  // Set TC as Match Frequency
    while (TC2->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    TC2->COUNT16.CC[0].reg = COUNTS_PER_TICK_ADC;     // Set the TC2 CC0 register as the TOP value in match frequency mode
    while (TC2->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable the TC2 interrupt request
    TC2->COUNT16.INTENSET.bit.MC0 = 1;
    while (TC2->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable TC2
    TC2->COUNT16.CTRLA.reg |= TC_CTRLA_ENABLE;        // Enable TC2
    while (TC2->COUNT16.STATUS.bit.SYNCBUSY);         // Wait for synchronization

    // Enable the TC2 interrupt vector use priority 1
    NVIC_SetPriority(TC2_IRQn, 1);                    // Set the NVIC priority for TC2 to 1
    NVIC_EnableIRQ(TC2_IRQn);
}