/*
 * I2C to USB Midi
 *
 * Wire.h on SAMD11C14 with NO_UART_ONE_WIRE_ONE_SPI uses
 * SAMD11CDEVBOARD pin 1 = 3.3V
 * SAMD11CDEVBOARD pin 2 = I2C SDA (PA14) SERCOM0
 * SAMD11CDEVBOARD pin 3 = I2C SCL (PA15) SERCOM0
 * SAMD11CDEVBOARD pin 4 = AIN2 (PA2)
 * SAMD11CDEVBOARD pin 5 = AIN0 (PA4)
 * SAMD11CDEVBOARD pin 6 = GND
 *
 * ONBOARD LED = PA09
 *
 ** This work may be reproduced, modified, distributed, performed, and 
** displayed for any non-commercial, non-profit purpose, but must acknowledge 
** the developer and the Marimbatron project. 
** Copyright is retained and must be preserved. The work is 
** provided as is; no warranty is provided, and users accept all liability.
** 
** Copyright 2024 Leo Kuipers.
 */ 
#include <Wire.h>
#include "MIDIUSB.h"
#include <SoftwareSerial.h>
#define LED PORT_PA09
#define I2C_CLOCK_SPEED 400000 // 400kHz
#define I2C_ADDRESS 0x01
#define DATA_LENGTH 5
uint8_t data[DATA_LENGTH];
// First parameter is the event type (0x09 = note on, 0x08 = note off).
// Second parameter is note-on/note-off, combined with the channel.
// Channel can be anything between 0-15. Typically reported to the user as 1-16.
// Third parameter is the note number (48 = middle C).
// Fourth parameter is the velocity (64 = normal, 127 = fastest).
void i2c_init(void) {
    // Enable the APB clock for SERCOM0
    PM->APBCMASK.reg |= PM_APBCMASK_SERCOM0;

    // Configure the GCLK for SERCOM0
    GCLK->CLKCTRL.reg = GCLK_CLKCTRL_ID_SERCOM0_CORE |
                        GCLK_CLKCTRL_GEN_GCLK0 |
                        GCLK_CLKCTRL_CLKEN;
    while (GCLK->STATUS.bit.SYNCBUSY);

    // Reset SERCOM0
    SERCOM0->I2CM.CTRLA.bit.SWRST = 1;
    while (SERCOM0->I2CM.CTRLA.bit.SWRST || SERCOM0->I2CM.SYNCBUSY.bit.SWRST);

    // Set up the SERCOM0 I2C peripheral
    SERCOM0->I2CM.CTRLB.reg = SERCOM_I2CM_CTRLB_SMEN; // Smart mode enable
    SERCOM0->I2CM.BAUD.reg = SystemCoreClock / (2 * I2C_CLOCK_SPEED) - 5; // Set the I2C clock speed
    SERCOM0->I2CM.CTRLA.reg = SERCOM_I2CM_CTRLA_MODE_I2C_MASTER | // Set the mode to I2C master
                              SERCOM_I2CM_CTRLA_ENABLE; // Enable the peripheral

    // Wait for synchronization
    while (SERCOM0->I2CM.SYNCBUSY.bit.ENABLE);

    // Set up the PA14 and PA15 pins for I2C
    PORT->Group[0].PINCFG[14].bit.PMUXEN = 1; // Enable the peripheral multiplexer for PA14
    PORT->Group[0].PMUX[7].bit.PMUXE = 2; // Set the multiplexer to peripheral function C (SERCOM)
    PORT->Group[0].PINCFG[15].bit.PMUXEN = 1; // Enable the peripheral multiplexer for PA15
    PORT->Group[0].PMUX[7].bit.PMUXO = 2; // Set the multiplexer to peripheral function C (SERCOM)
}

void i2c_request_data(void) {
    // Send the I2C start condition
    SERCOM0->I2CM.ADDR.reg = (I2C_ADDRESS << 1) | 0x1;

    // Wait for the data to be available
    for (int i = 0; i < DATA_LENGTH; i++) {
        while (!(SERCOM0->I2CM.INTFLAG.reg & SERCOM_I2CM_INTFLAG_SB));
        data[i] = SERCOM0->I2CM.DATA.reg;
    }

    // Send the I2C stop condition
    SERCOM0->I2CM.CTRLB.reg |= SERCOM_I2CM_CTRLB_CMD(3);
}

void noteOn(byte channel, byte pitch, byte velocity) {
  midiEventPacket_t noteOn = {0x09, 0x90 | channel, pitch, velocity};
  MidiUSB.sendMIDI(noteOn);
}

void noteOff(byte channel, byte pitch, byte velocity) {
  midiEventPacket_t noteOff = {0x08, 0x80 | channel, pitch, velocity};
  MidiUSB.sendMIDI(noteOff);
}

void setup() {
  // set SAMD to run at 48Mhz
  setup48MhzUSBClockRecoveryMode();
  // i2c_init();
  Wire.begin(); // join i2c bus (address optional for master)
  Wire.setClock(100000);

  Serial.begin(115200);
  PORT->Group[0].DIRSET.reg = LED; // set led pin as output
  PORT->Group[0].OUTCLR.reg = LED; // turn led pin off
}

void loop() {
  uint8_t sensorNumber;
  uint32_t sensorVelocity;
  PORT->Group[0].OUTTGL.reg = LED; // toggle led pin
  // i2c_request_data();
  // Serial.print("requesting data ... ");
  Wire.requestFrom(1, 5);    // requesting 5 bytes from address 1. Max number of bytes in a single transmission cannot be more than 32.
  // Serial.println(Wire.available());
    // Slave may send less than requested
  while(Wire.available()) {
      sensorNumber = Wire.read(); // Wire.read reads 1 byte at a time. uint8_t is 1 byte.
      sensorVelocity = Wire.read() | (Wire.read() << 8) | (Wire.read() << 16) | (Wire.read() << 24); // sensorVelocity is uint32_t, which is 4 bytes. So need to combine 4 reads into one uint32_t.
      Serial.print(sensorNumber);         // Print the value
      Serial.print(" ");
      Serial.println(sensorVelocity);
      if (sensorNumber < 255) {
        Serial.print("Hit on sensor "); Serial.print(sensorNumber); Serial.print(" -> "); Serial.println(sensorVelocity);
        Serial.println("Sending note on");
        noteOn(0, 48, 64);   // Channel 0, middle C, normal velocity
        MidiUSB.flush();
        delay(100);
        Serial.println("Sending note off");
        noteOff(0, 48, 64);  // Channel 0, middle C, normal velocity
        MidiUSB.flush();
      }
      else{
        Serial.println("no hit");
      }
  }
  // Serial.println();
  // Serial.println("Done");
// if (Wire.available() == 5) { // peripheral may send less than requested
  //   Serial.print("Wire available: "); Serial.println(Wire.available());
  //   sensorNumber = Wire.read(); // Wire.read reads 1 byte at a time. uint8_t is 1 byte.
  //   sensorVelocity = Wire.read() | (Wire.read() << 8) | (Wire.read() << 16) | (Wire.read() << 24); // sensorVelocity is uint32_t, which is 4 bytes. So need to combine 4 reads into one uint32_t.
  //   if (sensorNumber < 255) {
  //     Serial.print("Hit on sensor "); Serial.print(sensorNumber); Serial.print(" -> "); Serial.println(sensorVelocity);
  //   }
  //   else{
  //     Serial.println("no hit");
  //   }
  // }

  // Serial.println("Sending note on");
  // noteOn(0, 48, 64);   // Channel 0, middle C, normal velocity
  // MidiUSB.flush();
  delay(500);
  // Serial.println("Sending note off");
  // noteOff(0, 48, 64);  // Channel 0, middle C, normal velocity
  // MidiUSB.flush();
  // delay(1500);
}

void setup48MhzUSBClockRecoveryMode(){
  // from https://blog.thea.codes/understanding-the-sam-d21-clocks/
  /* Set the correct number of wait states for 48 MHz @ 3.3v */
  NVMCTRL->CTRLB.bit.RWS = 1;

  /* This works around a quirk in the hardware (errata 1.2.1) -
    the DFLLCTRL register must be manually reset to this value before
    configuration. */
  while(!SYSCTRL->PCLKSR.bit.DFLLRDY);
  SYSCTRL->DFLLCTRL.reg = SYSCTRL_DFLLCTRL_ENABLE;
  while(!SYSCTRL->PCLKSR.bit.DFLLRDY);

  /* Write the coarse and fine calibration from NVM. */
  uint32_t coarse =
      ((*(uint32_t*)FUSES_DFLL48M_COARSE_CAL_ADDR) & FUSES_DFLL48M_COARSE_CAL_Msk) >> FUSES_DFLL48M_COARSE_CAL_Pos;
  uint32_t fine =
      ((*(uint32_t*)FUSES_DFLL48M_FINE_CAL_ADDR) & FUSES_DFLL48M_FINE_CAL_Msk) >> FUSES_DFLL48M_FINE_CAL_Pos;

  SYSCTRL->DFLLVAL.reg = SYSCTRL_DFLLVAL_COARSE(coarse) | SYSCTRL_DFLLVAL_FINE(fine);

  /* Wait for the write to finish. */
  while (!SYSCTRL->PCLKSR.bit.DFLLRDY) {};

  SYSCTRL->DFLLCTRL.reg |=
      /* Enable USB clock recovery mode */
      SYSCTRL_DFLLCTRL_USBCRM |
      /* Disable chill cycle as per datasheet to speed up locking.
        This is specified in section 17.6.7.2.2, and chill cycles
        are described in section 17.6.7.2.1. */
      SYSCTRL_DFLLCTRL_CCDIS;

  /* Configure the DFLL to multiply the 1 kHz clock to 48 MHz */
  SYSCTRL->DFLLMUL.reg =
      /* This value is output frequency / reference clock frequency,
        so 48 MHz / 1 kHz */
      SYSCTRL_DFLLMUL_MUL(48000) |
      /* The coarse and fine values can be set to their minimum
        since coarse is fixed in USB clock recovery mode and
        fine should lock on quickly. */
      SYSCTRL_DFLLMUL_FSTEP(1) |
      SYSCTRL_DFLLMUL_CSTEP(1);

  /* Closed loop mode */
  SYSCTRL->DFLLCTRL.bit.MODE = 1;

  /* Enable the DFLL */
  SYSCTRL->DFLLCTRL.bit.ENABLE = 1;

  /* Wait for the write to complete */
  while (!SYSCTRL->PCLKSR.bit.DFLLRDY) {};      

  /* Setup GCLK0 using the DFLL @ 48 MHz */
  GCLK->GENCTRL.reg =
      GCLK_GENCTRL_ID(0) |
      GCLK_GENCTRL_SRC_DFLL48M |
      /* Improve the duty cycle. */
      GCLK_GENCTRL_IDC |
      GCLK_GENCTRL_GENEN;

  /* Wait for the write to complete */
  while(GCLK->STATUS.bit.SYNCBUSY) {};

  // for debugging: temporarily set 48MHz clock to GCLK4 as well and output to a pin to check accuracy.
  // GCLK->GENCTRL.reg =
  //     GCLK_GENCTRL_ID(4) |
  //     GCLK_GENCTRL_SRC_DFLL48M |
  //     GCLK_GENCTRL_IDC |
  //     GCLK_GENCTRL_GENEN |
  //     /* Enable outputting the clock to an I/O pin. */
  //     GCLK_GENCTRL_OE;

  // /* Wait for the write to complete */
  // while(GCLK->STATUS.bit.SYNCBUSY);
  // PORT->Group[1].DIRSET.reg = (1 << 10);
  // PORT->Group[1].PINCFG[10].reg |= PORT_PINCFG_PMUXEN;
  // PORT->Group[1].PMUX[10 >> 1].bit.PMUXE |= PORT_PMUX_PMUXE_H;
}