/*
** Firmware for Marimbatron Sensormodule
** Version 5
** 
** this firmware will:
** - blink the red and green leds
** - drive 1 WS2812B RGB led and cycle the colorwheel
** - read out 19 analog sensor inputs and output their value in the serial monitor
** - use ISR timers to get micros
** - use an ISR timer to read these sensor inputs
** - send data using I2C (Wire1)
** - Use DMA to send data to WS2812B WS2812Bs using PB10. Credits to https://www.lucadavidian.com/wifi-controlled-neo-pixels-strips/
**
** TODO: 
** - implement I2C communication protocol to communicate from main board to sensor board (i.e. set threshold values)
** - unroll ISR loops to speed things up (only if needed)
**
** Required libraries:
** wire.h
**
** This work may be reproduced, modified, distributed, performed, and 
** displayed for any non-commercial, non-profit purpose, but must acknowledge 
** the developer and the Marimbatron project. 
** Copyright is retained and must be preserved. The work is 
** provided as is; no warranty is provided, and users accept all liability.
** 
** Copyright 2024 Leo Kuipers.
*/

// timers
#include "Clocks.h"
volatile uint32_t micros_counter = 0; // holds the count of microseconds since the MCU started (don't want to use micros())

const unsigned long intervalEvent1 = 50000; // in microseconds
const unsigned long intervalEvent2 = 1000000; // in microseconds
unsigned long previousTimeEvent1 = 0;
unsigned long previousTimeEvent2 = 0;

// LEDs and WS2812B
#define GREENLEDPIN PORT_PB30 // PB30 = arduino 43
#define REDLEDPIN PORT_PB31 // PB31 = arduino 46
static bool flashLED;
const uint8_t ledOnRmax = 127;
const uint8_t ledOnGmax = 170;
const uint8_t ledOnBmax = 255;
#include "WS2812B.h" // 19 leds are defined in this file

// ADC
#include "ADC.c"

// array of the sensors, to be used for iterating
const int sensorArrayLength = 19;
const unsigned long sensorArray[sensorArrayLength] = 
{
  SENSOR0, SENSOR1, SENSOR2, SENSOR3, SENSOR4,
  SENSOR5, SENSOR6, SENSOR7, SENSOR8, SENSOR9,
  SENSOR10, SENSOR11, SENSOR12, SENSOR13, SENSOR14,
  SENSOR15, SENSOR16, SENSOR17, SENSOR18
};

#include "Sensor.h" // include sensor class
Sensor pads[sensorArrayLength]; // array that holds all sensor related data
const uint32_t DEFAULTTHRESHOLD = 2700; // 12-bit threshold, max value 4095.
const uint32_t DEFAULTSCANTIME = 2000; // scantime in us.
const uint32_t DEFAULTMASKTIME = 50000; // masktime in us.

// I2C communication
#include <Wire.h> 

void TC4_Handler() 
{ // 100us timer, priority 0
  // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope
  micros_counter = micros_counter + 100;
  // Clear the interrupt
  TC4->COUNT16.INTFLAG.bit.MC0 = 1;
}

void TC5_Handler() { // 200 us timer, priority 1
  // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope
  measureSensors();
  // Clear the interrupt
  TC5->COUNT16.INTFLAG.bit.MC0 = 1;
}

void measureSensors(void) 
{
  uint32_t padThreshold;
  uint32_t lastADCTime;
  uint32_t maskTime;
  unsigned long startScan;
  uint32_t currentMeasurement;
  bool maskTimeCheck;
  // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope

  for (int i = 0; i < sensorArrayLength; i++) { // iterate through all sensors 
    // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope
    startScan = micros_counter; // save current time
    lastADCTime = pads[i].getLastADCTime(); // save last 
    maskTime = pads[i].getMaskTime();
    // Next line is shorthand IF statement: checks if sensor should be masked
    // true if the measurement should not be masked. i.e. the measurement should be done.
    // startscan - lastADCTime >= masktime is done in this order to prevent issues from micros_counter overflow.
    // note (unsigned long) is typecasting the calculation between () to make sure the overflow prevention works.
    // see https://www.norwegiancreations.com/2018/10/arduino-tutorial-avoiding-the-overflow-issue-when-using-millis-and-micros/
    ((unsigned long)(startScan - lastADCTime) >= maskTime) ? maskTimeCheck = true : maskTimeCheck = false;

    if (maskTimeCheck) {
      // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope
      padThreshold = pads[i].getThreshold();
      setADCToPin(sensorArray[i]);
      currentMeasurement = readADC(true); // true = first measurement should be disregarded due to channel change
      if (currentMeasurement < padThreshold) { // measurement below threshold = hit detected. Start peak detection for scantime us.
        uint32_t scanTime = pads[i].getScanTime();
        uint32_t peakMeasurement = padThreshold;
        while ((unsigned long)(micros_counter - startScan) <= scanTime) { // stay measuring for as long as scanTime is. Again using micros_counter overflow prevention
          // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope
          currentMeasurement = readADC(false); // disregarding first measurement is not neccesary because we didn't change the input channel
          if (currentMeasurement < peakMeasurement) { // peak detection for local minimum
            peakMeasurement = currentMeasurement;
          }
        }
        // when scantime is done, store the peak value as velocity and set hit to true to indicate a valid hit is detected.
        pads[i].setVelocity(peakMeasurement); // peak detection done. Store peak value.
        pads[i].setHit(true); // set hit to true to indicate the pad was hit.
        pads[i].setLastADCTime(startScan + scanTime); // store lastADCTime, required for masktime calculation.
      } 
      else {
        if (!pads[i].getHit()) { // do not do this if there's a hit detected. Wait for hit to be set to false.
          pads[i].setVelocity(currentMeasurement); // save measurement (which is > threshold) as velocity, as to be able to show the value
        }
      }
    }
  }
}

void I2CEvent()
{ // send [255, 0, 0, 0, 0] when no hit is detected. Otherwise send sensor number and velocity of just 1 sensor and reset the hit state of that sensor
  uint32_t velocity;
  bool hitDetected = false;
  uint8_t data[] = {255,0,0,0,0};
  // Serial.println("Received I2C event");
  for (uint8_t i = 0; i < sensorArrayLength; i++) {
    if (pads[i].getHit()){ // if true: hit is detected
      hitDetected = true;
      velocity = pads[i].getVelocity();
      Serial.print("HIT s"); Serial.print(i); Serial.print(" ");Serial.print(velocity);
      Serial.println();
      // set sensor number and velocity to data variable when hit is detected
      data[0] = i;
      data[1] = velocity & 0xFF;
      data[2] = (velocity >> 8) & 0xFF;
      data[3] = (velocity >> 16) & 0xFF;
      data[4] = (velocity >> 24) & 0xFF;
      PORT->Group[1].OUTSET.reg = GREENLEDPIN; // turn green led pin on
      flashLED = true;
      pads[i].setHit(false); // reset hit
      break; // break the for loop
    }
  }
  Wire1.write(data, 5); // even when no hit is detected an answer will still be send (255, 0, 0, 0, 0)
}

void setup() 
{
  // set SAMD21 to run at 48Mhz
  setup48MhzUSBClockRecoveryMode();
  // setup 3 hardware timers
  setupHardwareTimers();

  // configure LED pins
  PORT->Group[1].DIRSET.reg = GREENLEDPIN | REDLEDPIN; // set led pins as output
  PORT->Group[1].OUTCLR.reg = REDLEDPIN; // turn red led pin off
  PORT->Group[1].OUTSET.reg = GREENLEDPIN; // turn green led pin on

  // configure WS2812B
  WS2812B_init();

  // configure ADCs
  setupADC();

  // configure I2C
  Wire1.begin(1); // start I2C with address 1
  Wire1.onRequest(I2CEvent); // register I2C event

  // configure serial port
  Serial.begin(115200); // enable serial communication  
  while (!Serial && micros_counter < 5000000); // wait for serial bus to activate (in order to communicate immediately after reset) or timeout
  delay(100);
  Serial.println(F("\nMarimbatron Sensor board firmware v5."));
  // configure pads
  setupPads();

  PORT->Group[1].OUTCLR.reg = GREENLEDPIN; // turn green led pin off

}

void loop() 
{
  unsigned long currentTime = micros_counter;
  uint32_t ADCMeasurement;
  static uint8_t eventCase = 0;

  static uint8_t ledOnR = 127;
  static uint8_t ledOnG = 170;
  static uint8_t ledOnB = 255;


  /* This is event 1 -> check for pad hit */
  if ((unsigned long)(currentTime - previousTimeEvent1) >= intervalEvent1) {
    PORT->Group[1].OUTTGL.reg = REDLEDPIN; // toggle red led pin
    if (flashLED) {
      PORT->Group[1].OUTCLR.reg = GREENLEDPIN; // turn green led pin off
      flashLED = false;
    }

   /* Update the timing for the next time around */
    previousTimeEvent1 = currentTime;
  }

  /* This is event 2 -> set LED colors based on incoming I2C master commands */
  if ((unsigned long)(currentTime - previousTimeEvent2) >= intervalEvent2) {
    ledOnR = ledOnR+30;
    ledOnG = ledOnG+30;
    ledOnB = ledOnB+30;
    if (ledOnR > ledOnRmax) ledOnR = 0;
    if (ledOnG > ledOnGmax) ledOnG = 0;
    if (ledOnB > ledOnBmax) ledOnB = 0;

    switch (eventCase)
    {
    case 0:
      WS2812B_set_pixel(0, ledOnR, 0, 0);
      WS2812B_set_pixel(1, 0, ledOnG, 0);
      WS2812B_set_pixel(2, 0, 0, ledOnB);
      WS2812B_set_pixel(3, ledOnR, 0, 0);
      WS2812B_set_pixel(4, 0, ledOnG, 0);
      WS2812B_set_pixel(5, 0, 0, ledOnB);
      WS2812B_set_pixel(6, ledOnR, 0, 0);
      WS2812B_set_pixel(7, 0, ledOnG, 0);
      WS2812B_set_pixel(8, 0, 0, ledOnB);
      WS2812B_set_pixel(9, ledOnR, 0, 0);
      WS2812B_set_pixel(10, 0, ledOnG, 0);
      WS2812B_set_pixel(11, 0, 0, ledOnB);
      WS2812B_set_pixel(12, ledOnR, 0, 0);
      WS2812B_set_pixel(13, 0, ledOnG, 0);
      WS2812B_set_pixel(14, 0, 0, ledOnB);
      WS2812B_set_pixel(15, ledOnR, 0, 0);
      WS2812B_set_pixel(16, 0, ledOnG, 0);
      WS2812B_set_pixel(17, 0, 0, ledOnB);
      WS2812B_set_pixel(18, ledOnR, 0, 0);
      WS2812B_update();  
      eventCase = 1;
      break;
    case 1:
      WS2812B_set_pixel(0, 0, ledOnG, 0);
      WS2812B_set_pixel(1, 0, 0, ledOnB);
      WS2812B_set_pixel(2, ledOnR, 0, 0);
      WS2812B_set_pixel(3, 0, ledOnG, 0);
      WS2812B_set_pixel(4, 0, 0, ledOnB);
      WS2812B_set_pixel(5, ledOnR, 0, 0);
      WS2812B_set_pixel(6, 0, ledOnG, 0);
      WS2812B_set_pixel(7, 0, 0, ledOnB);
      WS2812B_set_pixel(8, ledOnR, 0, 0);
      WS2812B_set_pixel(9, 0, ledOnG, 0);
      WS2812B_set_pixel(10, 0, 0, ledOnB);
      WS2812B_set_pixel(11, ledOnR, 0, 0);
      WS2812B_set_pixel(12, 0, ledOnG, 0);
      WS2812B_set_pixel(13, 0, 0, ledOnB);
      WS2812B_set_pixel(14, ledOnR, 0, 0);
      WS2812B_set_pixel(15, 0, ledOnG, 0);
      WS2812B_set_pixel(16, 0, 0, ledOnB);
      WS2812B_set_pixel(17, ledOnR, 0, 0);
      WS2812B_set_pixel(18, 0, ledOnG, 0);
      WS2812B_update();  
      eventCase = 2;
      break;
    case 2:
      WS2812B_set_pixel(0, 0, 0, ledOnB);
      WS2812B_set_pixel(1, ledOnR, 0, 0);
      WS2812B_set_pixel(2, 0, ledOnG, 0);
      WS2812B_set_pixel(3, 0, 0, ledOnB);
      WS2812B_set_pixel(4, ledOnR, 0, 0);
      WS2812B_set_pixel(5, 0, ledOnG, 0);
      WS2812B_set_pixel(6, 0, 0, ledOnB);
      WS2812B_set_pixel(7, ledOnR, 0, 0);
      WS2812B_set_pixel(8, 0, ledOnG, 0);
      WS2812B_set_pixel(9, 0, 0, ledOnB);
      WS2812B_set_pixel(10, ledOnR, 0, 0);
      WS2812B_set_pixel(11, 0, ledOnG, 0);
      WS2812B_set_pixel(12, 0, 0, ledOnB);
      WS2812B_set_pixel(13, ledOnR, 0, 0);
      WS2812B_set_pixel(14, 0, ledOnG, 0);
      WS2812B_set_pixel(15, 0, 0, ledOnB);
      WS2812B_set_pixel(16, ledOnR, 0, 0);
      WS2812B_set_pixel(17, 0, ledOnG, 0);
      WS2812B_set_pixel(18, 0, 0, ledOnB);
      WS2812B_update();  
      eventCase = 0;
      break;    
    }
   /* Update the timing for the next time around */
    previousTimeEvent2 = currentTime;
  }
}

void setupPads()
{
  for (int i = 0; i <= sensorArrayLength; i++) {
    pads[i].setThreshold(500); // temporarily setting this very low now because no sensor is attached.
    pads[i].setScanTime(DEFAULTSCANTIME);
    pads[i].setMaskTime(DEFAULTMASKTIME);
    pads[i].setLastADCTime(micros_counter);
  }
  pads[0].setThreshold(3300); // for testing, there is a sensor connected to sensor 8.
  pads[1].setThreshold(DEFAULTTHRESHOLD); // for testing, there is a sensor connected to sensor 8.
  pads[2].setThreshold(DEFAULTTHRESHOLD); // for testing, there is a sensor connected to sensor 8.

}
