/*
** Firmware for Marimbatron Sensormodule
** Version 6
** 
** this firmware will:
** - Blink all 19 WS2812B LEDs
** - read out 19 analog sensor inputs and output their value in the serial monitor
** - blink the red and green leds
** - send data using I2C (Wire1) when requested

** - use ISR timers to get micros
** - use an ISR timer to read these sensor inputs
** - Use DMA to send data to WS2812B WS2812Bs using PB10. Credits to https://www.lucadavidian.com/wifi-controlled-neo-pixels-strips/
**
** TODO: 
** V set PA14, PB15, PA15 as input, pullup
** V Set I2C address based on jumper 1, 3 and 4 (PA14, PB15, PA15) -> gnd = active
** V setup() -> measure current sensor values -> 15 times and take average, set as threshold: currentvalue - autoThreshold value (200?)
** V set sensor LEDs for wicki-hayden and for piano layout
** - lightup sensor LEDs on hit
** - store configurable values in flash
**
** - implement I2C communication protocol to communicate from main board to sensor board (i.e. set threshold values)
** - unroll ISR loops to speed things up (only if needed)
**
** Required libraries:
** wire.h
**
** This work may be reproduced, modified, distributed, performed, and 
** displayed for any non-commercial, non-profit purpose, but must acknowledge 
** the developer and the Marimbatron project. 
** Copyright is retained and must be preserved. The work is 
** provided as is; no warranty is provided, and users accept all liability.
** 
** Copyright 2024 Leo Kuipers.
*/

// timers
#define FIRMWAREVERSION 6
#include "Clocks.h"
volatile uint32_t micros_counter = 0; // holds the count of microseconds since the MCU started (don't want to use micros())

const unsigned long intervalEvent1 = 50000; // in microseconds
const unsigned long intervalEvent2 = 40000; // 40ms = 25x per second (in microseconds)
unsigned long previousTimeEvent1 = 0;
unsigned long previousTimeEvent2 = 0;

const uint8_t sensorArrayLength = 19;

// LEDs and WS2812B
#define GREENLEDPIN PORT_PB30 // PB30 = arduino 43
#define REDLEDPIN PORT_PB31 // PB31 = arduino 46
static bool flashLED;
const uint8_t ledOnRmax = 127;
const uint8_t ledOnGmax = 170;
const uint8_t ledOnBmax = 255;
bool sensorHit[sensorArrayLength];
#include "WS2812B.h" // 19 leds are defined in this file
typedef struct {
    uint8_t R;
    uint8_t G;
    uint8_t B;
} RGB;
const RGB colorLayoutWickiHayden[sensorArrayLength] = 
{
    {ledOnRmax, 0, 0}, // sensor 0
    {ledOnRmax, 0, 0}, // sensor 1
    {ledOnRmax, 0, 0}, // sensor 2
    {0, 0, ledOnBmax}, // sensor 3
    {0, 0, ledOnBmax}, // sensor 4
    {0, 0, ledOnBmax}, // sensor 5
    {0, 0, ledOnBmax}, // sensor 6
    {ledOnRmax, 0, 0}, // sensor 7
    {ledOnRmax, 0, 0}, // sensor 8
    {ledOnRmax, 0, 0}, // sensor 9
    {ledOnRmax, 0, 0}, // sensor 10
    {ledOnRmax, 0, 0}, // sensor 11
    {ledOnRmax, 0, 0}, // sensor 12
    {0, 0, ledOnBmax}, // sensor 13
    {0, 0, ledOnBmax}, // sensor 14
    {0, 0, ledOnBmax}, // sensor 15
    {ledOnRmax, 0, 0}, // sensor 16
    {ledOnRmax, 0, 0}, // sensor 17
    {ledOnRmax, 0, 0}  // sensor 18 
};
const RGB colorLayoutPiano[sensorArrayLength] = 
{
    {0, 0, 0}, // sensor 0
    {0, 0, ledOnBmax}, // sensor 1
    {0, 0, ledOnBmax}, // sensor 2
    {0, 0, ledOnBmax}, // sensor 3
    {0, 0, ledOnBmax}, // sensor 4
    {0, 0, ledOnBmax}, // sensor 5
    {0, 0, ledOnBmax}, // sensor 6
    {0, 0, ledOnBmax}, // sensor 7
    {0, 0, ledOnBmax}, // sensor 8
    {0, 0, 0}, // sensor 9
    {0, 0, 0}, // sensor 10
    {ledOnRmax, 0, 0}, // sensor 11
    {ledOnRmax, 0, 0}, // sensor 12
    {0, 0, 0}, // sensor 13
    {ledOnRmax, 0, 0}, // sensor 14
    {ledOnRmax, 0, 0}, // sensor 15
    {ledOnRmax, 0, 0}, // sensor 16
    {0, 0, 0}, // sensor 17
    {0, 0, 0} // sensor 18  
};
RGB targetSensorColor[sensorArrayLength]; // array that holds the color that the sensors should be
RGB currentSensorColor[sensorArrayLength]; // array that holds the color that the sensors are currently
const float colorTransitionSpeed = 0.3;
const uint8_t WS2812BtoSensorMap[sensorArrayLength] = 
{
 0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 1, 3, 5, 7, 9, 11, 13, 15, 17
};
uint8_t currentLayout = 0; // 0 = wicki hayden, 1 = piano
// ADC
#include "ADC.c"

// array of the sensors, to be used for iterating
const unsigned long sensorArray[sensorArrayLength] = 
{
  SENSOR0, SENSOR1, SENSOR2, SENSOR3, SENSOR4,
  SENSOR5, SENSOR6, SENSOR7, SENSOR8, SENSOR9,
  SENSOR10, SENSOR11, SENSOR12, SENSOR13, SENSOR14,
  SENSOR15, SENSOR16, SENSOR17, SENSOR18
};

#include "Sensor.h" // include sensor class
Sensor pads[sensorArrayLength]; // array that holds all sensor related data
const uint16_t DEFAULTTHRESHOLD = 2600; // 12-bit threshold, max value 4095.
const uint16_t DEFAULTSCANTIME = 2000; // scantime in us.
const uint16_t DEFAULTMASKTIME = 50000; // masktime in us.

// I2C communication
#include <Wire.h> 

// configurable variables, to be stored in flash
uint16_t autoThresholdValue = 200;


void TC4_Handler() 
{ // 100us timer, priority 0
  // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope
  micros_counter = micros_counter + 100;
  // Clear the interrupt
  TC4->COUNT16.INTFLAG.bit.MC0 = 1;
}

void TC5_Handler() { // 200 us timer, priority 1
  // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope
  measureSensors();
  // Clear the interrupt
  TC5->COUNT16.INTFLAG.bit.MC0 = 1;
}

void measureSensors(void) 
{
  uint16_t padThreshold;
  uint32_t lastADCTime;
  uint32_t maskTime;
  unsigned long startScan;
  uint16_t currentMeasurement;
  bool maskTimeCheck;
  // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope

  for (uint8_t i = 0; i < sensorArrayLength; i++) { // iterate through all sensors 
    // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope
    startScan = micros_counter; // save current time
    lastADCTime = pads[i].getLastADCTime(); // save last 
    maskTime = pads[i].getMaskTime();
    // Next line is shorthand IF statement: checks if sensor should be masked
    // true if the measurement should not be masked. i.e. the measurement should be done.
    // startscan - lastADCTime >= masktime is done in this order to prevent issues from micros_counter overflow.
    // note (unsigned long) is typecasting the calculation between () to make sure the overflow prevention works.
    // see https://www.norwegiancreations.com/2018/10/arduino-tutorial-avoiding-the-overflow-issue-when-using-millis-and-micros/
    ((unsigned long)(startScan - lastADCTime) >= maskTime) ? maskTimeCheck = true : maskTimeCheck = false;

    if (maskTimeCheck) {
      // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope
      padThreshold = pads[i].getThreshold();
      setADCToPin(sensorArray[i]);
      currentMeasurement = readADC(true); // true = first measurement should be disregarded due to channel change
      if (currentMeasurement < padThreshold) { // measurement below threshold = hit detected. Start peak detection for scantime us.
        uint32_t scanTime = pads[i].getScanTime();
        uint16_t peakMeasurement = padThreshold;
        while ((unsigned long)(micros_counter - startScan) <= scanTime) { // stay measuring for as long as scanTime is. Again using micros_counter overflow prevention
          // PORT->Group[1].OUTTGL.reg = TESTPIN; // toggle TEST pin to be able to check with oscilloscope
          currentMeasurement = readADC(false); // disregarding first measurement is not neccesary because we didn't change the input channel
          if (currentMeasurement < peakMeasurement) { // peak detection for local minimum
            peakMeasurement = currentMeasurement;
          }
        }
        // when scantime is done, store the peak value as velocity and set hit to true to indicate a valid hit is detected.
        pads[i].setVelocity(peakMeasurement); // peak detection done. Store peak value.
        pads[i].setHit(true); // set hit to true to indicate the pad was hit.
        sensorHit[i] = true; // save sensor hit status
        pads[i].setLastADCTime(startScan + scanTime); // store lastADCTime, required for masktime calculation.
        Serial.print("hit on s"); Serial.print(i); Serial.print(" ");Serial.print(peakMeasurement);Serial.println();
      } 
      else {
        if (!pads[i].getHit()) { // do not do this if there's a hit detected. Wait for hit to be set to false.
          pads[i].setVelocity(currentMeasurement); // save measurement (which is > threshold) as velocity, as to be able to show the value
        }
      }
    }
  }
}

void I2CEvent()
{ // send [255, 0, 0] when no hit is detected. Otherwise send sensor number and velocity of just 1 sensor and reset the hit state of that sensor
  uint16_t velocity;
  bool hitDetected = false;
//  uint8_t data[] = {255,0,0,0,0};
  uint8_t data[] = {255,0,0};
  // Serial.println("Received I2C event");
  for (uint8_t i = 0; i < sensorArrayLength; i++) {
    if (pads[i].getHit()){ // if true: hit is detected
      hitDetected = true;
      velocity = pads[i].getVelocity();
      Serial.print("send hit to I2C s"); Serial.print(i); Serial.print(" ");Serial.print(velocity);Serial.println();
      // set sensor number and velocity to data variable when hit is detected
      data[0] = i;
      data[1] = velocity & 0xFF;
      data[2] = (velocity >> 8) & 0xFF;
      // data[3] = (velocity >> 16) & 0xFF;
      // data[4] = (velocity >> 24) & 0xFF;
      // PORT->Group[1].OUTSET.reg = GREENLEDPIN; // turn green led pin on
      // flashLED = true;
      pads[i].setHit(false); // reset hit
      break; // break the for loop
    }
  }
  Wire1.write(data, 5); // even when no hit is detected an answer will still be send using sensor number 255 (255, 0, 0, 0, 0)
}

void setup() 
{
  uint8_t I2CAddress;
  bool I2CAddress0, I2CAddress1, I2CAddress2;

  // set SAMD21 to run at 48Mhz
  setup48MhzUSBClockRecoveryMode();
  // setup 3 hardware timers
  setupHardwareTimers();

  // configure LED pins
  PORT->Group[1].DIRSET.reg = GREENLEDPIN | REDLEDPIN; // set led pins as output
  PORT->Group[1].OUTCLR.reg = REDLEDPIN; // turn red led pin off
  PORT->Group[1].OUTSET.reg = GREENLEDPIN; // turn green led pin on

  // configure WS2812B
  WS2812B_init();

  // configure ADCs
  setupADC();

  // configure I2C
  // set PA14 to input, enabling pull resistor and setting it to pull-up.
  PORT->Group[0].DIRCLR.reg = PORT_PA14;
  PORT->Group[0].PINCFG[14].reg = PORT_PINCFG_PULLEN | PORT_PINCFG_INEN;
  PORT->Group[0].OUTSET.reg = PORT_PA14;
  // set PB15 to input, enabling pull resistor and setting it to pull-up.
  PORT->Group[1].DIRCLR.reg = PORT_PB15;
  PORT->Group[1].PINCFG[15].reg = PORT_PINCFG_PULLEN | PORT_PINCFG_INEN;
  PORT->Group[1].OUTSET.reg = PORT_PB15;
  // set PA15 to input, enabling pull resistor and setting it to pull-up.
  PORT->Group[0].DIRCLR.reg = PORT_PA15;
  PORT->Group[0].PINCFG[15].reg = PORT_PINCFG_PULLEN | PORT_PINCFG_INEN;
  PORT->Group[0].OUTSET.reg = PORT_PA15;

  I2CAddress0 = ~PORT->Group[0].IN.reg & PORT_PA14 ? 1 : 0; // read value of input pin. Input has pull-up so the actual value is inverted. this converts back.
  I2CAddress1 = ~PORT->Group[0].IN.reg & PORT_PA15 ? 1 : 0; // read value of input pin
  I2CAddress2 = ~PORT->Group[1].IN.reg & PORT_PB15 ? 1 : 0; // read value of input pin
  I2CAddress = ((I2CAddress2 << 2) | (I2CAddress1 << 1) | I2CAddress0) + 1; // combine values of input pins into one integer and add 1 (otherwise you can get the reserved address 0)
  
  Wire1.begin(I2CAddress); // start I2C with address based on jumpers

  // configure serial port
  Serial.begin(115200); // enable serial communication  
  while (!Serial && micros_counter < 5000000); // wait for serial bus to activate (in order to communicate immediately after reset) or timeout
  delay(100);
  Serial.print("Marimbatron Sensor board firmware v");
  Serial.println(FIRMWAREVERSION);
  Serial.print("I2C address = ");
  Serial.println(I2CAddress);

  // configure pads
  setupPads();
  autoThreshold();
  setLayoutColor(currentLayout); // set layout colors to wicki hayden

  for (uint8_t i = 0; i < sensorArrayLength; i++) { // sensorHit array is used to track which sensor was hit to be able to let that LED light up in different color. default state is false
    sensorHit[i] = false;
  }
  Wire1.onRequest(I2CEvent); // start listening for I2C events
  PORT->Group[1].OUTCLR.reg = GREENLEDPIN; // turn green led pin off

}

void loop() 
{
  unsigned long currentTime = micros_counter;
  /* This is event 1 -> check for pad hit */
  if ((unsigned long)(currentTime - previousTimeEvent1) >= intervalEvent1) {
    PORT->Group[1].OUTTGL.reg = REDLEDPIN; // toggle red led pin
    if (flashLED) {
      PORT->Group[1].OUTCLR.reg = GREENLEDPIN; // turn green led pin off
      flashLED = false;
    }

    for (uint8_t i = 0; i < sensorArrayLength; i++) {
      // Serial.print(" s"); Serial.print(i); Serial.print(" ");Serial.print(pads[i].getVelocity());
      if (sensorHit[i]){ // if true: hit is detected
        PORT->Group[1].OUTSET.reg = GREENLEDPIN; // turn green led pin on
        flashLED = true;
        currentSensorColor[i].R = ledOnRmax;
        currentSensorColor[i].G = ledOnGmax;
        currentSensorColor[i].B = ledOnBmax;
        // Serial.print("*"); // hit indicator
        sensorHit[i] = false; // reset hit
      }
      // else {
      //   Serial.print(" "); // add a white space for better alignment
      // }
    }
    // Serial.println();

   /* Update the timing for the next time around */
    previousTimeEvent1 = currentTime;
  }

  /* This is event 2 -> set LED colors based on incoming I2C master commands */
  if ((unsigned long)(currentTime - previousTimeEvent2) >= intervalEvent2) {
    updateSensorColors();
    /* Update the timing for the next time around */
    previousTimeEvent2 = currentTime;
  }
}

void setupPads()
{
  for (uint8_t i = 0; i < sensorArrayLength; i++) {
    pads[i].setScanTime(DEFAULTSCANTIME);
    pads[i].setMaskTime(DEFAULTMASKTIME);
    pads[i].setLastADCTime(micros_counter);
  }
}

void autoThreshold()
{
  // read sensor values when they're not pressed. subtract autoThresholdValue and save that as threshold for that sensor
  uint16_t averageSensorValue, measuredSensorValue;
  for (uint8_t i = 0; i < sensorArrayLength; i++) {
    setADCToPin(sensorArray[i]);
    measuredSensorValue = readADC(false); // false = first measurement is not disregarded
    measuredSensorValue = 0; // but we still disregard it.
    for (uint8_t j = 0; j < 15; j++) { // measuredSensorValue is uint16_t so max value is 65535. ADC is 12 bit, so max value is 4095. Max ADC value fits 16 times in a uint16_t. For safety taking 15 samples. 
        measuredSensorValue = measuredSensorValue + readADC(false); // false = first measurement is not disregarded
    }
    averageSensorValue = measuredSensorValue / 15;
    pads[i].setThreshold(averageSensorValue - autoThresholdValue);
    Serial.print("autoThreshold Pad ");
    Serial.print(i);
    Serial.print(" -> ");
    Serial.print(averageSensorValue);
    Serial.print(" - ");
    Serial.print(autoThresholdValue);
    Serial.print(" = ");
    Serial.println(averageSensorValue - autoThresholdValue);
  }
}

void setLayoutColor(uint8_t layout)
{
  if (layout == 0) {
    for (uint8_t i = 0; i < sensorArrayLength; i++){
      targetSensorColor[i].R = colorLayoutWickiHayden[i].R;
      targetSensorColor[i].G = colorLayoutWickiHayden[i].G;
      targetSensorColor[i].B = colorLayoutWickiHayden[i].B;
    }  
  }
  else if (layout == 1) {
    for (uint8_t i = 0; i < sensorArrayLength; i++){
      targetSensorColor[i].R = colorLayoutPiano[i].R;
      targetSensorColor[i].G = colorLayoutPiano[i].G;
      targetSensorColor[i].B = colorLayoutPiano[i].B;
    }  
  }
}

void transitionColor(RGB* current, RGB target, float speed) { // calculate interpolated color
    current->R += speed * (target.R - current->R);
    current->G += speed * (target.G - current->G);
    current->B += speed * (target.B - current->B);
}

void updateSensorColors()
{
  for (uint8_t i = 0; i < sensorArrayLength; i++){
    transitionColor(&currentSensorColor[i], targetSensorColor[i], colorTransitionSpeed);
    WS2812B_set_pixel(WS2812BtoSensorMap[i], currentSensorColor[i].R, currentSensorColor[i].G, currentSensorColor[i].B); // set light levels
  }
  WS2812B_update();  
}