/// \file Custom_SPI_Master_On_D21_Arduino_1.8.ino
/// 	\brief Sample code to support SPI over custom pins.
///
/// A SPI connection on a Atmel SAMD21 processor is supported by an on-chip versatile  
/// communications inderface, callled SERCOM.
/// There are six SERCOMs. Any one can be use for Serial Peripheral Interface(SPI).
/// Some SERCOMs are used for other purposes.
///
/// There are four connection points for each SERCOM, called pads.
///
/// Each pin on the Arduino board may be assigned to one of several
/// connection points. There is a different set of connection points for each pin.
/// A program controls the assignment through pinPeripheral() or pinMode().
///
/// There is a list of the possible values that can be assigned to the various pins
/// on the comments in a file called variant.cpp. You will find that file in the
/// Arduino libraries supplied for the SAMD21 models.
///
///
/// In this sample code, the SPI signals are connected to these pins :
///
#include <SERCOM.h>
#include <wiring_private.h>

#define MY_SERCOM sercom3

static const uint8_t MY_MISO = 7;	// sercom3 PAD[3]
static const uint8_t MY_MOSI = PIN_WIRE_SDA; // sercom3 PAD[0]
static const uint8_t MY_SCK = PIN_WIRE_SCL; // sercom3 PAD[1]
static const uint8_t MY_SS = 6;

static const EPioType MY_MOSI_PIOTYPE = PIO_SERCOM;
static const EPioType MY_MISO_PIOTYPE = PIO_SERCOM_ALT;
static const EPioType MY_SCK_PIOTYPE = PIO_SERCOM;

/// There are only three connections to the device.
/// The _SS line is controlled by the sketch directly as an output.

static const SercomSpiTXPad MY_TXPAD = SPI_PAD_0_SCK_1; // PAD[0] is output is MOSI
static const SercomRXPad MY_RXPAD = SERCOM_RX_PAD_3; 	// PAD[3] is input is MISO

#define BUFFER_SIZE 32
#define MY_SERCOM sercom3
const char *MASTER_MESSAGE = "|-- Message from master Tian --|";
char inputBuffer[BUFFER_SIZE];

void setup()
{
	Serial.begin ( 115200 );
	while ( !Serial ) ; // wait for port to open
	showMenu();	
	clearInputBuffer();
	
	pinPeripheral( MY_MISO, MY_MISO_PIOTYPE );
	pinPeripheral( MY_MOSI, MY_MOSI_PIOTYPE );
	pinPeripheral( MY_SCK, MY_SCK_PIOTYPE );
	pinMode( MY_SS, OUTPUT);
	digitalWrite( MY_SS, HIGH ); // HIGH means not selected
}

void loop()
{
	if ( Serial.available() )
	{
		uint8_t clu = Serial.read();
		Serial.print ( (char)clu );
		switch ( clu )
		{
			case 'M':
				sendMessage();
				break;
		}
		showMenu();
	}
}

void showMenu()
{
	Serial.println ( F("Custom SPI Master on D21 for Arduino 1.8") );
	Serial.println ( F("M: Send a test message") );
}

void sendMessage()
{	
	uint16_t count;
	
	// Start up SPI
	MY_SERCOM.resetSPI();
  	MY_SERCOM.initSPI(MY_TXPAD, MY_RXPAD, SPI_CHAR_SIZE_8_BITS, MSB_FIRST);
	MY_SERCOM.initSPIClock(SERCOM_SPI_MODE_0, 1000000);
	//	_description.sercom->initSPIClock(SERCOM_SPI_MODE_0, 4000000);
	MY_SERCOM.enableSPI();

	// Wake up slave
	digitalWrite ( MY_SS, LOW );
	delay(10);
	// Exchange data
	for ( count = 0; count < BUFFER_SIZE; count++ )
	{
		inputBuffer[count] = MY_SERCOM.transferDataSPI(MASTER_MESSAGE[count]);
		delay(1);
	}
	
	// Turn off SPI
	MY_SERCOM.resetSPI();
	
	// Tell slave to rest
	digitalWrite(MY_SS, HIGH);
	
	reportInputBuffer();
}

void clearInputBuffer()
{
	uint16_t count;
	for ( count = 0; count < BUFFER_SIZE; count++ )
		inputBuffer[count] = '?';
}

void reportInputBuffer()
{
	uint16_t count;
	Serial.println (F( "Buffer contents:" ));
	for ( count = 0; count < BUFFER_SIZE; count++ )
	{
		if ( inputBuffer[count] > 0x1F )
			Serial.print ( (char)inputBuffer[count] );
		else
			Serial.print ( "!" );
	}
	Serial.println ();
	for ( count = 0; count < BUFFER_SIZE; count++ )
	{
		Serial.print ( (uint8_t)inputBuffer[count], HEX );
		Serial.print ( " " );
	}
	clearInputBuffer();
	Serial.println ();
}

	

